#!/bin/sh
#
# backup_iSeries.sh
#
# ME    4623    040208  Main script which performs the iSeries backup either via a given
#                       profile which was created by config_iseries.sh or on demand.
# ME    4919    051208  Window flashes from Motif GUI... put a read in if On Demand.
# ME    4951    051208  if no profiles exist, we should handle more gracefully and send to config.
# ME    4993    051408  give different status based on number objs skipped.
# ME    5100    051908  check for notbackedup file size greater than zero.
# ME    5119    052008  Add backup types for on-demand on-time backups. 
# ME    5133    052108  If incremental requested, but never had a master, run a master
# ME    5156    052108  don't remove perm profile if tmp backup
# ME    5178    052708  redirect stderr from updatedb to /dev/null to hide Pass took x seconds.
# ME    5229    052908  set sync bit on backup if client marked as sync to vault
# ME    5257    060308  call iSeries_util.php to set elapsed time for backups.
# ME    5290    060508  Fix misspelling in profile Dialy
#

. /etc/default/bp.ini 2> /dev/null
if [ "$BPDIR" = "" ]
then
	export BPDIR="/usr/bp"
fi
BPBIN=$BPDIR/bin
BPLOG=$BPDIR/logs.dir
iSERIESLOG=$BPLOG/iSeries_backup$$.log
BPPRO=$BPDIR/profiles.dir
i5PRO=$BPPRO/iSeries
mkdir -p $i5PRO

ECHO="echo -e"


#########################################################################
#
# cleanup
#
# End state routine to clean up any temporary files and exit.
#
#########################################################################
cleanup()
{
	Status=$1
	# keep the number of iSeries backup logs from growing exponetially
	NumiSeriesLogs=`ls $BPLOG/iSeries_backup* 2> /dev/null | wc -l`
	while [ $NumiSeriesLogs -gt 20 ]
	do
		ls -tr $BPLOG/iSeries_backup* > /tmp/blogs$$ 2> /dev/null
		thisLog=`head -1 /tmp/blogs$$`
		rm -f $thisLog 2> /dev/null
		NumiSeriesLogs=`ls $BPLOG/iSeries_backup* 2> /dev/null | wc -l`
	done
	rm -f /tmp/blogs$$
	rm -f /tmp/bpt$$
	rm -f /tmp/err$$
	rm -f /tmp/.config_profile_name
	rm -f $iSERIESLOCK
	# 5119
	# 5156
	if [ "$OnDemand" = "true" -a "$RemoveProfile" != "no" ]
	then
		rm -f $i5PRO/$iSeriesProfile.spr
		rm -f $i5PRO/$iSeriesProfile.incl
		rm -f $i5PRO/$iSeriesProfile.excl
		rm -f $i5PRO/.${iSeriesProfile}_IPW
	fi
	stty sane
	exit $Status
}


#########################################################################
#
# profile_read
#
# Read the profile entry requested and echo it for assignment in calling
# routine
#
#########################################################################
profile_read()
{
	Section="$1"
	Keyword="$2"
	$BPBIN/bputil -g "$Section" "$Keyword" "xxx" $iSERIESPROFILE
}

#########################################################################
#
# 4951
# runConfigiSeries()
#
# If no profiles are present or user doesn't want to use one, run this 
# config code.
#
#########################################################################
runConfigiSeries()
{
	OnDemand="true"
	$ECHO "Would you like to save this backup as a new profile? <y or n - default n>: \c"
	read ANS
	if [ "$ANS" != "y" -a "$ANS" != "Y" ]
	then
		config_opt="-t"
	else
		config_opt="-s"
		RemoveProfile="no"
	fi
	$ECHO "`date`: On demand iSeries backup -- calling config_iSeries.sh." >> $iSERIESLOG
	$BPBIN/iSeries/config_iSeries.sh $config_opt
	RC=$?
	if [ $RC -ne 0 ]
	then
		$ECHO "Cannot perform backup because backup configuration has failed."
		$ECHO "Please address the reported problems and try again."
		$ECHO "`date`: Cannot perform backup because configuration failed." >> $iSERIESLOG
		cleanup 1
	fi
	iSERIESPROFILE=`cat /tmp/.config_profile_name`
	if [ ! -s $iSERIESPROFILE ]
	then
		$ECHO "Cannot perform backup because no profile has been found."
		$ECHO "`date`: Cannot perform backup because no profile has been found." >> $iSERIESLOG
		cleanup 1
	fi
}

#########################################################################
#
# getProfileFromList
#
# Allow user to select a profile which already exists in:
#	/usr/bp/profiles.dir/iSeries
#
#########################################################################
getProfileFromList()
{
	clear
	$ECHO "iSeries Profile Selection:"
	> /tmp/proflist$$
	# 4951
	ProfileList=`ls $i5PRO/*.spr 2> /dev/null`
	listItem=1
	for profile in $ProfileList
	do
		$ECHO "$listItem.  $profile" 
		$ECHO "$profile" >> /tmp/proflist$$
		listItem=`expr $listItem + 1`
	done
	# 4951 if no profiles exist, send them to the on-demand code.
	if [ $listItem -eq 1 ]
	then
		$ECHO "No pre-existing profiles found. Please manually configure your backup."
		runConfigiSeries
		return 0
	fi
	${ECHO} "Enter selection for profile: \c" 
	read ANS
	case ${ANS} in
	[0-9]*)
		SelectedProfile=`cat /tmp/proflist$$ |  awk -vthisfile="$ANS" '{if(NR == thisfile) print $1}'`
		if [ ! -r "$SelectedProfile" ]
		then
			$ECHO "Invalid Selection. Please try again."
			sleep 2
			getProfileFromList
		fi
		;;
	*)
		$ECHO "Invalid Selection. Please try again."
		sleep 2
		getProfileFromList
		;;
	esac
	iSERIESPROFILE=$SelectedProfile
	$ECHO "`date`: Users selects from profiles: $iSERIESPROFILE" >> $iSERIESLOG
	$ECHO "You have selected profile $iSERIESPROFILE."
	$ECHO "Contents of this profile are:"
	more $iSERIESPROFILE
	return
}

repair_instance_id()
{
	node_no=`psql -U postgres bpdb -tAc "SELECT node_no FROM bp.nodes WHERE node_name = '${iSeriesClient}'"`

	result=`psql bpdb -U postgres -tAc "SELECT app_id FROM bp.node_applications WHERE node_no = ${node_no}"`
	if [ -z ${result} ]
	then
		psql bpdb -U postgres -tAc "INSERT INTO bp.node_applications (node_no, app_id, current, visible) VALUES(${node_no}, 1, true, true)" >> $iSERIESLOG
	fi

	result=`psql bpdb -U postgres -tAc "SELECT instance_id FROM bp.application_instances WHERE node_no = ${node_no} AND app_id = 1"`
	if [ -z ${result} ]
	then
		psql bpdb -U postgres -tAc "INSERT INTO bp.application_instances (node_no, app_id, current, sync) VALUES(${node_no}, 1, true, true)" >> $iSERIESLOG
	fi

	psql bpdb -U postgres -c "UPDATE bp.backups SET instance_id = (SELECT instance_id FROM bp.application_instances WHERE node_no = ${node_no}) WHERE backup_no = ${backupNumber}" >> $iSERIESLOG
	psql bpdb -U postgres -c "UPDATE bp.jobs SET instance_id = (SELECT instance_id FROM bp.application_instances WHERE node_no = ${node_no}) WHERE job_no = ${jobNumber}" >> $iSERIESLOG
}
#
# MAIN PROGRAM
#
# iSeries backups will be placed in the Default backup device folder.
# iSeries_util.php -d returns the pathname for the D2D device
# directory.

iSERIESPROFILE=""
if [ $# -gt 1 ]
then
	while getopts 'p:' i; do
		case "$i" in
		p )	iSERIESPROFILE="$OPTARG"
			if [ ! -f "$iSERIESPROFILE" ]
			then
				if [ ! -f $i5PRO/$iSERIESPROFILE ]
				then
					$ECHO "Cannot locate profile $iSERIESPROFILE"
					$ECHO "${0}: Usage - $0 [-p <profile_name>]"
					exit 1
				else
					iSERIESPROFILE="$i5PRO/$iSERIESPROFILE"
				fi
			fi
		;;
		* )	$ECHO "${0}: Usage - $0 [-p <profile_name>]"
			exit 1
		;;
		esac
	done
fi


# If no profile is supplied on the command line we must be running an
# on-demand backup. Call config_iSeries.sh so a profile can be built.


if [ "$iSERIESPROFILE" = "" ]
then
	$ECHO "Would you like to select an existing profile for your backup? <y or n - default n>: \c"
	read ANS
	if [ "$ANS" = "y" -o "$ANS" = "Y" ]
	then
		getProfileFromList
		if [ $? -ne 0 ]
		then
			$ECHO "Cannot perform backup because user profile select failed."
			$ECHO "Please check $iSERIESLOG and correct any problems. You may need to recreate your profile."
			$ECHO "Press <Enter>: \c"
			read
			$ECHO "`date`: Cannot perform backup because configuration failed." >> $iSERIESLOG
			cleanup 1
		else
			OnDemand="true"
			# 5156
			RemoveProfile="no"
		fi
	else
		# 4951
		runConfigiSeries
	fi
	$ECHO "`date`: On demand iSeries backup -- Using profile $iSERIESPROFILE." >> $iSERIESLOG
fi

iSeriesClient=`profile_read "iSeriesBackup" "ClientName"`
iSERIESLOCK="/var/lock/$iSeriesClient"
if ! lockfile -r 0 $iSERIESLOCK 2> /dev/null
then
	$ECHO ""
	$ECHO "Processing already in progress for iSeries client $iSeriesClient. Cannot continue." 
	$ECHO "`date`: Processing already in progress for iSeries client $iSeriesClient. Cannot continue."  >> $iSERIESLOG
	sleep 5
	exit 1
fi

trap 'cleanup $?' EXIT

iSeriesProfile=`profile_read "iSeriesBackup" "ProfileName"`
# Needed for Migration
if [ "$iSeriesProfile" = "xxx" ]
then
	iSeriesProfile=$iSeriesClient
	$BPBIN/bputil -p "iSeriesBackup" "ProfileName" "$iSeriesProfile" $iSERIESPROFILE
fi


iSeriesBackupPath=`profile_read "iSeriesBackup" "DevicePath"`
iSeriesBackupName=`profile_read "iSeriesBackup" "DeviceName"`
iSeriesBackupStrategy=`profile_read "iSeriesBackup" "BackupStrategy"`
iSeriesExclList=`profile_read "iSeriesBackup" "ExcludeListFile"`
iSeriesInclList=`profile_read "iSeriesBackup" "IncludeListFile"`
iSeriesUserName=`profile_read "iSeriesBackup" "UserName"`
iSeriesSyncFlag=`profile_read "iSeriesBackup" "SyncBackups"`
export IPW=`cat $i5PRO/.${iSeriesProfile}_IPW`

iSeriesLogLevel=`$BPBIN/bputil -g debugging iseriesclient 0 $BPDIR/bpinit/master.ini`
MaxBlockSize=`$BPBIN/bputil -g iSeriesAgent MaxBlockSize 0 $BPDIR/bpinit/master.ini`
SavActWait=`$BPBIN/bputil -g iSeriesAgent WaitForCheckPoint 0 $BPDIR/bpinit/master.ini`
PreCheck=`$BPBIN/bputil -g iSeriesAgent PreCheck 0 $BPDIR/bpinit/master.ini`
Threading=`$BPBIN/bputil -g iSeriesAgent Threading 1 $BPDIR/bpinit/master.ini`
Timeout=`$BPBIN/bputil -g iSeriesAgent Timeout 8 $BPDIR/bpinit/master.ini`
LastMasterBackupTime=`$BPBIN/iSeries/iSeries_util.php -m $iSeriesClient`
Now=`date`
Today=`$ECHO $Now | awk '{print $1}'`
ThisBackupTime=`date --date="$Now" +%s`

$ECHO "`date`: iSeries backup -- Last master for $iSeriesClient was $LastMasterBackupTime." >> $iSERIESLOG

EListOpt=
IListOpt=
if [ "$iSeriesExclList" != "xxx" ]
then
	sed -i '/^$/d' $iSeriesExclList
	EListOpt="-e $iSeriesExclList"
fi

if [ "$iSeriesInclList" != "xxx" ]
then
	sed -i '/^$/d' $iSeriesInclList
	IListOpt="-i $iSeriesInclList"
fi

# 5119
# 5290
case "$iSeriesBackupStrategy" in
"WeeklyMastDailyIncr")
	if [ "$Today" = "Sat" ]
	then
		DateOption=""
		BackupType="master"
	else
		# 5133
		if [ "$LastMasterBackupTime" = "0" ]
		then
			DateOption=""
			BackupType="master"
		else
			DateOption="--date $LastMasterBackupTime"
			BackupType="incremental"
		fi
	fi
;;
"OnDemandMaster")
	DateOption=""
	BackupType="master"
;;
"OnDemandIncremental")
	# 5133
	if [ "$LastMasterBackupTime" = "0" ]
	then
		DateOption=""
		BackupType="master"
	else
		DateOption="--date $LastMasterBackupTime"
		BackupType="incremental"
	fi
;;
"DailySelective")
	BackupType="selective"
;;
"OnDemandSelective")
	BackupType="selective"
;;
esac

processID=$$
$BPBIN/bptool -j create $iSeriesClient $iSeriesBackupName $BackupType $processID > /tmp/bpt$$ 2>&1
bptoolRC=$?
cat /tmp/bpt$$ >> $iSERIESLOG
jobNumber=`cat /tmp/bpt$$ | grep "^job_id" 2> /dev/null | awk -F"=" '{print $2}'`
backupNumber=`cat /tmp/bpt$$ | grep "^backup_id" 2> /dev/null | awk -F"=" '{print $2}'`


ulimit -c unlimited
if [ $bptoolRC -eq 0 ]
then
	repair_instance_id
	$ECHO "`date`: iSeries backup beginning. This could take some time. Please wait...\c"
	iSeriesJobNumber=`cat /tmp/bpt$$ | grep "^Successfully" | awk -F"#" '{print $2}' | awk -F"," '{print $1}'`
	$ECHO "`date`: iSeries backup -- Creating job number $jobNumber backup number $backupNumber." >> $iSERIESLOG
	$BPBIN/bputil -j $jobNumber "Receiving files from $iSeriesClient."

	# logging transfer
	cat $iSERIESLOG > $BPLOG/iSeries_backup$[$backupNumber % 10].log
	rm $iSERIESLOG
	iSERIESLOG=$BPLOG/iSeries_backup$[$backupNumber % 10].log
	Catalog=$BPDIR/catalog.dir/iSeries_$iSeriesClient-$[$backupNumber % 10].catalog

	RCFILE=/tmp/rc$$
	(
		$BPBIN/backupFTP --server $iSeriesClient --backup ${iSeriesBackupPath}backup_$backupNumber $EListOpt $IListOpt $DateOption --credentials ${iSeriesUserName}:ENV:IPW --notbackedup 6 --maxblocksize $MaxBlockSize --wait $SavActWait --precheck $PreCheck --threading $Threading --Catalog $Catalog --verbose=$iSeriesLogLevel --timeout $Timeout >> $iSERIESLOG 2>&1 6> /tmp/err$$
		echo $? > $RCFILE
	) > /dev/null 2>&1 & 

	while [ ! -s $RCFILE ]
	do
		sleep 5
		backup_size=`ls -lh ${iSeriesBackupPath}backup_$backupNumber 2>/dev/null | awk '{print $5}'`
		if [ -n "$backup_size" ]
		then
			$BPBIN/bputil -j $jobNumber "Retrieved ${backup_size}B"
		fi
		if [ `psql bpdb -U postgres -tAc "select status & 32768 from bp.jobs where job_no = $jobNumber"` != 0 ]
		then
			pkill -15 -g $$ backupFTP
		fi
	done

	RC=`cat $RCFILE`
	rm -f $RCFILE

	sort -f $Catalog > /tmp/cat$$
	mv /tmp/cat$$ $Catalog

	$ECHO "\n"
	# 5100 only add the notbackedup file list if the file size if larger than zero
	if [ -s /tmp/err$$ ]
	then
		$ECHO "The following files were not backed up:" | tee -a $iSERIESLOG
		cat /tmp/err$$ | tee -a $iSERIESLOG
	fi
	case $RC in
	2) # 4993 backupFTP exits with 2 if too many files skipped
		$ECHO "The backup to $iSeriesClient FAILED with return code $RC." | tee -a $iSERIESLOG
		$ECHO "This indicates that too many files were skipped. You may wish to check permissions." | tee -a $iSERIESLOG
		$ECHO "For a list of skipped files, click on the failed backup in the User Interface." | tee -a $iSERIESLOG
		$BPBIN/iSeries/iSeries_util.php -j $backupNumber -b "`$ECHO \"These files were not backed up:\n\"` `cat /tmp/err$$`"
		$BPBIN/bputil -j $jobNumber "Running updatedb on iSeries backup number $backupNumber"
		$BPBIN/updatedb -p ${iSeriesBackupPath}backup_$backupNumber 2> /dev/null
		# 5257
		After=`date`
		BackupCompleteTime=`date --date="$After" +%s`
		ElapsedTime=`expr $BackupCompleteTime - $ThisBackupTime`
		$BPBIN/iSeries/iSeries_util.php -t $ElapsedTime $backupNumber
		if [ $? -ne 0 ]
		then
			$ECHO "Could not set elapsed time for backup." | tee -a $iSERIESLOG
			if [ "$OnDemand" = "true" ]
			then
				$ECHO "Please press <Enter> to return: \c"
				read
			fi
		fi
		$BPBIN/bputil -j $jobNumber "Backup $backupNumber failed. Cleaning up."
		$BPBIN/bptool -j update_status $jobNumber failed > /tmp/bpt$$ 2>&1
		# 4919
		if [ "$OnDemand" = "true" ]
		then
			$ECHO "Please press <Enter> to return: \c"
			read
		fi
		rm -f /tmp/err$$
		exit $RC
	;;
	1) # 4993 backupFTP exits with 1 if Success*
		$ECHO "The backup of $iSeriesClient has warnings with return code $RC." | tee -a $iSERIESLOG
		$ECHO "This indicates that too many files were skipped. You may wish to check permissions." | tee -a $iSERIESLOG
		$ECHO "For a list of skipped files, click on the failed backup in the User Interface." | tee -a $iSERIESLOG
		$BPBIN/bputil -j $jobNumber "Running updatedb on iSeries backup number $backupNumber"
		$BPBIN/updatedb -p ${iSeriesBackupPath}backup_$backupNumber 2> /dev/null
		# 5257
		After=`date`
		BackupCompleteTime=`date --date="$After" +%s`
		ElapsedTime=`expr $BackupCompleteTime - $ThisBackupTime`
		$BPBIN/iSeries/iSeries_util.php -t $ElapsedTime $backupNumber
		if [ $? -ne 0 ]
		then
			$ECHO "Could not set elapsed time for backup." | tee -a $iSERIESLOG
			if [ "$OnDemand" = "true" ]
			then
				$ECHO "Please press <Enter> to return: \c"
				read
			fi
		fi
		$BPBIN/bputil -j $jobNumber "Backup $backupNumber has warnings."
		$BPBIN/bptool -j update_status $jobNumber warning > /tmp/bpt$$ 2>&1
		# 5229 if client is syncable, set the sync bit on the backup.
		if [ "$iSeriesSyncFlag" = "y" ]
		then
			$BPBIN/iSeries/iSeries_util.php -s $backupNumber
		fi
		# 4919
		if [ "$OnDemand" = "true" ]
		then
			$ECHO "You may review status and see a file listing of this backup via the"
			$ECHO "User Interface."
			$ECHO "Please press <Enter> to return: \c"
			read
		fi
		rm -f /tmp/err$$
	;;
	0) # 4993 backupFTP exits with 0 if no files skipped
		
		$ECHO "The backup of $iSeriesClient is SUCCESSFUL." | tee -a $iSERIESLOG
		$ECHO "`date`: iSeries backup -- Complete." >> $iSERIESLOG
		$BPBIN/bputil -j $jobNumber "Running updatedb on iSeries backup number $backupNumber"
		$BPBIN/updatedb -p ${iSeriesBackupPath}backup_$backupNumber 2> /dev/null
		$BPBIN/bptool -j update_status $jobNumber success > /tmp/bpt$$ 2>&1
		# 5257
		After=`date`
		BackupCompleteTime=`date --date="$After" +%s`
		ElapsedTime=`expr $BackupCompleteTime - $ThisBackupTime`
		$BPBIN/iSeries/iSeries_util.php -t $ElapsedTime $backupNumber
		if [ $? -ne 0 ]
		then
			$ECHO "Could not set elapsed time for backup." | tee -a $iSERIESLOG
			if [ "$OnDemand" = "true" ]
			then
				$ECHO "Please press <Enter> to return: \c"
				read
			fi
		fi
		$ECHO "`date`: iSeries backup -- database update complete." >> $iSERIESLOG
		$BPBIN/bputil -j $jobNumber "Done."
		# 5229 if client is syncable, set the sync bit on the backup.
		if [ "$iSeriesSyncFlag" = "y" ]
		then
			$BPBIN/iSeries/iSeries_util.php -s $backupNumber
		fi
		# 4919
		if [ "$OnDemand" = "true" ]
		then
			$ECHO "You may review status and see a file listing of this backup via the"
			$ECHO "User Interface."
			$ECHO "Please press <Enter> to return: \c"
			read
		fi
		exit $RC
	;;
	143) # cancelled
		$ECHO "The backup to $iSeriesClient was cancelled." | tee -a $iSERIESLOG
		$BPBIN/bputil -j $jobNumber "Backup $backupNumber was cancelled."

		if [ "$OnDemand" = "true" ]
		then
			$ECHO "Please press <Enter> to return: \c"
			read
		fi
		rm -f /tmp/err$$
		exit $RC
	;;
	*) # backupFTP exits with 127 or something else happened
		$ECHO "The backup to $iSeriesClient FAILED with return code $RC." | tee -a $iSERIESLOG
		$ECHO "Please verify the client name and username-password  credentials you" | tee -a $iSERIESLOG
		$ECHO "supplied are correct and rerun this configuration utility." | tee -a $iSERIESLOG
		$BPBIN/iSeries/iSeries_util.php -j $backupNumber -b "`$ECHO \"These files were not backed up:\n\"` `cat /tmp/err$$`"
		$BPBIN/bputil -j $jobNumber "Backup $backupNumber failed. Cleaning up."
		$BPBIN/bptool -j update_status $jobNumber failed > /tmp/bpt$$ 2>&1
		# 4919
		if [ "$OnDemand" = "true" ]
		then
			$ECHO "Please press <Enter> to return: \c"
			read
		fi
		rm -f /tmp/err$$
		exit $RC
	;;
	esac
else
	$ECHO "The backup to $iSeriesClient FAILED with return code $bptoolRC." | tee -a $iSERIESLOG
	$ECHO "Please verify the client name and username-password  credentials you" | tee -a $iSERIESLOG
	$ECHO "supplied are correct and rerun this configuration utility." | tee -a $iSERIESLOG
	$BPBIN/bptool -j update_status $jobNumber failed > /tmp/bpt$$ 2>&1
	# 4919
	if [ "$OnDemand" = "true" ]
	then
		$ECHO "Please press <Enter> to return: \c"
		read
	fi
	exit $bptoolRC

fi
rm -f /tmp/bpt$$
