#!/usr/bin/perl

use strict;
use warnings;
# CentOS doesn't have the DateTime CPAN module available as a package
use Date::Parse;
use POSIX 'ceil';
use Fcntl ':mode';

# Reads a legacy (bkup) catalog file and transforms it into a comma-delimited
# file that matches the catalog format that accompanies XML archive profiles.

# This script also supports scanning a directory and reporting the same data;
# if done, it wants the directory as the archive path argument followed by an
# extra argument that is the pathname prefix for files printed, e.g. for a
# localdir archive: $0 /archives/LocalDir/etc /etc

# Output format: newline-delimited lists of comma-delimited (CSV-format) fields:
# 0: blk offset
# 1: blocks
# 2: directory
# 3: filename
# 4: type
# 5: size
# 6: mod_date

# Input looks like this:
#
# Contents of backup file backup_6.gz
# Permiss  Uid/Gid Bytes  Last Modified     Filename
# ============================================================================
# rw-rw-rw-  0/0      120 Nov 19 22:58 2006 C:/path/somefile.log, 1 blocks (Extended Properties)
# rw-rw-rw-  0/0   288844 Nov 19 22:58 2006 C:/path/somefile.log, 565 blocks
# rwxrwxrwx  0/0      144 Nov 19 22:58 2006 C:/path, 0 blocks^H^H^H^H^H^H^H^H (Directory)(Extended Properties)
# rwxrwxrwx  0/0        0 Nov 19 22:58 2006 C:/path, 0 blocks^H^H^H^H^H^H^H^H (Directory)
# ...

$\ = "\n";
$, = ' ';
my $bputil = "/usr/bp/bin/bputil";

my ($path, $prefix, $tape);
if ($#ARGV < 0) {
	my $name = $0;
	$name =~ s/.*\/([^\/]*)/$1/ if $name =~ /\//;
	print "Usage: ".$name." <archive path> [prefix if dir] [is_tape?0|1]";
	exit 1;
}
$path = shift;
$prefix = shift if (-d $path);
$tape = shift;

my $is_tape = 0;
if ($is_tape eq '1') {
	print "Tape not yet supported"; exit 1;
	$is_tape = 1;
}

# Don't die() on error; print to stdout and don't print the script line
if (!$path) {print "Error: an archive catalog path must be supplied"; exit 1; }
if (!open(FILE, "<$path")) {print "Error: cannot open archive catalog"; exit 1; }

&bkupList() if not defined $prefix;
&dirList() if defined $prefix;

sub bkupList()
{
	my $offset = 0;
	my $foundhdr = 0;
	my ($prev, $cur) = ('', '');
	while (<FILE>) {
		if (!$foundhdr) {
			$foundhdr = 1 if (substr($_, 0, 10) eq '==========');
			next;
		}
		# Process prev on each loop; check $cur for symlink type
		$prev = $cur;
		$cur = $_;
		next if ($prev eq '');
		# Offsets aren't accurate, so just count
		&printFile($prev, $cur, \$offset);
	}
	&printFile($cur, '', \$offset);
	close(FILE);
}

sub printFile()
{
	my ($this, $next, $offset_ref) = (shift, shift, shift);
	my ($fpath, $props, $date, $blocks, $size, $datestr, $type, $pos, $dir, $file);

	$_ = $this;
	return if (substr($_, 0, 22) eq 'Symbolic link to ===> ');

	# There's not always a space after the permissions if uid/gid are 3 digits
	if (length($_) < 10) { print STDERR "Bad line: $_"; return; }
	if (substr($_, 9, 1) ne ' ') { $_ = substr($_, 0, 9).' '.substr($_, 9); }
	if (length($next) >= 10 and substr($next, 9, 1) ne ' ') {
		$next = substr($next, 0, 9).' '.substr($next, 9);
	}

	my @arr = split(' ',$_,8);
	if ($#arr < 7) {
		print STDERR "Bad line: $_"; return;
	}

	$arr[7] =~ s///g;
	$arr[7] =~ m/(.*), ([0-9]+) blocks\s* (\([^\)]*\)\s*|linked to.*)*$/g;
	$fpath = $1;
	$blocks = $2;
	$size = $arr[2];
	$datestr = "${arr[3]} ${arr[4]} ${arr[5]} ${arr[6]}";

	$props = ((defined $3) ? $3 : '');
#	print "props: $3" if (defined $3);
	return if ($props =~ m/Extended Properties/);

	# Filter out duplicate entries that vary only in size/blocks
	my @nextarr = split(' ',$next,8);
	if ($#nextarr >= 7) {
		$nextarr[7] =~ s///g;
		$nextarr[7] =~ m/(.*), ([0-9]+) blocks\s* (\([^\)]*\)\s*|linked to.*)*$/g;
		my $nextfpath = $1;
		my $nextdatestr = "${nextarr[3]} ${nextarr[4]} ${nextarr[5]} ${nextarr[6]}";
		return if ($arr[0].' '.$arr[1].' '.$datestr.' '.$fpath eq
		           $nextarr[0].' '.$nextarr[1].' '.$nextdatestr.' '.$nextfpath);
	}

	$$offset_ref += 1;

	if (substr($_, 0, 22) eq 'Symbolic link to ===> ') { $type = ord('2') }
	elsif ($props eq '')                  { $type = ord('0') }
	elsif ($props =~ m/Directory/)        { $type = ord('5') }
	elsif ($props =~ m/linked to/)        { $type = ord('1') } # hardlink
	elsif ($props =~ m/Raw Device/)       { $type = ord('R') }
	elsif ($props =~ m/Raw Virtual file/) { $type = ord('S') }
	elsif ($props =~ m/Virtual file/)     { $type = ord('v') }
	elsif ($props =~ m/XFS ACL/)          { $type = ord('A') }
	elsif ($props =~ m/Registry Key/)     { $type = ord('K') }
	elsif ($props =~ m/Disk Image/)       { $type = ord('i') }
	elsif ($props =~ m/BareMetal/)        { $type = ord('i') }
	elsif ($props =~ m/Incomplete/)       { $type = ord('I') }
	elsif ($props =~ m/Resource fork/)    { $type = ord('S') }
	else                                  { $type = ord('0') } # regular file

	$pos = rindex($fpath, '/') + 1;
	$dir = substr($fpath, 0, $pos);
	$file = substr($fpath, $pos);

	$file =~ s/"/""/g;
	$dir =~ s/"/""/g;

	$date = str2time($datestr);
	$date = 0 if (!defined($date));
	print qq($$offset_ref,$blocks,"$dir","$file",$type,$size,$date);
}

sub dirList()
{
	my $prevfile = '';
	my $count = 0;
	$path =~ s/\/$//;
	$prefix =~ s/\/$//;
	my @flist = `find -P $path -printf '\%P\n'`;
	if ($? >> 8) { print STDERR "Could not execute `find` command"; return; }

	while ($_ = shift @flist) {
		chomp $_;
		my ($virtfile, @stat, $badprev, $pos);
		my ($blocks, $dir, $file,$type, $size, $date);

		$badprev = 0;
		if ($prevfile ne '') {
			# Some files have newlines in them; if we didn't find the last file
			# then try it with this name appended with a newline between them.
			$prevfile = "$prevfile\n$_";
			@stat = stat("$path/$prevfile");
			$_ = $prevfile if (@stat);
			$badprev = 1 if (!@stat);
		}
		@stat = stat("$path/$_") if (!@stat);
		if (!@stat) {
			$prevfile = $_;
			next;
		}

		print STDERR "Could not stat file: $prevfile" if $badprev;
		$prevfile = '';
		$type = S_IFMT($stat[2]);
		if    ($type == S_IFREG)  { $type = ord('0') }
		elsif ($type == S_IFDIR)  { $type = ord('5') }
		elsif ($type == S_IFLNK)  { $type = ord('2') }
		elsif ($type == S_IFBLK)  { $type = ord('4') }
		elsif ($type == S_IFCHR)  { $type = ord('3') }
		else                      { $type = ord('0') }

		if ($_ ne '') { $virtfile = "$prefix/$_"; }
		else          { $virtfile = $prefix;      }

		$count++;
		$pos = rindex("$virtfile", '/') + 1;
		$dir = substr("$virtfile", 0, $pos);
		$file = substr("$virtfile", $pos);
		$file =~ s/"/""/g;
		$dir =~ s/"/""/g;

		$blocks = ceil($stat[7] + 511);
		$size = $stat[7];
		$date = $stat[9];

		print qq($count,$blocks,"$dir","$file",$type,$size,$date);
	}
}

