#!/usr/bin/perl 

use strict;
use warnings;

use DBI;
use Fcntl ':mode';

######################################## Main begins ##########################
$\ = "\n";
$, = ' ';

if ($#ARGV < 6) {
	my $name = $0;
	$name =~ s/.*\/([^\/]*)/$1/ if $name =~ /\//;
	print "Usage: ".$name." <check|migrate> <migrate directory> <logFile> <debug> <syncUser> <dpuname> <encryptedVault> <clientlist>";
	print "Example: ".$name. " [check|migrate] /backups/VMware_UEB/syncd/migrate /backups/VMWARE_UEB/syncd/migrate/abc.log VMware_UEB VMware_UEB 1 winxp-64bit win2008sql";
	exit 1;
}

my $bpdir = `grep BPDIR= /etc/default/bp.ini | cut -d= -f2`;
$bpdir =~ s/\s+$//;
if ($bpdir eq '') {
	print "Failed to determine install directory";
	exit 1;
}
my $bputil = "$bpdir/bin/bputil";
my $archiverw = "$bpdir/bin/bputil --archiverw ";

my ($action, $migrationDir, $logFile, $DBG, $syncUser, $dpuName, $encryptedVault, @clients) = (shift, shift, shift, shift, shift, shift, shift, @ARGV);
if ($DBG eq '1') {
	print STDERR "Migration Home: $migrationDir";
	print STDERR "Sync User for migration: $syncUser";
	print STDERR "System to migrate: $dpuName";
	foreach my $client (@clients) {
		print STDERR "Client to migrate: $client";
	}
}

my $dbdump = "$migrationDir/dbdump.sql";
my $dpuClients = "$migrationDir/dpu_clients.rpt";
my $cryptMaster = "$migrationDir/cryptoDaemonMasterKeys";

# Don't die() on error; print to stdout and don't print the script line
if (!$migrationDir) {print "Error: migration home must be supplied"; exit 1; }
if (! -x $migrationDir) {print "Error: cannot open migration home " .$migrationDir; exit 1; }
if (! -e $dpuClients) { print "Error: cannot find backup information"; exit 1; }
if (! -e $dbdump) { print "Error: cannot find backup metadata"; exit 1; }

if ($action ne 'check' && $action ne 'migrate') {
	print "Error: Action has to be one of check or migrate";
	exit 1;
}
if ($action eq 'check') {
	if ($DBG eq 1) {
		print STDERR "Validating vaulted data for selected clients";
	}
	# We perform check per client. Client list should be of size 1
	if (@clients ne 1) {
		print "Usage: Only one client name is accepted during check";
		exit 1;
	}
}
my $reportFile = "$migrationDir/vaultMigration71.rpt";
my $decryptUtil = "$bpdir/bin/mr71DecryptUtil";
my $nonDecryptUtil = "$bpdir/bin/bpcrypt";

my $ENCRYPTED_BKUPS = 0;
my $ENCRYPTED_SIZE  = 0;
my $DECRYPTED_SIZE  = 0;
my $DECRYPTED_BKUPS = 0;
my $FATAL			= '';
my $MESSAGES		= '';

my $sqlUser = "postgres";
my $sqlDB = "bpdb";
my $P_SYNTHETIC = 0x00000001;
my $P_IMPORTED  = 0x00000020;

my $pid = $$; 
my $ST_SUCCESS  = 0x00000800;
my $ST_FAILED  = 0x00001000;
my $ST_JOB_FINAL = 0x00000040;
my $ST_SUMMARY = 0x00000020;
my $ST_COMPLETE = 0x00000400;
my $FINAL_SUCCESS = $ST_SUCCESS | $ST_JOB_FINAL | $ST_SUMMARY | $ST_COMPLETE;

&checkOrMigrateVaultedBackupInfo();
if ($action eq 'check') {
	# Caller wants these back in the same order
	printf STDOUT "ENCRYPTED_SIZE:$ENCRYPTED_SIZE:";
	printf STDOUT "ENCRYPTED_BKUPS:$ENCRYPTED_BKUPS:";
	printf STDOUT "DECRYPTED_SIZE:$DECRYPTED_SIZE:";
	printf STDOUT "DECRYPTED_BKUPS:$DECRYPTED_BKUPS:";
	printf STDOUT "FATAL:$FATAL:";
	printf STDOUT "MESSAGES:$MESSAGES";

}
if ($DBG) {
	print STDERR "Finished gathering vaulted data";
}
exit 0;

###############################################################################
sub checkOrMigrateVaultedBackupInfo() {
	# Read dpuclientsrpt to get data we need 
	# Additional backup detail is obtained from the dump file
	my @dpuclientsrpt;
	open (FILE, $dpuClients)
		and push @dpuclientsrpt, <FILE>
		and close FILE;

		if ($DBG) {
			 print STDERR "Loaded dpu clients report";
		}
	
	@dpuclientsrpt = map {chomp; $_ } @dpuclientsrpt;
	my $identity = '';
	my $role = '';		
	# For this source,  change role to 'replication source' and add identity
	if ($action eq 'migrate') {
		my $dbConn = DBI->connect("dbi:Pg:dbname=$sqlDB;host=$dpuName", 
									$sqlUser, '');
		if (!defined $dbConn) {
			print STDERR  "Failed to connect to $dpuName. Is the source active?";
			return;
		}
		my $result = $dbConn->prepare("SELECT LOWER(install_id) FROM bp.state");
		if (!defined $result) {
			print STDERR "Failed to get system identity for $syncUser";
			$dbConn->disconnect();
			return;
		}
		my $resultSql = $result->execute();
		if (!defined $resultSql) {
			print STDERR "Failed to get system identity for $syncUser";
			$result->finish();
			$dbConn->disconnect();
			return;
		}
		$identity = $result->fetchrow_array();
		$result->finish();
		$dbConn->disconnect();
		
		$role = 'Replication Source';	
		updateSystemClient($role, $identity);
	}
	foreach my $client (@clients) {
		# First check if $client is available on the source
		my $clientPresent = isClientPresentOnSource($client);
		if (!$clientPresent) {
			$FATAL .= "[$client] Client not found on source system $syncUser";
			next;
		}
		if ($DBG eq '1') {
			print STDERR "Looking for client [$client] from DPU Client Report.";
		}

		my $section = '';
		my @lines;
		my $clientFound = 0;
		my $legacySQLProcessed = 0; # Will have atmost one lastSQL per client

		for (my $i = 0; $i <= $#dpuclientsrpt; $i++) {
			my $line = $dpuclientsrpt[$i];
			my ($backupType, $srcFile, $startTime, $backupSize, $instId, 
				$backupNo, $application, $encrypted) = ('','',0,0,0,0,0,'');
			my ($legacyInstanceName, $legacyDBName, $legacyFileOrder) = 
												('', '','');
			my $backupSubType = ''; #For legacy SQL backups

			my ($modifyStartTime, $delta, $altFnameUsed, $found) = (0,0,0,0);
			my $deltaMast = '';
			my $fname = '';
			my @fields;
			
			next if $line =~ /^\s*$/;
			if ($line =~ /^\[(.*)\]/) {
				$section = $1;
				next;
			}

			# get file-level backup information
			if ($section eq $client and 
				$line =~ /\s*Item[0-9]+=([^\s]+[^\,]+)/) {
				my $appName = "file-level";
				if ($DBG eq '1') {
					print STDERR "$section: $client, value: $1";
				}
				@fields = split ',', $1;
				if ($#fields < 4) { 
					if ($DBG) {
						print STDERR "Invalid line"; 
					}
					next;
				}
				$found = 1;
				$clientFound = 1;
				$backupType = $fields[0];

				# PROCESS MASTER RECORD
				# SAMPLE: Item1=Master,/backups/D2Dbackups/,D2DBackups,2744,1355370640
				if ($backupType eq 'Master') {
					$backupSize = $fields[3];
					$startTime = $fields[4];
					$fname="lastMaster";
				}

				# PROCESS INCREMENTATL RECORD
				# SAMPLE: Item2=Incremental,/backups/D2Dbackups/,D2DBackups,61,1355372082
				if ($backupType eq 'Incremental') {
					$backupSize = $fields[3];
					$startTime = $fields[4];
					$fname="lastIncremental";
				}

				# PROCESS BARE METAL RECORD
				# SAMPLE: Item3=BareMetal,/backups/D2Dbackups/,D2DBackups,24168,1355499052
				if ($backupType eq 'BareMetal') {
					$backupSize = $fields[3];
					$startTime = $fields[4];
					$fname="lastBaremetal";
				}

				# SAMPLE: Item4=SQL,/backups/D2Dbackups/,WINXP\MSDE2000,db2000_1,Full,1360357205,D2DBackups,1
				if ($backupType eq 'SQL') {
					$legacyInstanceName = $fields[2];
					$legacyDBName = $fields[3];
					$backupSubType = $fields[4];
					$startTime = $fields[5];
					$legacyFileOrder = $fields[7];
					$fname = "lastSQL";
				} 

				#Get app_id for file-level instance 
				my $appId = getAppIdFromDump($appName);  
				if ($appId eq -1) {
					$FATAL .= "[$client] Failed to find valid application";
					$FATAL .= " Id for type $appName";
					next; #Skip this one.
				}
				$instId = getInstInfoFromDump($client, $appId);

				# Verify that the backup file exists:
				$srcFile = "/backups/$syncUser/syncd/$client/archives/$fname";
				
			} # Ends client section 

			# Process application level backup 
			if ($section eq "$client-applications" and 
				$line =~ /\s*Item[0-9]+=([^\s]+[^\,]+)/) {
				if ($DBG eq '1') {
					print STDERR "$section: $client, value: $1";
				}
				@fields = split ',', $1;
				if ($#fields < 6) { 
					if ($DBG) {
						print STDERR "Invalid line"; 
					} 
					next;
				}
				$found = 1;
				$clientFound = 1;
				$application = 1;
				$backupType = $fields[0];
				$backupNo = $fields[1];
				$backupSize = $fields[4];
				$startTime = $fields[5];
				$instId = $fields[6];
				$instId =~ s/\s+$//;
				if ($backupType =~ /[^\s]+Full/) {
					$fname = 'lastFull';
				} elsif ($backupType =~ /[^\s]+Diff/) {
					$fname = 'lastDifferential';
				} elsif ($backupType =~ /[^\s]+Trans/) {
					$fname = "trans-$backupNo";
				} elsif ($backupType =~ /[^\s]+Incr/) {
					$fname = "incr-$backupNo";
				}

				$srcFile = "/backups/$syncUser/syncd/$client/";
				$srcFile .= "appInstance/$instId/$fname";

			} # Ends client and applications section. 
			 
			if (!$found) {
				next; #Skip below, continue
			}

			# Clean up start time
			$startTime =~ s/\s+$//;
			# validate file, and check for encryption
			my @statSrcFile = stat("$srcFile");
			if (!@statSrcFile) {
				if ($backupType eq 'Incremental' ||$backupType =~ /[^\s]+Diff/) {
					# It could be a delta 
					$srcFile = $srcFile . '.delta';
					$altFnameUsed = 1;
					@statSrcFile = stat("$srcFile");
					if (!@statSrcFile) {
					  $MESSAGES .= "[$client] $srcFile does not exist. Skipping";
						next;
					}
				} else {
					$MESSAGES .= "[$client] $srcFile does not exist. Skipping";
					next;
				}
			}

			if ($backupType eq 'SQL') {
				#Get lastSQL file size
				$backupSize = $statSrcFile[7] / 1048576; # bytes to MB
			}

			# Check if already completed. ## moved to here after srcFile fixup bugid:16256
			if (alreadyCompleted($srcFile, $backupSize) eq 1) {
				if ($DBG) {
					print STDERR "File $srcFile already completed";
				}
				next;
			}

			if (!$altFnameUsed) {
				$encrypted = is_encrypted($srcFile);
			} else {
				# TODO: If delta we cannot determine if it is encrypted or not
				# We'll assume unencrypted
				$encrypted = 'n';
			}
			
			my $backupNumeric = getBackupTypeNumeric($backupType);
			if ($backupNumeric eq 0) {
				$FATAL .= "[$client] Failed to get valid backup type";
				$FATAL .= " value for $backupType";
				next;
			}
			
			# Legacy SQL is different. A single archive file lastSQL 
			# corresponds to multiple entries in the dpu_clients report. 
			# All of the line entries need to be processed at a single time
			if ($backupType eq 'SQL') {
				if (!$legacySQLProcessed) {
					#This is the first Item=SQL,.. entry
					checkOrMigrateLegacySQLBackups($client, $instId, 
													$backupSize, @dpuclientsrpt);
					$legacySQLProcessed = 1;
				} else {
					# first entry has already been seen. Skip the remaining.
					next;
				}
			} elsif ($application) {
				#Check if backup is on source
				if (!isBackupOnSource($backupNo)) {
					# Not present
					$backupNo = getLatestBackupNoFromSource(
												$backupNumeric, $instId, 0);
					$modifyStartTime = 1;
					if ($backupNo == 0) {
						$FATAL .= " [$client] Failed to get valid backup id";
						$FATAL .= " for type $backupType at $startTime";
						next; #Skipping
					}
				}
			} else {
				# Usual file-level. Contact source and get a backup number
				# for start_time, type, instance_id combination
				$backupNo = getBackupNoFromSource($backupNumeric, 
													$instId, $startTime);
				if ($backupNo == 0) {
					# We don't have backup for this time, get latest 
					$backupNo = getLatestBackupNoFromSource(
													$backupNumeric, $instId, 0);
					$modifyStartTime = 1;
					if ($backupNo == 0) {
						$FATAL .= " [$client] Failed to get valid backup id";
						$FATAL .= " for type $backupType at $startTime";
						next; #Skipping
					}
				}
			}
				
			# Print out what we found for this backup:
			if ($DBG eq '1') {
				print STDERR "Backup #$backupNo of client: $client";
				print STDERR "File Path     : $srcFile";
				print STDERR "Backup Type   : $backupType";
				print STDERR "Instance Info : $instId";
				print STDERR "Size          : $backupSize";
				print STDERR "Start Time    : $startTime";
				print STDERR "Encrypted     : $encrypted";
			}
			if ($action eq 'check') {
				if ($encrypted eq 'y') {
					$ENCRYPTED_SIZE += $backupSize;
					$ENCRYPTED_BKUPS++;
				} else {
					$DECRYPTED_SIZE += $backupSize;
					$DECRYPTED_BKUPS++;
				}
				next; # Skip processing
			}
				
			################# Check ends here. Migrate begins #################
			# Now we have a valid backup number, export blob, import it.
			if ($backupType eq 'SQL' && $legacySQLProcessed) {
				next; 
			}
			if ($DBG) { 
				print STDERR "Processing metadata for backup $backupNo";
			}

			my $exportedXML = exportBlobXML($backupNo, $srcFile, $backupSize);
			if ($exportedXML eq '') {
				# Failed. 
				next;
			}
			
			if ($modifyStartTime) {
				#TODO: Modify backup properties
				if ($DBG)	{
					print STDERR "Adjusting time stamp and size"; 
				}
				$exportedXML =~ s/<start_time>[0-9]+<\/start_time>/<start_time>$startTime<\/start_time>/g;
				$exportedXML =~ s/<total_megs>[0-9]+<\/total_megs>/<total_megs>$backupSize<\/total_megs>/g;
			}

			my $newBackupId = importBlobXML($exportedXML, $srcFile, $backupSize);
			if ($newBackupId <= 0) {
				next; #Failed. Move to next one
			}
			
			if (updateBackupPids($newBackupId) < 0) {
				# Do not fail. The backup is still good.
				print STDERR "Failed to update process id for backup $newBackupId. A status bit error message will be seen on the backup";
			}
			print STDOUT "[$client] Migration in progress";
			print STDOUT "$srcFile will be available as backup #$newBackupId";

			if (copyVaultedData($newBackupId, $srcFile,  
							$backupNumeric, $client, $instId, $backupSize) != 0 ||
				runUpdatedb($newBackupId,$srcFile, $backupSize) != 0) {
				next; #Failed. Move to next one
			}
			my $msg = "Complete at ".(localtime);
			if ($DBG) {
				print STDERR $msg;
			}
			updateResults(1, $srcFile, $backupSize, $msg);
		} #Ends looking at dpu_clients report 

		if (!$clientFound && $action eq 'check') {
			# Didn't find client in dpu_clients.rpt
			$FATAL .= "[$client] Client information not found in metadata";
		}

	} ### foreach my $client (@clients) 

	$role = 'DPU Configured for Vaulting';
	$identity = '';

	updateSystemClient($role, $identity);
}

###############################################################################
sub checkOrMigrateLegacySQLBackups() {
	my ($client, $instanceId, $lastSQLSize, @dpuclientsrpt) = 
							(shift, shift, shift, @_);
	my $section = '';

	my $firstSQL = 0;	
	my $result   = 0; # 1 is success
	my $resultMsg = '';
	# This file has already been checked.
	my $srcFile = "/backups/$syncUser/syncd/$client/archives/lastSQL";	
	for (my $i = 0; $i <= $#dpuclientsrpt; $i++) {
		my $line = $dpuclientsrpt[$i];
		my ($instanceName, $DBName, $sqlType, $startTime, $fileOrder, $fname)=
										( '','','',0,0,'');
		my @fields;
		my $modifyStartTime = 0;
		next if $line =~ /^\s*$/;
		if ($line =~ /^\[(.*)\]/) {
			$section = $1;
			next;
		}

		my $cmd = '';
		# get file-level backup information
		if ($section eq $client and 
			$line =~ /\s*Item[0-9]+=([^\s]+[^\,]+)/) {
			@fields = split ',', $1;
			if ($#fields < 7) { 
				next; #This will skip all non-SQL entries
			}
			my $backupType = $fields[0];
			if ($backupType ne 'SQL') {
				next;
			}
			$instanceName = $fields[2];
			$DBName = $fields[3];
			$sqlType = getSqlTypeNumeric($fields[4]);
			$startTime = $fields[5];
			$fileOrder = $fields[7];
			$fname = "lastSQL";
			if ($sqlType == 0) {
				print STDERR "Failed to get SQL type";
				next;
			}
			$firstSQL++;
			my $backupNo = getSQLBackupNoFromSource($instanceName,
									$DBName,$sqlType, $startTime, 
									$instanceId, $fileOrder);
			if ($backupNo == 0) {
				$backupNo = getLatestSQLBackupNoFromSource($instanceName,
									$DBName,$sqlType, $startTime, 
									$instanceId, $fileOrder);

				$modifyStartTime = 0;
				if ($backupNo == 0) {
					$FATAL .= " [$client] Failed to get valid backup id";
					$FATAL .= " for type $backupType at $startTime";
					$resultMsg = $FATAL;
					next; #Skipping
				}
			}
			
			if ($action eq 'check') {
				next; # Skip migration
			}
				
			######### Start migration #########
			print STDOUT "[$client] Migration of legacy SQL backups in progress";

			if ($firstSQL == 1) {
				# Extract lastSQL to a directory.
                                # check if src is encrypted (dictates what utility to read the backup)
				if (is_encrypted($srcFile) eq 'y') {
					$cmd = "$decryptUtil -m $cryptMaster -r $reportFile ";
				}
				else {
					$cmd = "$nonDecryptUtil ";
				}
				$cmd .= "-d $srcFile | $bpdir/bin/bkup -xf - -zWHERE=/backups/";
				$cmd .= "$syncUser/syncd/$client/archives/SQL_restore 2>>$logFile";
				if ($DBG) { print STDERR "migration-sql cmd: $cmd";}
				if (system($cmd) != 0) {
					print STDERR "Failed to extract lastSQL of $client";
					$resultMsg = "Failed to extract lastSQL of $client";
					last; # Error, don't try other line items from report.
				}
			}
			my $sqlPath = "/backups/$syncUser/syncd/$client/archives/";
			$sqlPath .= "SQL_restore/$client/sql";
			if ($DBG) {
				print STDERR "SQL Path is $sqlPath";
			}
			my $sqlSrcFile = '';
			if ($sqlType == 1) {
				$sqlSrcFile = "$sqlPath/$instanceName/$DBName/full";
			} elsif ($sqlType == 2) {
				$sqlSrcFile = "$sqlPath/$instanceName/$DBName/diff";
			} elsif ($sqlType == 4) {
				$sqlSrcFile = "$sqlPath/$instanceName/$DBName/trans_$fileOrder";
			}
			my @statSrcFile = stat("$sqlSrcFile");
			if (!@statSrcFile) {
				print STDOUT "$sqlSrcFile doesn't exist. Skipping";
				next;
			}
			my $sqlSize = $statSrcFile[7];
			my $exportedXML = exportBlobXML($backupNo, $srcFile, $lastSQLSize);
			if ($exportedXML eq '') {
				next;
			}
			if ($modifyStartTime) {
				#TODO: Modify backup properties?
				if ($DBG)	{
					print STDERR "Adjusting time stamp and size"; 
				}
				$exportedXML =~ s/<start_time>[0-9]+<\/start_time>/<start_time>$startTime<\/start_time>/g;
				$exportedXML =~ s/<total_megs>[0-9]+<\/total_megs>/<total_megs>$sqlSize<\/total_megs>/g;
			}
			my $newBackupId = importBlobXML($exportedXML, $srcFile, $lastSQLSize);
			if ($newBackupId <= 0) {
				next; #Failed. Move to next one
			}
			if (updateBackupPids($newBackupId) < 0) {
				# Do not fail. The backup is still good.
				print STDERR "Failed to update process id for backup $newBackupId. A status bit error message will be seen on the backup";
			}

			print STDOUT "Migrating $fields[4] of database $DBName belonging to $instanceName as backup #$newBackupId";
        		$sqlSrcFile =~ s/\\/\\\\/g;
			if (copyVaultedData($newBackupId, $sqlSrcFile,  
						$sqlType, $client, $instanceId, $lastSQLSize) != 0) {## ||
				##runUpdatedb($newBackupId,$sqlSrcFile, $lastSQLSize) != 0) {
				$resultMsg = "Failed to copy vaulted data.";
				next; #Failed. Move to next one
			}
			# Success
			$result = 1;
		} #End client section
	} # End $#dpuclientsrpt
	## TODO and SUCCESS
	# In the end, remove SQL_restore directory
	if ($action eq 'migrate') {
		my $cmd = "rm -rf /backups/$syncUser/syncd/$client/archives/SQL_restore 2>>$logFile";
		if (system($cmd) != 0) {
			print STDERR "Failed to remove temorary directory";
		}
		## update results
		updateResults($result, $srcFile, $lastSQLSize, $resultMsg);
	}
}

###############################################################################
sub updateBackupPids() {
	if (@_ != 1) {
		print STDERR "DBG WARNING! updateBackupPids accepts 1 arguments!";
		print STDERR "DBG @_";
	}
	my $backupId = shift;
	if ($DBG){
		print STDERR "Updating backup process id";
	}
	my $localDBConn = DBI->connect("dbi:Pg:dbname=$sqlDB", $sqlUser, '');
	if (!defined $localDBConn) {
		print STDERR "Failed to connect to database";
		return 1;
	}
	
	my $result = $localDBConn->prepare('UPDATE bp.backups SET pid = ? WHERE backup_no=?');
	if (!defined $result) {
		print STDERR "Failed to update process id for backup $backupId";
		$result->finish();
		$localDBConn->disconnect();
		return 1 ;
	}
	my $resultSql = $result->execute($pid, $backupId);
	if (!defined $resultSql) {
		print STDERR "Failed to update process id for backup $backupId";
		$result->finish();
		$localDBConn->disconnect();
		return 1;
	}
	$result->finish();
	$localDBConn->disconnect();
	return 0;
}

###############################################################################
sub getSQLBackupNoFromSource() {
	if (@_ != 6) {
		print STDERR "DBG WARNING! getSQLBackupNoFromSource accepts 6 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($instanceName, $DBName, $sqlType, $startTime, $instanceId, $fileOrder)=
		(shift, shift, shift, shift, shift, shift);
	my $id = 0;
	my $dbConn = DBI->connect("dbi:Pg:dbname=$sqlDB;host=$dpuName", $sqlUser, '');
	if (!defined $dbConn) {
		print STDERR "Failed to connect to database on $dpuName. Is the source active?";
		goto FINAL;		
	}
	my $sql = "SELECT s.sql_ref FROM bp.sql_backups AS s JOIN bp.backups AS b";
	$sql .= " ON s.sql_ref=b.backup_no WHERE b.start_time=$startTime AND";
	$sql .= " s.database_name='$DBName' AND s.sql_type=$sqlType AND b.instance_id=$instanceId";
	$sql .= " AND s.file_order_no=$fileOrder AND s.server_instance_name='$instanceName'";
	$sql .= " AND b.tape_no != 4294967295";
        $sql =~ s/\\/\\\\/g;
	if ($DBG) {
		print STDERR "getSQLBackupNoFromSource() sql: $sql";
	}

	my $result = $dbConn->prepare($sql);
	if (!defined $result) {
		print "Failed to find backup number for sql backup of type $sqlType";
		goto FINISH;
	}
	my $resultSql = $result->execute();
	if (!defined $resultSql) {
		print "Failed to find backup number for sql backup of type $sqlType";
		goto FINISH;
	}
	$id = $result->fetchrow_array();

FINISH:
	$result->finish();
FINAL:
	$dbConn->disconnect();
	return $id;
}

###############################################################################
sub getLatestSQLBackupNoFromSource() {
	if (@_ != 6) {
		print STDERR "DBG WARNING! getSQLBackupNoFromSource accepts 6 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($instanceName, $DBName, $sqlType, $startTime, $instanceId, $fileOrder)=
		(shift, shift, shift, shift, shift, shift);
	my $id = 0;
	my $dbConn = DBI->connect("dbi:Pg:dbname=$sqlDB;host=$dpuName", $sqlUser, '');
	if (!defined $dbConn) {
		print STDERR "Failed to connect to database on $dpuName. Is the source active?";
		goto FINAL;		
	}

	my $sql = "SELECT max(s.sql_ref) FROM bp.sql_backups AS s JOIN bp.backups AS b";
	$sql .= " ON s.sql_ref=b.backup_no WHERE ";
	$sql .= " s.database_name='$DBName' AND s.sql_type=$sqlType AND b.instance_id=$instanceId";
	$sql .= " AND s.file_order_no=$fileOrder AND s.server_instance_name='$instanceName' ";
	$sql .= " AND (b.status & $FINAL_SUCCESS) = $FINAL_SUCCESS AND (b.properties & $P_IMPORTED) = 0";
	$sql .= " AND (b.properties & $P_SYNTHETIC) = 0 AND b.tape_no != 4294967295";
	$sql =~ s/\\/\\\\/g;
        if ($DBG) {
                print STDERR "getLatestSQLBackupNoFromSource() sql: $sql";
        }
	
	my $result = $dbConn->prepare($sql);
	if (!defined $result) {
		print STDERR "Failed to get backup number for instance $instanceName";
		goto FINISH;
	}
	my $resultSql = $result->execute();
	if (!defined $resultSql) {
		print STDERR "Failed to get backup number for instance $instanceName";
		goto FINISH;
	}
	$id = $result->fetchrow_array();

FINISH:
	$result->finish();
FINAL:
	$dbConn->disconnect();
	return $id;
}

###############################################################################
sub getSqlTypeNumeric() {
	if (@_ != 1) {
		print STDERR "DBG WARNING! getSqlTypeNumeric accepts 1 arguments!\n";
		print STDERR "DBG @_";
	}
	my $sqlStr = shift;
	if ($sqlStr eq 'Full') {
		return 1;
	} elsif ($sqlStr eq 'Diff') {
		return 2;
	} elsif ($sqlStr eq 'Trans') {
		return 4;
	} else {
		return 0;
	}
}

###############################################################################
sub updateSystemClient() {
	if (@_ != 2) {
		print STDERR "DBG WARNING! updateSystemClient accepts 2 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($role, $identity) = (shift, shift);
	if ($action eq 'check') {
		return;
	}
	my $dbConn = DBI->connect("dbi:Pg:dbname=$sqlDB", $sqlUser, '');
	if (!defined $dbConn) {
		print STDERR  "Failed to connect to database";
		return;
	}
	my $sql = "UPDATE bp.systems SET role=?,identity=? WHERE name=?";
	my $result = $dbConn->prepare($sql);

	if (!defined $result) {
		print STDERR "Failed to update system role and identity for $syncUser";
		$dbConn->disconnect();
		return;
	}
	my $resultSql = $result->execute($role, $identity, $dpuName);
	if (!defined $resultSql) {
		print STDERR "Failed to update system role and identity for $syncUser";
		$result->finish();
		$dbConn->disconnect();
		return;
	}
}

###############################################################################
# Connects to source and exports the blob
# Params: backup number of the vaulted backup, as available on source
# Return: string containing metadata
sub exportBlobXML() {
	if (@_ != 3) {
		print STDERR "DBG WARNING! &exportBlobXML accepts 3 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($id, $srcFile, $size) = (shift, shift, shift);
	
	my $xml = `$bputil --remotecmd $dpuName "bputil --hactar_blob export $id"`;
	if (!defined $xml) {
		my $msg =  "Failed to export backup metadata for $id";
		updateResults(0, $srcFile, $size, $msg);
		return '';
	} else {
		$xml =~ s/^(?:.*\n){1,4}//;
#		$xml =~ s/(?:.*\n){1}/i/;
		return $xml;
	}
}

###############################################################################
# Imports the blob
# Params: XML string
# Return: backupi number, 0 otherwize
sub importBlobXML() {
	if (@_ != 3) {
		print STDERR "DBG WARNING! &importBlobXML accepts 3 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($xml,$srcFile, $size) = (shift, shift, shift);
	my $returnVal = 0;
	my $msg = '';

	# Get default device
	my $deviceNo = `$bputil -g 	\"Configuration Options\" \"DefaultTapeDevice\" 0 \"$bpdir/bpinit/master.ini\"`;
	$deviceNo =~ s/\s+$//;
	if ($deviceNo eq 0) {
		$msg = "Failed to get default device information";
		updateResults(0, $srcFile,$size, $msg);
		return 0;
	}
	if ($DBG) {
		print STDERR "Importing backup to device $deviceNo";
	}
	$xml =~ s/[\t ]+$//;
	$xml =~ s/^[\t ]+//;

	my $cmd = "$bputil --hactar_blob \"import\" \"$deviceNo\" \'$xml\'";
	if (!defined $returnVal) {
		$msg = "Failed to import backup metadata.";
		$msg .= " Backup record couldn't be created";
		updateResults(0, $srcFile, $size, $msg);
		return 0;

	}
	$returnVal = `$cmd`;
	$returnVal =~ s/[^0-9]//g;
	if ($DBG) {
		print STDERR "Backup number: $returnVal";
	}
	if (!defined $returnVal || $returnVal <= 0) {
		$msg = "Failed to import backup metadata.";
		$msg .= " Backup record couldn't be created";
		updateResults(0, $srcFile, $size, $msg);
		return 0;
	} else {
		return $returnVal;
	}
}

###############################################################################
# Transfer/Copy file
#Params: srcFile, backup_no
#Return: 0 on success

sub copyVaultedData() {
	if (@_ != 6) {
		print STDERR "DBG WARNING! &copyVaultedData accepts 6 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($id, $src, $backupNumeric, $client, $oldInstance, $size) = 
							(shift, shift, shift, shift, shift, shift);
	my $msg = '';	
	# Make a local connection
	my $localDBConn = DBI->connect("dbi:Pg:dbname=$sqlDB", $sqlUser, '');
	if (!defined $localDBConn) {
		print "Failed to connect to database";
		return 1;
	}
	
	my $result = $localDBConn->prepare('SELECT dev_nr_name FROM bp.devices WHERE tape_no=(SELECT tape_no FROM bp.backups WHERE backup_no= ?)');
	if (!defined $result) {
		print STDERR "Failed to get device name for backup #$id";
		$result->finish();
		$localDBConn->disconnect();
		updateBackupStatus($id, 1);
		return 1;
	}
	my $resultSql = $result->execute($id);
	if (!defined $resultSql) {
		print STDERR "Failed to get device name for backup #$id";
		$result->finish();
		$localDBConn->disconnect();
		updateBackupStatus($id, 1);
		return 1;
	}
	my $deviceDir = $result->fetchrow_array();
	$result->finish();
	
	$result = $localDBConn->prepare('SELECT instance_id FROM bp.backups WHERE backup_no= ?');
	if (!defined $result) {
		print STDERR "Failed to get instance id for #$id";
		$result->finish();
		$localDBConn->disconnect();
		updateBackupStatus($id, 1);
		return 1;
	}
	$resultSql = $result->execute($id);
	if (!defined $resultSql) {
		print STDERR "Failed to get instance id for #$id";
		$result->finish();
		$localDBConn->disconnect();
		updateBackupStatus($id, 1);
		return 1;
	}
	my $instanceId = $result->fetchrow_array();
	$result->finish();

	# Insert mypid as 
	$localDBConn->disconnect();

	my $dest = "$deviceDir/backup_$id";
	my $tempExists = 0;
	my $cmd = '';
	if (index($src, ".delta") > 0) {
		if ($DBG) {
			print STDERR "Patching delta file";
		}
		#  Patch delta files 
		my $parentFile = getParentBackupFile($backupNumeric, $instanceId, $deviceDir);
		if ($parentFile eq '') {
			# If master was not successful, we don't migrate differential
			$msg = "Master backup does not exist/is not succsesful";
			updateBackupStatus($id, 1);
			updateResults(0, $src, $size,  $msg);
			return 1;
		}
		if ($backupNumeric eq 2) {
			$parentFile = "/backups/$syncUser/syncd/$client/archives/lastMaster";
		} else {
			$parentFile = "/backups/$syncUser/syncd/$client/appInstance/$oldInstance/lastFull"
		}
		my $tempDiffFile = "/backups/$syncUser/syncd/$client/tempDifferentialFile";
		$cmd = "$bpdir/bin/bpsync -v -p $tempDiffFile -d $src $parentFile 2>>$logFile";
		if ($DBG) {
			print STDERR "Command: $cmd";
		}
		$result  = system($cmd);
		if ($result ne 0) {
			updateBackupStatus($id, 1);
			$msg = "Failed to create differential archive from delta";
			updateResults(0, $src, $size,  $msg);
			return 1;
		}
		$tempExists = 1;
		$src = $tempDiffFile; # Src will now be the newly patched file
	}
	my $isEncrypted = is_encrypted($src);
	
	# Pipe through archiverw (with smgr)
   	my $options = "cs";
    my $bufSize = "1048576"; ##TODO ok for all ?
	## Note, this rounds down, but compress should reduce the size we actually use
	my $sizeM = $size.'M'; 
	if ($isEncrypted eq 'y') {
		if ($encryptedVault eq 1 && 
			setBackupEncryptionProperty($id) eq 0) {
			# Encrypt backup only if we could update the backup property
			$options .="e";
		}
		#$cmd = "$decryptUtil -m $cryptMaster -r $reportFile -d $src > $dest 2>/dev/null";
		$cmd = "$decryptUtil -m $cryptMaster -r $reportFile -d $src | $archiverw w  $dest 0 $sizeM $bufSize $options 1>>$logFile 2>>$logFile";
	} else {
		#$cmd = "cp $src $dest";
		$cmd = "$nonDecryptUtil -d $src | $archiverw w $dest 0 $sizeM $bufSize $options 1>>$logFile 2>>$logFile";
	}
	if ($DBG) { 
		print STDERR "migration cmd: $cmd";
	}
	$result = system($cmd);
	if ($result eq 0) {
		updateBackupStatus($id, 0);
	} else {
		updateBackupStatus($id, 1);
		$msg = "Failed to write backup file for backup $id.";
		if ($isEncrypted eq 'y') {
			$msg .= " Possible reason: Failed to decrypt backup. Is the passphrase correct?"
		}
		if ($DBG) {
			print STDERR $msg;
		}
		updateResults(0, $src, $size, $msg);
	}
	if ($tempExists eq 1) {
		system("rm -rf /backups/$syncUser/syncd/$client/archives/tempDifferentialFile");
	}
	return $result;
}

###############################################################################
# Returns 0 on success, 1 if failed to set the property
sub setBackupEncryptionProperty() {
	if (@_ != 1) {
		print STDERR "DBG WARNING! &setBackupEncryptionProperty accepts 1 arguments!\n";
		print STDERR "DBG @_";
	}
	my $id = shift;
	my $returnVal = 1;

	# Set P_ENCRYPTED property 
	my $localDBConn = DBI->connect("dbi:Pg:dbname=$sqlDB", $sqlUser, '');
	if (!defined $localDBConn) {
		print STDERR "Failed to connect to database";
		return 1;
	}
	my $sql = "UPDATE bp.backups SET properties = (properties | ?) WHERE backup_no=?";
	my $result = $localDBConn->prepare($sql);
	if (!defined $result) {
		print STDERR "Failed to update encryption property for backup $id";
		goto FINISH;
	}
	my $P_ENCRYPTED = 0x00000100;
	my $resultSql = $result->execute($P_ENCRYPTED,$id);
	if (!defined $resultSql) {
		print STDERR "Failed to update encryption property for backup $id";
		goto FINISH;
	}
	$returnVal = 0;
FINISH:
	$result->finish();
	$localDBConn->disconnect();
	return $returnVal;
}

###############################################################################
sub updateBackupStatus() {
	if (@_ != 2) {
		print STDERR "DBG WARNING! &updateBackupStatus accepts 3 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($backupId, $failed) = (shift, shift);
	my $status = 0;
	if ($failed eq 1) {
		if ($DBG) {
			print STDERR "Updating status of backup #$backupId to failed";
		}
		$status = $ST_JOB_FINAL | $ST_COMPLETE | $ST_FAILED;
	} else {
		if ($DBG) {
			print STDERR "Updating status of backup #$backupId to success";
		}
		$status = $FINAL_SUCCESS;
	}
	my $localDBConn = DBI->connect("dbi:Pg:dbname=$sqlDB", $sqlUser, '');
	if (!defined $localDBConn) {
		print STDERR "Failed to connect to database";
		return;
	}
	
	my $result = $localDBConn->prepare('UPDATE bp.backups SET status=? WHERE backup_no=?');
	if (!defined $result) {
		print STDERR "Failed to update backup status";
		$result->finish();
		$localDBConn->disconnect();
		return;
	}
	my $resultSql = $result->execute($status, $backupId);
	if (!defined $resultSql) {
		print STDERR "Failed to update backup status";
		$result->finish();
		$localDBConn->disconnect();
		return;
	}
	$result->finish();
	$localDBConn->disconnect();
}

###############################################################################
sub getParentBackupFile() {
	if (@_ != 3) {
		print STDERR "DBG WARNING! &getParentBackupFile accepts 3 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($childType, $instance, $deviceDir) = (shift, shift, shift);
	my $parentType = 0;
	if ($DBG) {
		print STDERR "Child type: $childType";
		print STDERR "Child instance : $instance";
		print STDERR "Child device: $deviceDir";
	}
	if ($childType == 2) {
		$parentType = 1;
	} elsif ($childType == 6) {
		$parentType = 5;
	} elsif ($childType == 10) {
		$parentType = 9;
	}  elsif ($childType == 13) {
		$parentType = 12;
	} elsif ($childType == 1032) {
		$parentType = 1031;
	} else {
		print STDERR "Error finding parent backup type";
		return '';
	}
	my $localSystem = 1;
	my $parentId = 0;
	$parentId = getLatestBackupNoFromSource($parentType, $instance, $localSystem);
	if ($parentId == 0) {
		return '';
	} else {
		return "$deviceDir/backup_$parentId";
	}
}

##############################################################################
sub runUpdatedb() {
	if (@_ != 3) {
		print STDERR "DBG WARNING! &runUpdatedb accepts 3 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($id, $srcFile, $size) = (shift, shift, shift);
	my $msg = '';
	my $localDBConn = DBI->connect("dbi:Pg:dbname=$sqlDB", $sqlUser, '');
	if (!defined $localDBConn) {
		print STDERR "Failed to connect to database.";
		return 1;
	}
	
	my $result = $localDBConn->prepare('SELECT dev_nr_name FROM bp.devices WHERE tape_no=(SELECT tape_no FROM bp.backups WHERE backup_no= ?)');
	if (!defined $result) {
		$msg = "Failed to get device path for backup $id. Updatedb failed";
		updateResults(0, $srcFile, $size, $msg);
		$result->finish();
		$localDBConn->disconnect();
		return 1;
	}
	my $resultSql = $result->execute($id);
	if (!defined $resultSql) {
		$msg = "Failed to get device path for backup $id. Updatedb failed";
		updateResults(0, $srcFile, $size, $msg);
		$result->finish();
		$localDBConn->disconnect();
		return 1;
	}
	my $deviceDir = $result->fetchrow_array();
	$result->finish();
	$localDBConn->disconnect();
	
	my $bkupFile = "$deviceDir/backup_$id";
	my $isEncrypted = is_encrypted($bkupFile);
	my $isCompressed = `file $bkupFile | grep -q "gzip compressed data"`;
	my $exit_value = $? >> 8;
	my $cmd = '';
	if ($exit_value eq '0') {
		$cmd = "gunzip -q -c $bkupFile 2>/dev/null |";
	} else {
		$cmd = "cat $bkupFile |";
	}

	if ($isEncrypted eq 'y') {
		# decrypt and then pipe to updatedb
		$cmd .= "$bpdir/bin/bpcrypt -t -d | $bpdir/bin/updatedb -b $id -p - 1>>$logFile 2>>$logFile";
	} else {
		$cmd .= "$bpdir/bin/updatedb -b $id -p - 1>>$logFile 2>>$logFile";

	}
	if ($DBG) {
		print STDERR "Command to run: $cmd";
	}
	$result = system($cmd);
	if ($DBG) {
		print STDERR "Updatedb command returned: $result";
	}
	if ($result ne 0) {
		$msg = "Updatedb failed on backup #$id";	
		updateResults(0, $srcFile, $size, $msg);
		return 1;
	}
	return 0;
}

###############################################################################
# Checks if a given backup is on the source
# Param: backup_no
# Return: 1 if available, 0 otherwise
sub isBackupOnSource() {
	if (@_ != 1) {
		print STDERR "DBG WARNING! &isBackupOnSource accepts 1 arguments!\n";
		print STDERR "DBG @_";
	}
	my $id = shift;
	my $returnVal = 1;
	my $dbConn = DBI->connect("dbi:Pg:dbname=$sqlDB;host=$dpuName", $sqlUser, '');
	if (!defined $dbConn) {
		print STDERR  "Failed to connect to $dpuName. Is the source active?";
		$returnVal = 0;
	}

	my $result = $dbConn->prepare("SELECT TRUE FROM bp.backups
									WHERE backup_no=$id AND tape_no != 4294967295");
	my $resultSql = $result->execute();
	if (!defined $resultSql || $result->rows == 0) {
		$returnVal = 0;	
	}
	$result->finish();
	$dbConn->disconnect();
	return $returnVal;
}

###############################################################################
# Queries the source to get the backup number. If exact match not found
# gets the backup_no of the last successful backup for given type, instance
# Params: backup_type, instance_id and start_time
# Returns: 0 if nothing found, or valid backup_no on success.
sub getBackupNoFromSource() {
	if (@_ != 3) {
		print STDERR "DBG WARNING! &getBackupNoFromSource accepts 3 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($type, $instanceId, $time) = (shift, shift, shift);
	my $id = 0;
	
	# Open a DB connection
	my $dbConn = DBI->connect("dbi:Pg:dbname=$sqlDB;host=$dpuName", $sqlUser, '');
	if (!defined $dbConn) {
		print STDERR "Failed to connect to $dpuName. Is the source active?";
		return $id;
	}

	# Find exact match 
	my $result = $dbConn->prepare("SELECT backup_no FROM bp.backups WHERE type = $type 
							AND instance_id = $instanceId AND start_time = 
							$time AND tape_no != 4294967295 AND 
							(properties & $P_SYNTHETIC ) = 0");
	
	my $resultSql = $result->execute();
	if (!defined $resultSql || $result->rows() != 1) {
		$result->finish();
		$dbConn->disconnect();
		return $id;
	}
	# Have actual backup
	$id = $result->fetchrow_array(); 
	$result->finish();
	$dbConn->disconnect();
	return $id;
}

###############################################################################
# Get last successful backup for the type,instance. 
# Params: backup_type, instance_id and start_time
# Returns: 0 if nothing found, or valid backup_no on success.
sub getLatestBackupNoFromSource() {
	if (@_ != 3) {
		print STDERR "DBG WARNING! &getLatestBackupNoFromSource accepts 3 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($type, $instanceId, $local) = (shift, shift, shift);
	my $id = 0;
	
	# Open a DB connection
	my $dbConn = '';

	if ($local == 0) {
		$dbConn = DBI->connect("dbi:Pg:dbname=$sqlDB;host=$dpuName", $sqlUser, '');
	} else {
		if ($DBG) {
			print STDERR "Connecting to local system";
		}
		$dbConn = DBI->connect("dbi:Pg:dbname=$sqlDB", $sqlUser, '');
	}
	if (!defined $dbConn) {
		print STDERR "Failed to connect to database on $dpuName. Is the source active?";
		return $id;
	}
	
	# Exclude P_SYNTHETIC, P_IMPORTED
	my $result = $dbConn->prepare("SELECT max(backup_no) FROM bp.backups WHERE
			instance_id = ? AND type = ? AND
			tape_no != 4294967295 AND 
			(properties & ?) = 0 AND 
			(properties & ?) = 0 AND
			(status & ?) = ? ");
	$result->bind_param(1, $instanceId);
	$result->bind_param(2, $type);
	$result->bind_param(3, $P_SYNTHETIC);
	$result->bind_param(4, $P_IMPORTED);
	$result->bind_param(5, $FINAL_SUCCESS);
	$result->bind_param(6, $FINAL_SUCCESS);
	my $resultSql = $result->execute();
	if (!defined $resultSql) {
		# We didn't find last backup for this (type, instance)
		print STDERR "Failed to find latest backup for instance $instanceId";
		$result->finish();
		$dbConn->disconnect();
		return $id;
	}
	if ($result->rows() != 1) {
		print STDERR "Failed to find latest backup for instance $instanceId";
		$result->finish();
		$dbConn->disconnect();
		return $id;
	}
	$id = $result->fetchrow_array();
	$result->finish();
	$dbConn->disconnect();
	return $id;
}

###############################################################################
### The alternative is to use bptool -e backup_file
# Looks at the file to check if it is encrypted
# Param: Filepath
# Return: y if encrypted, n otherwise.
sub is_encrypted()
{
	if (@_ != 1) {
		print STDERR "DBG WARNING! &is_encrypted should get exactly one argument!\n";
		print STDERR "DBG @_";
	}

	my $filename = shift;
	if ($DBG eq '1') {
		print STDERR "DBG checking encryption for $filename"; 
	}
	my $linkflag = `zcat -qf $filename 2>/dev/null | dd bs=1 skip=156 count=1 2>/dev/null`;
	if ($DBG eq '1') {
		print STDERR "DBG Encryption for $filename is:$linkflag:"; 
	}
	return ($linkflag eq 'E' ? 'y' : 'n');
}

###############################################################################
# Get Instance_id from dump. First find the node_no, then uses the
# node_no and app_id to find the matching instance_id
# Params: node name, app_id
# Return: instance_id if found, -1 otherwise.
sub getInstInfoFromDump() {
	if (@_ != 2) {
		print STDERR "DBG WARNING! &getInstInfoFromDump should get exactly 2 arguments!\n";
		print STDERR "DBG @_";
	}

	my ($nodeName, $appId) = (shift, shift);
	if (! open(FILE, "$dbdump")) {
		print STDERR "Cannot open state db dump $dbdump"; return '';
	}

	# Read through the dump until we get to the application_instances table
	my @ldlines;
	my $nodeFound = 0;
	my $appsFound = 0;
	my $infoFound = 0;;
	my $nodeNo = 0;
	my $instanceId = 0;

	while (<FILE>) {
		chomp;
		if (!$nodeFound && m/^COPY nodes /) {
			$nodeFound = 1; next;
		}
		next if !$nodeFound;
		if (!$infoFound && m/^[0-9]+\s+$nodeName/) {
			my @fields = split '\t';
			$nodeNo = $fields[0];
			# Get instance_id from application_instances, file is not closed yet.
			$infoFound = 1;
		}

		if (!$infoFound && (m/^\\\./ or m/^\s*$/)) {
			print STDERR "Node $nodeName not found in dump";
			close FILE;
			return -1;
		}

		#print "Line: $_\n";
		if ($infoFound && !$appsFound && m/^COPY application_instances/) {
			$appsFound = 1; next;
		}
		next if !$appsFound;
		if ($infoFound && m/^[0-9]+\s+$nodeNo/) {
			my @fields1 = split '\t';
			if ($fields1[2] eq $appId) {
				$instanceId = $fields1[0];
				close FILE;
				return $instanceId;
			} else {
				# Look at the next line..some where a line will exist which has nodeNo and app_id that match 
				next;
			}
		}
	}
	print STDERR "Instance Id not found for node $nodeName, appId $appId in dump (EOF)";
	close FILE;
	return -1;
}

###############################################################################
# convert from a string to a numeric value representing backup type
# Param: type string
# Return: Numeric value of backup type, -1 for invalid type
sub getBackupTypeNumeric() {
	if (@_ != 1) {
		print STDERR "DBG WARNING! &getBackupTypeNumeric should get exactly one argument!\n";
		print STDERR "DBG @_";
	}

	my $typestr = shift;
	if ($typestr eq 'Master') {
		return 1;
	}
	if ($typestr eq 'Incremental') {
		return 2;
	} 
	if ($typestr eq 'SQL') {
		return 4; # This is the type for selective backups (SQL).
	}
	if ($typestr eq 'ExchFull') {
		return 5;
	}
	if ($typestr eq 'ExchDiff') {
		return 6;
	}
	if ($typestr eq 'ExchIncr') {
		return 7;
	}
	if ($typestr eq 'MSSQLFull') {
		return 9;
	}
	if ($typestr eq 'MSSQLDiff') {
		return 10;
	}
	if ($typestr eq 'MSSQLTrans') {
		return 11;
	}
	if ($typestr eq 'VMESXFull') {
		return 12;
	}
	if ($typestr eq 'VMESXDiff') {
		return 13;
	}
	if ($typestr eq 'VMESXIncr') {
		return 14;
	}
	if ($typestr eq 'BareMetal') {
		return 1024;
	}
	if ($typestr eq 'HyperVFull') {
		return 1031;
	}
	if ($typestr eq 'HyperVDiff') {
		return 1032;
	}
	if ($typestr eq 'HyperVIncr') {
		return 1033;
	}
	print STDERR "DBG WARNING! &getBackupTypeNumeric did not find a match for $typestr \n";
	return 0;
}

###############################################################################
# Gets application id from the dbdump.sql
# Params: application name (e.g. file-level)
# Return: app_id if found, -1 otherwise;
sub getAppIdFromDump()
{
	if (@_ != 1) {
		print STDERR "DBG WARNING! &getAppIdFromDump should get exactly one argument!\n";
		print STDERR "DBG @_";
	}

	my $appName = shift;
	if (! open(FILE, "$dbdump")) {
		print STDERR "Cannot open state db dump $dbdump"; return '';
	}

	# Read through the dump until we get to the application_lookup table
	my $section = '';
	my @ldlines;
	my $found = 0;
	while (<FILE>) {
		chomp;
		if (!$found && m/^COPY application_lookup /) {
			$found = 1; next;
		}
		next if !$found;
		if (m/^[0-9]+\s+$appName\s+/) {
			# This is the one, return the 1st field (0 based)
			my @fields = split '\t';
			if ($DBG eq 1) {
				print STDERR @fields;
			}
			close FILE;
			#print STDERR "DBG Found application type record for app_id $app_id, $fields[2]";
			return $fields[0];
		}
		if (m/^\\\./ or m/^\s*$/) {
			print STDERR "Application $appName not found in dump";
			close FILE;
			return -1;
		}
	}
	print STDERR "Application $appName not found in dump (EOF)";
	close FILE;
	return -1;
}

###############################################################################
# Queries the source to check if the client is present
# Params: client name, system name
# Returns 1 if client is found, 0 otherwise.
sub isClientPresentOnSource() {
	if (@_ != 1) {
		print STDERR "DBG WARNING! &isClientPresentOnSource accepts 1 arguments!\n";
		print STDERR "DBG @_";
	}
	my $cName = shift;
	my $returnVal = 1; #Start by assuming we have it
	
	# Open a DB connection
	my $dbConn = DBI->connect("dbi:Pg:dbname=$sqlDB;host=$dpuName", $sqlUser, '');
	if (!defined $dbConn) {
		print STDERR "Failed to connect to $dpuName. Is the source active?";
		return 0;
	}

	my $result = $dbConn->prepare('SELECT node_no FROM bp.nodes WHERE CASE WHEN gcname IS NULL THEN node_name= ? ELSE gcname= ? END');
	if (!defined $result) {
		print STDERR "Failed to get node_no for $cName from $syncUser";
		$result->finish();
		$dbConn->disconnect();
		return 0;
	}
	my $sqlResult = $result->execute($cName, $cName);
	if (!defined $sqlResult) {
		print STDERR "Query to find node_no for $cName from $syncUser failed";
		$result->finish();
		$dbConn->disconnect();
		return 0;
	}
	if ($result->rows() == 0) {
		print STDERR "client $cName does not exist on $syncUser";
		$result->finish();
		$dbConn->disconnect();
		return 0;
	}
	$result->finish();
	$dbConn->disconnect();
	return $returnVal;
}

###############################################################################
# Checks if a given backup file having a given size is complete or not
# Params srcFile, size of srcfile
# Return 1 if complete. 0 if not
sub alreadyCompleted() {
	if (@_ != 2) {
		print STDERR "DBG WARNING! &alreadyCompleted accepts 2 arguments!\n";
		print STDERR "DBG @_";
	}
	my ($bkupfile, $bkupsize) = (shift, shift);

	#post results, decrement backup count and update reserverd space value
	my $cmd = "";

	$cmd = "$bputil -g \"Backups Success\" $bkupfile 0 $reportFile";
	if ($DBG) { 
		print STDERR "Already completed check: cmd = $cmd";
	}
	my $size = `$cmd`;
	$size =~ s/^\s+//;
	$size =~ s/\s+$//;
	if ($DBG) {
		print STDERR "Completed size: $size";
		print STDERR "Backup size: $bkupsize";
	}
	if ($size eq $bkupsize) {
		return 1;
	}
	return 0;
}


###############################################################################
sub updateResults()
{
	if (@_ != 4) {
		print STDERR "DBG WARNING! &updateResults should get exactly 4 arguments!\n";
		print STDERR "DBG @_";
	}

	my ($success, $srcFile, $backupSize, $msg) = (shift, shift, shift, shift);

	## post results, decrement remaining backup count, and update reserved space value

	## First the results
	my $cmd = "";
	if ($success eq '1') {
		## NOTE: posting the backup size for successes, is needed when checking to see if a backup was already run for the target backup...
		$cmd = "/usr/bp/bin/bputil -p \"Backups Success\" $srcFile $backupSize $reportFile";
	}
	else {
		$cmd = "/usr/bp/bin/bputil -p \"Backups Failed\" $srcFile '$msg' $reportFile";
	}
	if ($DBG) { 
		print STDERR "cmd post: $cmd";
	}
	`$cmd`;
	## Now the backup count, only if successful migration
	## Retrieve current count, so we can decrement it
	if ($success eq '1') {
		$cmd = "/usr/bp/bin/bputil -g \"Backups\" \"Total_Remaining\" 1 $reportFile";
		if ($DBG) { 
			print STDERR "get cmd: $cmd\n";
		}
		my $current_count = `$cmd`;
		$current_count =~ s/^\s+//;
		$current_count =~ s/\s+$//;
		if ($DBG) {
			print STDERR "Backup count: $current_count\n";
		}
		## post the count update
		$current_count--;
		$cmd = "/usr/bp/bin/bputil -p \"Backups\" \"Total_Remaining\" $current_count $reportFile";
		if ($DBG) {
			print STDERR "put cmd: $cmd\n";
		}
		`$cmd`;
	}

	## Finally, the space remaining
	## Retrieve TOTAL SPACE, so we can subtract our value
	## TOTAL SPACE stored as MB, no need to convert
	$cmd = "/usr/bp/bin/bputil -g \"Backups\" \"Total_Space_Needed\" 0 $reportFile";
	if ($DBG) {
		print STDERR "get cmd: $cmd\n";
	}
	my $tot_space_needed = `$cmd`;
	$tot_space_needed =~ s/^\s+//;
	$tot_space_needed =~ s/\s+$//;
	if ($DBG) {
		print STDERR "Backup size: $backupSize\n";
		print STDERR "Total Space Needed: $tot_space_needed\n";
	}
	if ($tot_space_needed > 0) {
		## convert to MB
		##$tot_space_needed *= 1024;
		## subtract used, and convert it to gb
		##my $now_needed = ($tot_space_needed - $backupSize) / 1024;
		my $now_needed = ($tot_space_needed - $backupSize);
		if ($now_needed < 0) { 
			$now_needed = 0;
		}
		## post the size update
		$cmd = "/usr/bp/bin/bputil -p \"Backups\" \"Total_Space_Needed\" $now_needed $reportFile";
		if ($DBG) {
			print STDERR "put cmd: $cmd\n";
		}
		`$cmd`;
	}
}
