-- Define any functions here.
-- Use "CREATE or REPLACE FUNCTION" since this script will run on both
-- a fresh install or upgrade.  If a function needs to be removed or if
-- it changes signatures, call drop_function_if_exists() to remove the
-- old function before creating a new one.

SET search_path to bp;


---------------------------------------------------------------------------------------
-- Returns TRUE if all functions specified were deleted successfully
CREATE OR REPLACE FUNCTION conditionally_drop_function(sch_name VARCHAR, func_name VARCHAR)
RETURNS BOOLEAN AS
$BODY$
	DECLARE fn_row RECORD;
	DECLARE p_string VARCHAR := '';
	DECLARE type_array INT8[];
	DECLARE type_rec RECORD;
	DECLARE i INTEGER;
	DECLARE type_oid INT8;

BEGIN
	-- Select all functions with this name and retrieve the argument types
	FOR fn_row IN
		SELECT CAST(proargtypes AS INT8[]) FROM pg_catalog.pg_proc p, pg_catalog.pg_namespace n
		WHERE p.pronamespace = n.oid
			AND lower(n.nspname) = lower(sch_name)
			AND lower(p.proname) = lower(func_name)
	LOOP
		-- Build the query string for this function by iterating over parameter type array
		-- and concatenating the typenames into the string
		type_array := fn_row.proargtypes;
		FOR i IN COALESCE(array_lower(type_array,1),0) .. COALESCE(array_upper(type_array,1),-1)
		LOOP
			type_oid := type_array[i];
			SELECT INTO type_rec typname FROM pg_catalog.pg_type WHERE oid = type_oid;
			IF (i > 0) THEN p_string := p_string || ','; END IF;
			p_string := p_string || type_rec.typname;
		END LOOP;

		-- Execute the query string
		p_string := 'DROP FUNCTION ' || sch_name || '.' || func_name || '(' || p_string || ')';
		EXECUTE p_string;
		-- RAISE NOTICE 'Executed: %', p_string;
		-- RAISE NOTICE 'DROP FUNCTION';
	END LOOP;

	RETURN TRUE;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;


---------------------------------------------------------------------------------------
-- Drop removed functions that were in previous product versions
SELECT conditionally_drop_function('bp', 'MarkNetpathsForPurge');
SELECT conditionally_drop_function('bp', 'PurgeUnreferencedNetpaths');


---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION ApproxCount(SchemaName VARCHAR, TableName VARCHAR)
RETURNS BIGINT AS
$BODY$

DECLARE
	approx_count BIGINT := 0;

BEGIN

-- This function returns the approximate count of rows in a table
SELECT INTO approx_count reltuples::numeric
	FROM pg_catalog.pg_class, pg_catalog.pg_namespace
	WHERE pg_class.relnamespace = pg_namespace.oid
	  AND pg_namespace.nspname = SchemaName
	  AND pg_class.relname = TableName;

RETURN approx_count;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
SELECT conditionally_drop_function('bp', 'ReallyPurgeFileset'); -- return value changed
CREATE OR REPLACE FUNCTION ReallyPurgeFileset(fs_no BIGINT, NetpathsPurge BOOLEAN,
	OUT unref_details INTEGER, OUT unref_netpaths INTEGER)
RETURNS SETOF RECORD
AS
$BODY$
-- Deletes the fileset and detail records for the given fileset_no.
-- If NetpathsPurge is true, deletes the netpaths records associated with the fileset.
-- Returns count of details purged and count of netpaths purged

BEGIN
-- RAISE NOTICE 'ReallyPurgeFileset: fileset_no %, NetpathsPurge is %', fs_no, NetpathsPurge;

IF (NetpathsPurge)
THEN
	-- Delete all netpaths referenced by detail entries in this fileset but not
	-- referenced by any detail entries not in this fileset
	DELETE FROM bp.netpaths
		USING bp.detail AS src
		WHERE src.node_no = netpaths.node_no
		AND src.dir_no = netpaths.dir_no
		AND src.file_no = netpaths.file_no
		AND src.lib_inst_no = netpaths.lib_inst_no
		AND src.fileset_no = fs_no
		AND (SELECT node_no FROM bp.netpaths
		     WHERE src.node_no = netpaths.node_no
		       AND src.dir_no = netpaths.dir_no
		       AND src.file_no = netpaths.file_no
		       AND src.lib_inst_no = netpaths.lib_inst_no
		       AND src.fileset_no != netpaths.fileset_no) IS NULL;
	GET DIAGNOSTICS unref_netpaths = ROW_COUNT;
ELSE
	unref_netpaths := 0;
END IF;

DELETE FROM bp.detail WHERE fileset_no = fs_no;
GET DIAGNOSTICS unref_details = ROW_COUNT;
DELETE FROM bp.filesets WHERE fileset_no = fs_no;

-- RAISE NOTICE 'Netpaths processed for fileset_no %: Total=%, Deleted=%, Unreferenced=%',
--			fs_no, total_count, delete_count, unref_count;

RETURN NEXT;
RETURN;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION PurgeEmptyBackups(PurgeDays INTEGER, NetpathsPurge BOOLEAN)
RETURNS INTEGER AS
$BODY$

DECLARE
	myWindow BIGINT;
	b bp.backups%ROWTYPE;
	f RECORD;
	COMMENT_REC CONSTANT INTEGER := 65525; 
	CNTL_REC CONSTANT INTEGER := 65527; 
	total_unref INTEGER := 0;
	bkup_count INTEGER := 0;
	P_SQL CONSTANT INTEGER := 128;
	fs_fn_result RECORD;

BEGIN

SELECT INTO myWindow EXTRACT(EPOCH FROM now())::BIGINT;
myWindow := myWindow - (PurgeDays * 3600 * 24);
-- RAISE NOTICE 'window is epoch % and timestamp %', myWindow, to_timestamp(myWindow);

-- Bug 8695: do not delete records where P_HASHED | P_DEDUPLICATED is still set
FOR b IN SELECT * FROM bp.backups WHERE total_netpaths = 0
			AND type IN (1,2,4,8,1024,4096)
			AND ((properties & (512 | 1024)) = 0)
			AND ((status & 512) = 0)
			AND start_time < myWindow
LOOP
	bkup_count := bkup_count + 1;

	-- Remove unreferenced netpaths entries if NetpathsPurge is true
	IF (NetpathsPurge)
	THEN
		FOR f IN SELECT fileset_no FROM bp.filesets WHERE backup_no = b.backup_no
		LOOP
			SELECT INTO fs_fn_result * FROM bp.ReallyPurgeFileset(f.fileset_no, NetpathsPurge);	
			total_unref := total_unref + fs_fn_result.unref_netpaths;
		END LOOP; 
	END IF;

	-- If it's a sql backup, delete the corresponding records in sql_backups
	IF ((b.properties & P_SQL) = P_SQL)
	THEN
		DELETE FROM bp.sql_backups WHERE bp.sql_backups.sql_ref = b.backup_no;
	END IF;
END LOOP;

IF(bkup_count > 0)
THEN
-- Bug 8695: do not delete records where P_HASHED | P_DEDUPLICATED is still set
	DELETE FROM bp.backup_msg WHERE backup_no IN
		(SELECT backup_no FROM bp.backups WHERE total_netpaths = 0
			AND type IN (1,2,4,8,1024,4096)
			AND ((properties & (512 | 1024)) = 0)
			AND ((status & 512) = 0)
			AND start_time < myWindow);

	-- The verify records
	DELETE FROM bp.backup_msg WHERE backup_no IN
		(SELECT backup_no FROM bp.backups where data_bkno IN
			(SELECT backup_no FROM bp.backups WHERE total_netpaths = 0
				AND type IN (1,2,4,8,1024,4096)
				AND ((properties & (512 | 1024)) = 0)
				AND ((status & 512) = 0)
				AND start_time < myWindow));

	DELETE FROM bp.lcs_hist WHERE backup_no IN
		(SELECT backup_no FROM bp.backups WHERE total_netpaths = 0
			AND type IN (1,2,4,8,1024,4096)
			AND ((properties & (512 | 1024)) = 0)
			AND ((status & 512) = 0)
			AND start_time < myWindow)
		AND node_no != COMMENT_REC AND node_no != CNTL_REC;

	-- The verify records
	DELETE FROM bp.lcs_hist WHERE backup_no IN
		(SELECT backup_no FROM bp.backups where data_bkno IN
			(SELECT backup_no FROM bp.backups WHERE total_netpaths = 0
				AND type IN (1,2,4,8,1024,4096)
				AND ((properties & (512 | 1024)) = 0)
				AND ((status & 512) = 0)
				AND start_time < myWindow))
		AND node_no != COMMENT_REC AND node_no != CNTL_REC;

	-- The verify records
	DELETE FROM bp.backups WHERE backup_no IN 
		(SELECT backup_no FROM bp.backups where data_bkno IN
			(SELECT backup_no FROM bp.backups WHERE total_netpaths = 0
				AND type IN (1,2,4,8,1024,4096)
				AND ((properties & (512 | 1024)) = 0)
				AND ((status & 512) = 0)
				AND start_time < myWindow));

	DELETE FROM bp.backups WHERE backup_no IN 
		(SELECT backup_no FROM bp.backups WHERE total_netpaths = 0
			AND type IN (1,2,4,8,1024,4096)
			AND ((properties & (512 | 1024)) = 0)
			AND ((status & 512) = 0)
			AND start_time < myWindow);
END IF;

RETURN bkup_count;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION ReallyPurgeBackups(NetpathsPurge BOOLEAN)
RETURNS INTEGER AS
$BODY$
-- This function deletes that backups that have been marked as logically deleted,
-- along with their associated filesets, backup_msg, lcs_hist and sql_backups
-- records.

DECLARE
	bkup RECORD;
	f RECORD;
	COMMENT_REC CONSTANT INTEGER := 65525; 
	CNTL_REC CONSTANT INTEGER := 65527; 
	total_unref INTEGER := 0;
	bkup_count INTEGER := 0;
	P_SQL CONSTANT INTEGER := 128;
	fs_fn_result RECORD;

BEGIN
-- RAISE NOTICE 'ReallyPurgeBackups: NetpathsPurge is %', NetpathsPurge;

-- Bug 8695: do not delete records where P_HASHED | P_DEDUPLICATED is still set
FOR bkup IN SELECT backup_no, type, properties FROM bp.backups WHERE tape_no = 4294967295
				AND ((properties & (512 | 1024)) = 0)
LOOP
	bkup_count := bkup_count + 1;

	IF (bkup.type != 128 AND bkup.type != 512)
	-- it is not a verify
	THEN
		-- Remove the fileset itself
		FOR f IN SELECT fileset_no FROM bp.filesets WHERE backup_no = bkup.backup_no
		LOOP
			SELECT INTO fs_fn_result * FROM bp.ReallyPurgeFileset(f.fileset_no, NetpathsPurge);	
			total_unref := total_unref + fs_fn_result.unref_netpaths;
		END LOOP; 

		-- If it's a sql backup, delete the corresponding records in sql_backups
		IF ((bkup.properties & P_SQL) = P_SQL)
		THEN
			DELETE FROM bp.sql_backups WHERE bp.sql_backups.sql_ref = bkup.backup_no;
		END IF;
	END IF;
END LOOP;

IF (bkup_count > 0)
THEN
-- Bug 8695: do not delete records where P_HASHED | P_DEDUPLICATED is still set
	DELETE FROM bp.backup_msg WHERE backup_no IN
		(SELECT backup_no FROM bp.backups WHERE tape_no = 4294967295 AND ((properties & (512 | 1024)) = 0));

	DELETE FROM bp.lcs_hist WHERE backup_no IN
		(SELECT backup_no FROM bp.backups WHERE tape_no = 4294967295 AND ((properties & (512 | 1024)) = 0))
		AND node_no != COMMENT_REC AND node_no != CNTL_REC;

	DELETE FROM bp.backups WHERE backup_no IN 
		(SELECT backup_no FROM bp.backups WHERE tape_no = 4294967295 AND ((properties & (512 | 1024)) = 0));
END IF;

-- RAISE NOTICE 'Purged % backups, total unreferenced netpaths = %', bkup_count, total_unref;

RETURN bkup_count;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
SELECT conditionally_drop_function('bp', 'UnreferencedFilesets'); -- return value changed
CREATE OR REPLACE FUNCTION UnreferencedFilesets(DoPurge BOOLEAN, NetpathsPurge BOOLEAN,
	OUT unref_filesets INTEGER, OUT unref_details INTEGER, OUT unref_netpaths INTEGER)
RETURNS SETOF RECORD
AS
$BODY$
-- This function will either report the number of unreferenced filesets (DoPurge=false), or
-- actually remove them (DoPurge=true).  Note that NetpathsPurge is a setting in master.ini that
-- indicates whether or not the associated netpaths record should be deleted(T) or updated(F)
-- if the removal of this fileset and its corresponding detail records have rendered some
-- netpaths records as unreferenced by any fileset.

-- Returns one record with 3 columns: (unref_filesets, unref_details, unref_netpaths)

DECLARE
	fs RECORD;
	fs_fn_result RECORD;

BEGIN
-- RAISE NOTICE 'UnreferencedFilesets: NetpathsPurge is %, DoPurge is %', NetpathsPurge, DoPurge;
unref_filesets := 0;
unref_details := 0;
unref_netpaths := 0;

FOR fs IN SELECT fileset_no FROM bp.filesets
			WHERE backup_no NOT IN (SELECT backup_no FROM bp.backups)
LOOP
	unref_filesets := unref_filesets + 1;
	IF (DoPurge) 
	THEN
		SELECT INTO fs_fn_result * FROM bp.ReallyPurgeFileset(fs.fileset_no, NetpathsPurge);	
		unref_details := unref_details + fs_fn_result.unref_details;
		unref_netpaths := unref_netpaths + fs_fn_result.unref_netpaths;
	END IF;
END LOOP;

-- IF (DoPurge)
-- THEN
--	RAISE NOTICE 'Purged % unreferenced filesets, resulting in % unreferenced netpaths ',
--			unref_filesets, unref_netpaths; 
-- ELSE
--	RAISE NOTICE 'Unreferenced filesets count: %', unref_filesets;
-- END IF;

RETURN NEXT;
RETURN;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION UnreferencedDetails(DoPurge BOOLEAN)
RETURNS BIGINT AS
$BODY$
-- This function will either report the number of unreferenced details (DoPurge=false), or
-- actually remove them (DoPurge=true).
DECLARE
	unref_details BIGINT := 0;

BEGIN
-- RAISE NOTICE 'UnreferencedDetails: DoPurge is %', DoPurge;

IF(DoPurge)
THEN
	DELETE FROM bp.detail WHERE fileset_no NOT IN (SELECT fileset_no FROM bp.filesets);
	GET DIAGNOSTICS unref_details = ROW_COUNT;
ELSE
	SELECT INTO unref_details count(*) FROM bp.detail WHERE fileset_no NOT IN
			(SELECT fileset_no FROM bp.filesets);
END IF;

-- RAISE NOTICE 'Unreferenced detail count: %', unref_details;
RETURN unref_details;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION SetLibraryPurgeStatus(status INTEGER, cmt_str VARCHAR)
RETURNS INTEGER AS
$BODY$

BEGIN
	UPDATE bp.library SET purge_status = status, comment = cmt_str
	WHERE library_no = 1;

RETURN status;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION PurgeUnreferencedNodeNetpaths(NodeNo INTEGER,
	StartDirNo INTEGER, LastDirNo INTEGER)
RETURNS BIGINT AS
$BODY$

DECLARE
	begin_num_netpaths BIGINT := 0;
	netpaths_purged BIGINT := 0;
	netpaths_remaining BIGINT;
	msg VARCHAR;
	dir_no_cond VARCHAR;
	query VARCHAR;

BEGIN

-- create string representing dir_no conditional in query
dir_no_cond := 'dir_no >= ' || StartDirNo;
IF (LastDirNo > 0)
THEN
	dir_no_cond := dir_no_cond || ' AND dir_no < ' || LastDirNo;
END IF;

LOCK TABLE bp.netpaths, bp.detail;

SELECT INTO begin_num_netpaths bp.ApproxCount('bp', 'netpaths');

query := 'DELETE FROM bp.netpaths'
	|| ' WHERE node_no = ' || NodeNo
	|| ' AND ' || dir_no_cond
	||   ' AND (node_no, dir_no, file_no, lib_inst_no)'
	||   ' NOT IN (SELECT node_no, dir_no, file_no, lib_inst_no'
	||           ' FROM bp.detail WHERE node_no = ' || NodeNo
	||                            ' AND ' || dir_no_cond || ')';
EXECUTE query;

GET DIAGNOSTICS netpaths_purged = ROW_COUNT;

netpaths_remaining := begin_num_netpaths - netpaths_purged;
UPDATE bp.library SET purge_status = 0, total_netpaths = netpaths_remaining,
	unref_netpaths = 0, comment = '' WHERE library_no = 1;

-- RAISE NOTICE 'Unreferenced netpaths count: %', netpaths_purged;
RETURN netpaths_purged;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION UnreferencedNetpaths(DoPurge BOOLEAN)
RETURNS BIGINT AS
$BODY$
-- This function will either report the number of unreferenced netpaths (DoPurge=false), or
-- actually remove them (DoPurge=true).
DECLARE
	unref_netpaths BIGINT := 0;

BEGIN
-- RAISE NOTICE 'UnreferencedNetpaths: DoPurge is %', DoPurge;

IF(DoPurge)
THEN
	DELETE FROM bp.netpaths WHERE node_no NOT IN (SELECT node_no FROM bp.nodes);
	GET DIAGNOSTICS unref_netpaths = ROW_COUNT;
ELSE
	SELECT INTO unref_netpaths count(*) FROM bp.netpaths WHERE node_no NOT IN
			(SELECT node_no FROM bp.nodes);
END IF;

-- RAISE NOTICE 'Unreferenced netpaths count: %', unref_netpaths;
RETURN unref_netpaths;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION PurgeDirnames()
RETURNS INTEGER AS
$BODY$

DECLARE
	d_count BIGINT := 0;
	ISLNK CONSTANT INTEGER := 49;
	SYMLNK CONSTANT INTEGER := 50;

BEGIN
	DELETE FROM bp.d_master WHERE dir_no NOT IN (SELECT DISTINCT dir_no FROM bp.netpaths
		UNION SELECT DISTINCT start_dir_no FROM bp.backups
		UNION SELECT DISTINCT link_to_dir_no FROM bp.netpaths
		        WHERE file_type = ISLNK OR file_type = SYMLNK);
	GET DIAGNOSTICS d_count = ROW_COUNT;

RETURN d_count;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION PurgeFilenames()
RETURNS INTEGER AS
$BODY$

DECLARE
	f_count BIGINT := 0;
	ISLNK CONSTANT INTEGER := 49;
	SYMLNK CONSTANT INTEGER := 50;

BEGIN
	DELETE FROM bp.f_master WHERE file_no NOT IN (SELECT DISTINCT file_no FROM bp.netpaths
		UNION SELECT DISTINCT link_to_file_no FROM bp.netpaths
		        WHERE file_type = ISLNK OR file_type = SYMLNK);
	GET DIAGNOSTICS f_count = ROW_COUNT;

RETURN f_count;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;



---------------------------------------------------------------------------------------
SELECT conditionally_drop_function('bp', 'ImportTapeCreateReport');
CREATE OR REPLACE FUNCTION ImportTapeCreateReport()
RETURNS VARCHAR AS
$BODY$

DECLARE
	node_row RECORD;
	fs_row RECORD;
	rpt_row RECORD;
	fs_query TEXT;
	node_number INTEGER;
	nbackups INTEGER := 0; 
	msg VARCHAR;
	str VARCHAR;
	backup_details VARCHAR[];
	field_lengths INTEGER[];

BEGIN

field_lengths[0] = length('tape');
field_lengths[1] = length('client');
field_lengths[2] = length('backup');

-- Process each node separately for performance
FOR node_row IN SELECT node_no FROM bp.nodes ORDER BY node_no
LOOP
	node_number := node_row.node_no;
	-- Obtain all filesets with missing netpaths
	fs_query := 'SELECT DISTINCT fileset_no FROM bp.detail'
	          || ' WHERE fileset_no IN (SELECT fileset_no FROM bp.filesets)'
	          ||   ' AND node_no = ' || node_number
	          ||   ' AND (dir_no, file_no, lib_inst_no)'
	          ||       ' NOT IN (SELECT dir_no, file_no, lib_inst_no'
	          ||               ' FROM bp.netpaths WHERE node_no = ' || node_number || ')';
	FOR fs_row IN EXECUTE fs_query
	LOOP
		-- Obtain tape_no, node_name, backup_no for all filesets
		SELECT INTO rpt_row backups.tape_no, node_name, backup_no
			FROM bp.backups
				JOIN bp.nodes USING (node_no)
				JOIN bp.filesets USING (backup_no)
			WHERE bp.nodes.node_no = node_number
			  AND bp.filesets.fileset_no = fs_row.fileset_no;

		-- Store details (in groups of 3) and max field widths for later reporting
		backup_details[(3*nbackups)+0] := rpt_row.tape_no;
		backup_details[(3*nbackups)+1] := rpt_row.node_name;
		backup_details[(3*nbackups)+2] := rpt_row.backup_no;
		field_lengths[0] := int4larger(field_lengths[0], length(rpt_row.tape_no));
		field_lengths[1] := int4larger(field_lengths[1], length(rpt_row.node_name));
		field_lengths[2] := int4larger(field_lengths[2], length(rpt_row.backup_no));
		nbackups := nbackups + 1;
	END LOOP;
END LOOP;

IF (nbackups <= 0)
THEN
	msg := '\nNo backups need to be imported.\n';
ELSE
	IF (nbackups = 1) THEN
		str := 'backup needs';
	ELSE
		str := 'backups need';
	END IF;
	msg := '\nThe following ' || nbackups || ' ' || str || ' to be imported because some'
	    || '\ndatabase information necessary to restore the data is incomplete.'
	    || '\nPlease refer to the section "Importing Backups" in the "Backups"'
	    || '\nchapter of the DPU Manual for instructions on how to import a backup.\n\n';

	-- Add formatted query output to msg string
	msg := msg || ' ' || rpad('tape', field_lengths[0])
	         || ' | ' || rpad('client', field_lengths[1])
	         || ' | ' || 'backup' || '\n'
	         || repeat('-', field_lengths[0]+1) || '+++'
	         || repeat('-', field_lengths[1])   || '+++'
	         || repeat('-', field_lengths[2]+1);

	FOR i IN COALESCE(array_lower(backup_details,1),0) .. (COALESCE(array_upper(backup_details,1),-1) / 3)
	LOOP
		-- RAISE NOTICE '% | % | %', backup_details[(3*i)+0], backup_details[(3*i)+1], backup_details[(3*i)+2];
		msg := msg || '\n ' || rpad(backup_details[(3*i)+0], field_lengths[0])
				   || ' | ' || rpad(backup_details[(3*i)+1], field_lengths[1])
				   || ' | ' || rpad(backup_details[(3*i)+2], field_lengths[2]);
	END LOOP;
END IF;
-- RAISE NOTICE '%', msg;

RETURN msg;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

---------------------------------------------------------------------------------------

CREATE OR REPLACE FUNCTION restart_sequence(sequence_name VARCHAR, query VARCHAR)
RETURNS BOOLEAN AS
$BODY$
	DECLARE result BIGINT;
	DECLARE rows INTEGER;
BEGIN
	EXECUTE query;
	-- check query before attempting to store result
	GET DIAGNOSTICS rows = ROW_COUNT;
	IF (rows != 1) THEN
		RAISE NOTICE 'Unexpected number of rows: %', rows;
		RETURN FALSE;
	END IF;

	--DROP SEQUENCE IF EXISTS sequence_name (when we upgrade postgres > 8.1)
	EXECUTE query INTO result;

	-- ensure sequence min value is not violated
	IF (result = 0) THEN
		result := 1;
	END IF;

	EXECUTE 'ALTER SEQUENCE ' || sequence_name || ' RESTART WITH ' || result;

	RETURN TRUE;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;
	
---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION bp.update_modification_time() RETURNS TRIGGER AS
$modification_tag$
	DECLARE name VARCHAR;
	name_row RECORD;
	new_record TEXT;
	old_record TEXT;
	different BOOLEAN := false;
BEGIN
	-- Updates are special. Can happen to existing records, without 
	-- change in data. E.g., bp.storage.is_online used to be updated 
	-- to the same value every xx seconds by populate_alerts.
	IF (TG_OP = 'UPDATE') THEN
		IF ((TG_RELNAME = 'nodes') OR (TG_RELNAME = 'bp.nodes')) THEN
		-- Or OLD = NEW -- 
			IF (OLD.node_no = NEW.node_no AND
			OLD.node_name = NEW.node_name AND
			OLD.type = NEW.type AND
			OLD.os = NEW.os AND
			OLD.vendor = NEW.vendor AND
			OLD.status = NEW.status AND
			OLD.last_master	= NEW.last_master AND
			OLD.last_master_start_time = NEW.last_master_start_time AND
			OLD.last_incremental = NEW.last_incremental AND
			OLD.last_incremental_start_time = NEW.last_incremental_start_time AND
			OLD.last_selective = NEW.last_selective AND
			OLD.last_selective_start_time = NEW.last_selective_start_time AND
			OLD.last_baremetal 	= NEW.last_baremetal AND
			OLD.last_baremetal_start_time = NEW.last_baremetal_start_time AND
			OLD.priority = NEW.priority AND
			OLD.options = NEW.options AND
			OLD.features = NEW.features AND
			OLD.fstatus = NEW.fstatus AND
			OLD.properties = NEW.properties AND
			OLD.bm_media_test_date = NEW.bm_media_test_date AND
			OLD.version = NEW.version) THEN
				RETURN NULL;
			-- ignore changes to some columns. 
			ELSIF (OLD.last_master	!= NEW.last_master OR
			OLD.last_master_start_time != NEW.last_master_start_time OR
			OLD.last_incremental != NEW.last_incremental OR
			OLD.last_incremental_start_time != NEW.last_incremental_start_time OR
			OLD.last_selective != NEW.last_selective OR
			OLD.last_selective_start_time != NEW.last_selective_start_time OR
			OLD.last_baremetal 	!= NEW.last_baremetal OR
			OLD.last_baremetal_start_time != NEW.last_baremetal_start_time) THEN
				RETURN NULL;
			END IF;
		ELSIF ((TG_RELNAME = 'devices') OR (TG_RELNAME = 'bp.devices')) THEN
			IF (OLD.device_no = NEW.device_no AND
			OLD.dev_name = NEW.dev_name AND
			OLD.device_type = NEW.device_type AND
			OLD.node_no = NEW.node_no AND
			OLD.capacity = NEW.capacity AND
			OLD.blocksize = NEW.blocksize AND
			OLD.cur_fileset = NEW.cur_fileset AND
			OLD.comment = NEW.comment AND
			OLD.status = NEW.status AND
			OLD.cur_k_position = NEW.cur_k_position AND
			OLD.time_stamp = NEW.time_stamp AND
			OLD.scsi_blksize = NEW.scsi_blksize AND
			OLD.is_scsi = NEW.is_scsi AND
			OLD.dev_rw_name = NEW.dev_rw_name AND
			OLD.dev_nr_name = NEW.dev_nr_name AND
			OLD.dev_cmd_name = NEW.dev_cmd_name AND
			OLD.pid = NEW.pid AND
			OLD.same_as_devices = NEW.same_as_devices AND
			OLD.auto_switch = NEW.auto_switch AND
			OLD.tape_class = NEW.tape_class AND
			OLD.reservation_pid = NEW.reservation_pid AND
			OLD.scsi_caps = NEW.scsi_caps AND
			OLD.scsi_lun = NEW.scsi_lun AND
			OLD.scsi_id = NEW.scsi_id AND
			OLD.scsi_ha = NEW.scsi_ha AND
			OLD.tested = NEW.tested AND
			OLD.drive_level = NEW.drive_level AND
			OLD.fd = NEW.fd AND
			OLD.tape_no = NEW.tape_no AND
			OLD.properties = NEW.properties AND
			OLD.juke_drive_no = NEW.juke_drive_no AND
			OLD.no_label = NEW.no_label AND
			OLD.mux_maxcnt = NEW.mux_maxcnt AND
			OLD.mux_sched = NEW.mux_sched AND
			OLD.mux_running = NEW.mux_running AND
			OLD.is_online = NEW.is_online AND
			OLD.slot_no = NEW.slot_no AND
			OLD.dev_changer_name = NEW.dev_changer_name AND
			OLD.use_juke = NEW.use_juke AND
			OLD.juke_libno = NEW.juke_libno AND
			OLD.comment_short = NEW.comment_short AND
			OLD.total_write_usages = NEW.total_write_usages AND
			OLD.total_write_time = NEW.total_write_time AND
			OLD.backup_no = NEW.backup_no AND
			OLD.juke_slot_info = NEW.juke_slot_info AND
			OLD.juke_node_no = NEW.juke_node_no AND
			OLD.is_deleted = NEW.is_deleted AND
			OLD.dedup = NEW.dedup AND
			OLD.storage_id = NEW.storage_id) THEN
				RETURN NULL;
			ELSIF (OLD.node_no != NEW.node_no OR 
				OLD.cur_fileset != NEW.cur_fileset OR
				OLD.cur_k_position != NEW.cur_k_position OR
				OLD.reservation_pid != NEW.reservation_pid OR
				OLD.juke_drive_no != NEW.juke_drive_no OR
				OLD.mux_running != NEW.mux_running OR
				OLD.mux_sched != NEW.mux_sched OR
				OLD.backup_no != NEW.backup_no OR 
				OLD.status != NEW.status OR 
				OLD.pid != NEW.pid OR
				OLD.time_stamp != NEW.time_stamp) THEN
				RETURN NULL;
			END IF;
		ELSIF ((TG_RELNAME = 'lcs') OR (TG_RELNAME = 'bp.lcs')) THEN
			IF (OLD.lcs_no = NEW.lcs_no AND
				OLD.is_online = NEW.is_online AND
				OLD.start_time = NEW.start_time AND
				OLD.sched_name = NEW.sched_name AND
				OLD.device_no = NEW.device_no AND
				OLD.status = NEW.status AND
				OLD.last_start_time = NEW.last_start_time AND
				OLD.last_end_time = NEW.last_end_time AND
				OLD.total_netpaths = NEW.total_netpaths AND
				OLD.total_megs = NEW.total_megs AND
				OLD.job_no = NEW.job_no AND
				OLD.failure_report_is_online = NEW.failure_report_is_online AND
				OLD.failure_report_notification_type = NEW.failure_report_notification_type AND
				OLD.backup_summary_is_online = NEW.backup_summary_is_online AND
				OLD.backup_summary_notification_type = NEW.backup_summary_notification_type AND
				OLD.weekly_summary_is_online = NEW.weekly_summary_is_online AND
				OLD.weekly_summary_notification_type = NEW.weekly_summary_notification_type AND
				OLD.cmd_before = NEW.cmd_before AND
				OLD.cmd_after = NEW.cmd_after AND
				OLD.server_name = NEW.server_name AND
				OLD.last_message = NEW.last_message AND
				OLD.working_element = NEW.working_element AND
				OLD.is_recurring = NEW.is_recurring) THEN
				RETURN NULL;
			ELSIF (OLD.last_start_time != NEW.last_start_time OR
					OLD.last_end_time != NEW.last_end_time OR
					OLD.total_netpaths != NEW.total_netpaths OR
					OLD.total_megs != NEW.total_megs OR
					OLD.job_no != NEW.job_no OR
					OLD.last_message != NEW.last_message OR
					OLD.status != NEW.status OR
					OLD.working_element != NEW.working_element) THEN
				RETURN NULL;
			END IF;
		ELSIF ((TG_RELNAME = 'lcs_client') OR (TG_RELNAME = 'bp.lcs_client')) THEN
			IF (OLD.status != NEW.status OR
					OLD.last_job != NEW.last_job OR
					OLD.last_job_status != NEW.last_job_status OR
					OLD.last_backup != NEW.last_backup OR
					OLD.last_verify_job != NEW.last_verify_job OR
					OLD.last_verify_status != NEW.last_verify_status OR
					OLD.last_verify_backup != NEW.last_verify_backup OR
					OLD.last_message != NEW.last_message OR
					OLD.last_queued_time != NEW.last_queued_time) THEN
				RETURN NULL;
			END IF;
		ELSIF ((TG_RELNAME = 'systems') OR (TG_RELNAME = 'bp.systems')) THEN
			IF (OLD.check_timestamp != NEW.check_timestamp OR
				OLD.is_active != NEW.is_active) THEN
				RETURN NULL;
			END IF;
		END IF;  
	END IF;		
-- Try update first, usually entry for table name will exist --
UPDATE bp.modification_time SET last_changed=extract(epoch FROM now()) WHERE table_name=TG_RELNAME;
	RETURN NULL;
IF NOT FOUND THEN
	INSERT INTO bp.modification_time(table_name) values (TG_RELNAME, extract(epoch FROM now()));
		RETURN NEW;
END IF;
END$modification_tag$
LANGUAGE 'plpgsql';

---------------------------------------------------------------------------------------
-- BOOL vacuum_threshold_bpdb(varchar SchemaName, varchar tableName, integer bloatThr)
-- Example: vacuum_threshold_bpdb('bp', 'hash_values',25)
-- Returns TRUE if table bloat exceeds bloatTrhr FALSE otherwise
-- NOTE: bloatThr must be between 0 and 100 (inclusive) since it denotes percent bloat
---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION vacuum_threshold_bpdb(varchar, varchar, integer)
RETURNS BOOLEAN AS
$BODY$

  DECLARE bloatStatus BOOLEAN;
  psqlStatement text := '';

  theSchema ALIAS FOR $1;
  theTab ALIAS FOR $2;
  theThreshold ALIAS FOR $3;
  --theThrPercent INTEGER := 50;

BEGIN

    /*SELECT CASE WHEN EXISTS (
    --  SELECT bp.domain_exists_bpdb('bp', 'bloat_pct_rangetype')
    --) THEN TRUE ELSE FALSE END INTO bloatStatus;

    --IF NOT bloatStatus THEN
    --  CREATE DOMAIN bloat_pct_rangetype AS INTEGER CHECK (VALUE >= 0 AND VALUE <= 100);
    END IF;*/

    --theThrPercent := theThreshold::bloat_pct_rangetype;

    SELECT
    CASE WHEN EXISTS (
        SELECT bloatpct FROM bloat_tables_bpdb WHERE
        schemaname = theSchema AND relname=theTab
        --AND bloatpct >= theThrPercent
        AND bloatpct >= theThreshold
    )
    THEN TRUE ELSE FALSE END INTO bloatStatus;

    /*-- vacuum if bloat exceeds threshold (boohoo can't run vacuum inside fn)
    IF bloatStatus THEN
      psqlStatement := 'VACUUM ANALYZE ' || theSchema || '.' || theTab;
      --EXECUTE psqlStatement;
      --RAISE NOTICE '%', psqlStatement;
    END IF;
    */
    RETURN bloatStatus;

END$BODY$ 
LANGUAGE 'plpgsql' VOLATILE;
---------------------------------------------------------------------------------------

/*  @desc: bpGetRestoreGroupOptimized returns a tuple of backup ids + supporting info for the all backups that need 
           to be restored in preparation for the given backup id to be restored;
    @input args: backup_id (integer) types in the order - full, diff, incr (array) and a filter (string)
    @output: getrestoregrouptuple (set of rows) that belong to the group of backups to be restored if successful 
              if no-records are found or non supported backup type, it returns a single row (0,0,0,0,0)
    @error: not implemented (and defaults to postgres exception handling)
*/

---------------------------------------------------------------------------------------
\set ON_ERROR_STOP off
  DROP TYPE getrestoregrouptuple CASCADE;
  CREATE TYPE getrestoregrouptuple AS(T_Backup_No BIGINT, T_Status BIGINT, T_Exit_Code BIGINT, T_Type INTEGER, T_Start_Time BIGINT);
\set ON_ERROR_STOP on
---------------------------------------------------------------------------------------
SELECT conditionally_drop_function('bp', 'bpGetRestoreGroupOptimized');
CREATE OR REPLACE FUNCTION bpGetRestoreGroupOptimized(integer, integer[], varchar)
RETURNS SETOF bp.getrestoregrouptuple AS
$BODY$

DECLARE

  -- Record holds the rows returned
  theRecord bp.getrestoregrouptuple%ROWTYPE;

  -- return values per row
  T_Backup_No BIGINT; T_Status BIGINT; T_Exit_Code BIGINT; T_Type INTEGER; T_Start_Time BIGINT;

  -- instance_id, latest-full, latest-full-or-diff
  theInstanceId INTEGER;
  latestFull INTEGER;
  latestFullOrDiff INTEGER;

  -- supported backup types (look at include/backtype.h)
  IsTypeSharePt BOOLEAN := FALSE;
  IsTypeOracle BOOLEAN := FALSE;
  IsTypeUCS BOOLEAN := FALSE;
  IsTypeExchange BOOLEAN := FALSE;
  IsTypeNoDeps BOOLEAN := FALSE;
  IsTypeFLDiff BOOLEAN := FALSE;
  IsTypeAppSupp BOOLEAN := FALSE;

  -- Assume result is not null 
  IsResultNull INTEGER := 1;

  -- placeholder for query
  theQuery TEXT;
  theTempTable VARCHAR;

  -- parameters passed to function
  theBackupId ALIAS FOR $1;
  theBTypes ALIAS FOR $2;
  theFilter ALIAS FOR $3;

BEGIN

    theInstanceId := 0; latestFull := 0; latestFullOrDiff := 0;
    --theRecord := (0,0); 
    -- Get instance_id for this backup
    theInstanceId := (SELECT instance_id FROM bp.backups WHERE bp.backups.backup_no = theBackupId);

    -- Get type (Sharepoint, Oracle, UCS, NoDeps, File-level Diff and Supported Apps - MSSQL, VMESX, HyperV, NDMP)
    -- Note: the order of theBTypes matters with theBTypes[1] indicating FULL backup: ordering of arguments 
    -- are tightly coupled to bpGetRestoreGroupOptimized() function in bp/libsrc/bpsql.c
    IsTypeSharePt := (SELECT theBTypes[1] = 1036);
    IsTypeOracle := (SELECT theBTypes[1] = 1034);
    IsTypeUCS := (SELECT theBTypes[1] = 1039);
    IsTypeExchange := (SELECT theBTypes[1] = 5);
    IsTypeNoDeps := (SELECT theBTypes[1] IN (1,4,1024,1038));
    IsTypeFLDiff := (SELECT theBTypes[1] IN (2));
    IsTypeAppSupp := (SELECT theBTypes[1] IN (9,10,11,12,13,14,1031,1032,1033,4098,4099,4100));

    -- temp table name
    theTempTable := (SELECT 'bpGRG' || md5(random()::VARCHAR));

    -- Get the group corresponding to this instance_id, backup_no and any type with filter applied
    theQuery := 'CREATE TEMPORARY TABLE ' || theTempTable || ' AS 
    (SELECT backup_no, exit_code, properties, start_time, status, instance_id, type, tape_no FROM bp.backups
        WHERE instance_id = '|| theInstanceId || ' AND backup_no <= ' || theBackupId || ' AND '|| theFilter || ');';

    EXECUTE theQuery;

    theQuery := 'SELECT count(1) FROM '|| theTempTable;

    EXECUTE theQuery INTO IsResultNull;

    IF (IsResultNull = 0) THEN
      FOR theRecord IN (SELECT 0,0,0,0,0)
      LOOP
        RETURN NEXT theRecord;
      END LOOP;
    ELSE

    -- Get latestFull
    theQuery := 'SELECT rg.backup_no FROM ' || theTempTable || ' rg WHERE type = ' || theBTypes[1] || ' AND
      rg.backup_no <= ' || theBackupId || ' ORDER BY backup_no DESC LIMIT 1;';

    EXECUTE theQuery INTO latestFull;

    -- (type == BT_SELE || type == BT_MAST || type == BT_IMAGE_BACKUP || type == BT_APPLN_META) OR (type == BT_RAE_(FULL,DIFF,INCR))
    IF (IsTypeNoDeps OR IsTypeUCS) THEN
      FOR theRecord IN EXECUTE 'SELECT rg.backup_no, rg.status, rg.exit_code, rg.type, rg.start_time FROM ' || theTempTable || '  rg
      WHERE rg.backup_no = ' || theBackupId || ' ORDER BY backup_no ASC;'
        LOOP
          RETURN NEXT theRecord;
        END LOOP;

    -- (type == BT_DIFF)
    ELSIF (IsTypeFLDiff) THEN
      FOR theRecord IN EXECUTE 'SELECT rg.backup_no, rg.status, rg.exit_code, rg.type, rg.start_time FROM ' || theTempTable || '  rg
      WHERE backup_no IN ( (SELECT ' || theBackupId || '  AS backup_no) UNION ALL (SELECT backup_no FROM ' || theTempTable || '  sg
      WHERE sg.type = 1 AND backup_no <= ' || theBackupId || '  ORDER BY backup_no DESC LIMIT 1)) ORDER BY backup_no ASC;'
        LOOP
          RETURN NEXT theRecord;
      END LOOP;

    -- (type == BT_ORACLE_(FULL,INCR)
    ELSIF (IsTypeOracle) THEN
      FOR theRecord IN EXECUTE 'SELECT rg.backup_no, rg.status, rg.exit_code, rg.type, rg.start_time FROM ' || theTempTable || '  rg
      WHERE rg.backup_no >= ' || latestFull || ' AND backup_no <= ' || theBackupId || '  AND (type IN (' || theBTypes[1] || ', ' || theBTypes[2] || ')
      OR backup_no = ' || theBackupId || ' ) ORDER BY backup_no ASC;'
        LOOP
          RETURN NEXT theRecord;
      END LOOP;

    -- (type == BT_EXCHANGE_(FULL=5,INCR=7)
    ELSIF (IsTypeExchange) THEN
      FOR theRecord IN EXECUTE 'SELECT rg.backup_no, rg.status, rg.exit_code, rg.type, rg.start_time FROM ' || theTempTable || ' rg
      WHERE rg.backup_no >= ' || latestFull || ' AND backup_no <= ' || theBackupId || '  AND (type IN (' || theBTypes[1] || ', ' || theBTypes[3] || ')
      OR backup_no = ' || theBackupId || ' ) ORDER BY backup_no ASC;'
        LOOP
          RETURN NEXT theRecord;
      END LOOP;

    -- (type = BT_SHAREPT_(FULL,DIFF)
    ELSIF (IsTypeSharePt) THEN
      FOR theRecord IN EXECUTE '(SELECT rg.backup_no, rg.status, rg.exit_code, rg.type, rg.start_time FROM ' || theTempTable || ' rg
      WHERE backup_no = ' || latestFull || ') UNION (SELECT sg.backup_no, sg.status, sg.exit_code, sg.type, sg.start_time FROM ' || theTempTable || ' sg
      WHERE backup_no = ' || theBackupId || ' AND type = ' || theBTypes[2] || ') ORDER BY backup_no ASC;'
        LOOP
          RETURN NEXT theRecord;
      END LOOP;

    ELSIF (IsTypeAppSupp) THEN
      -- Get latestFullOrDiff after latestFull
      theQuery := 'SELECT rg.backup_no FROM ' || theTempTable || ' rg WHERE type IN (' || theBTypes[1] || ', ' || theBTypes[2] || ')
         AND rg.backup_no <= ' || theBackupId || ' ORDER BY rg.backup_no DESC LIMIT 1;';
      EXECUTE theQuery INTO latestFullOrDiff;
      -- RAISE NOTICE '% % % % % ', latestFull, latestFullOrDiff, theBTypes[1], theBTypes[2], theBTypes[3];

      --FOR theRecord IN SELECT rg.backup_no, rg.status, rg.exit_code, rg.type, rg.start_time
      FOR theRecord IN EXECUTE 'SELECT rg.backup_no, rg.status, rg.exit_code, rg.type, rg.start_time FROM ' || theTempTable || ' rg
      WHERE ( (rg.type = ' || theBTypes[1] || ' AND rg.backup_no >= ' || latestFull || ') OR
      (rg.type = ' || theBTypes[2] || ' AND rg.backup_no >= ' || latestFullOrDiff || ') OR
      (rg.type = ' || theBTypes[3] || ' AND rg.backup_no <= ' || theBackupId || ' AND rg.backup_no >= ' || latestFullOrDiff || ')) ORDER BY backup_no ASC;'
      LOOP
        RETURN NEXT theRecord;
      END LOOP;
    ELSE
      FOR theRecord IN (SELECT 0,0,0,0,0)
      LOOP
        RETURN NEXT theRecord;
      END LOOP;
    END IF;
    END IF;

    theQuery := 'DROP TABLE ' || theTempTable;
    EXECUTE theQuery;
    --DROP TABLE bpGetRestoreGroupByInstanceID;

    RETURN;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;
---------------------------------------------------------------------------------------

---------------------------------------------------------------------------------------
/*  @desc: bpGetBackupFootprintPerSystem returns a summary footprint for
           backups not logically deleted for a given system 
           
    @input args: 
		bp.systems.system_id (integer, required)
		filesystem block size (integer, required) 
		distinct hash_id required (boolean, required) 

    @output: footprint (bigint) sum of hashed_length for the system
 
    @error: defaults to postgres exception handling

*/
---------------------------------------------------------------------------------------
SELECT conditionally_drop_function('bp', 'bpGetBackupFootprintPerSystem');
CREATE OR REPLACE FUNCTION bpGetBackupFootprintPerSystem(INTEGER, INTEGER, BOOLEAN)
RETURNS BIGINT AS

$BODY$

DECLARE

	theSystemId ALIAS FOR $1;
	theDiskBlockSize ALIAS FOR $2;
	theDistinctHashId ALIAS FOR $3;

	-- sum total over all nodes in this system
	theSystemSumHashedLength REAL := 0;

	-- checks if backups/nodes exist
	theNodesExist BOOLEAN := FALSE;

	-- check if compressed_length column exists
	hasCompressedLengthColumn BOOLEAN := FALSE;

	-- placeholder for query
	theQuery TEXT;
	theStoredLengthQuery TEXT;
   
BEGIN
	hasCompressedLengthColumn := (
        	(SELECT COUNT(*) FROM pg_catalog.pg_class c, pg_catalog.pg_namespace n, pg_catalog.pg_attribute a
        	WHERE a.attnum > 0
                	AND NOT a.attisdropped
                	AND a.attrelid = c.oid
                	AND c.relnamespace = n.oid
                	AND lower(n.nspname) = lower('bp')
                	AND lower(c.relname) = lower('hash_values')
                	AND lower(a.attname) = lower('compressed_length')
        	) > 0);

	theNodesExist := (SELECT EXISTS (SELECT node_no from bp.nodes where system_id = theSystemId LIMIT 1));

	IF (theNodesExist) THEN

        -- get backups for this system with good status code and not logically deleted 
        -- Status bits for 36928: ST_FAILED ST_CANCELLED ST_JOB_FINAL 
        EXECUTE 'CREATE TEMPORARY TABLE bpBackupNumPerSystemId AS SELECT b.backup_no FROM bp.backups b 
        WHERE b.tape_no != 4294967295 AND ((b.status & 36888) = 0) AND b.node_no IN 
        (SELECT n.node_no FROM bp.nodes n LEFT JOIN bp.systems s ON n.system_id = s.system_id 
          WHERE s.system_id = ' || theSystemId || ');';

        -- if distinct hash_id required - apply filter
        theQuery := 'SELECT SUM(uniqueHashedLength) FROM (SELECT ';
        IF (theDistinctHashId) THEN
            theQuery := theQuery || 'DISTINCT';
        END IF;

	IF (hasCompressedLengthColumn) THEN
		theStoredLengthQuery := '(CASE WHEN ((compressed_length IS NOT NULL) AND (compressed_length > 0)) THEN compressed_length ELSE hashed_length END)';
	ELSE
		theStoredLengthQuery := 'hashed_length';
	END IF;

        -- get footprint per system 
        --theQuery := theQuery || ' hash_id, count(*), SUM((hashed_length+' ||
        theQuery := theQuery || ' hash_id, count(*), SUM((' || theStoredLengthQuery || '+' ||
        theDiskBlockSize || '-1)/'||theDiskBlockSize||'*'||theDiskBlockSize
        ||')/count(*) AS uniqueHashedLength FROM bp.backups_hash_values bhv JOIN bpBackupNumPerSystemId bsys ON 
        bhv.backup_no = bsys.backup_no NATURAL JOIN bp.hash_values GROUP BY hash_id) AS theResult;';

        EXECUTE theQuery INTO theSystemSumHashedLength;

        IF (theSystemSumHashedLength > 0) THEN
            theSystemSumHashedLength = CEIL(theSystemSumHashedLength);
		END IF;

        DROP TABLE bpBackupNumPerSystemId;
	
    END IF;
	-- debugging purpose only
	--RAISE NOTICE '%', theQuery;
	--RAISE NOTICE 'distinct=% system=% footprint=%', theDistinctHashId, theSystemId, theSystemSumHashedLength;

	RETURN theSystemSumHashedLength::BIGINT;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

---------------------------------------------------------------------------------------

---------------------------------------------------------------------------------------

/*  @desc: bpGetRelatedBackupsOptimized returns a tuple of backup ids  supporting info for the all backups
           that will be affected if the specified backup is deleted, sorted sequentially by time.
    @input args: backup_id (integer) types in the order - full, diff, incr (array) and a filter (string)
    @output: bpbackupstuple (set of rows) that belong to the group of backups to be restored if successful
              if no-records are found or non supported backup type, it returns a single row (0,0,0,0,0,0)
    @error: not implemented (and defaults to postgres exception handling)
*/

---------------------------------------------------------------------------------------
\set ON_ERROR_STOP off
  DROP TYPE getrelatedbackupstuple CASCADE;
  CREATE TYPE getrelatedbackupstuple AS(T_Backup_No BIGINT, T_Properties BIGINT, T_Status BIGINT, T_Exit_Code BIGINT, T_Type INTEGER, T_Start_Time BIGINT);
\set ON_ERROR_STOP on
CREATE OR REPLACE FUNCTION bpGetRelatedBackupsOptimized(integer, integer[])
RETURNS SETOF bp.getrelatedbackupstuple AS
$BODY$

DECLARE

    -- Record holds the rows returned
    theRecord bp.getrelatedbackupstuple%ROWTYPE;

    -- return values per row
    T_Backup_No BIGINT; T_Status BIGINT; T_Properties BIGINT; T_Exit_Code BIGINT; T_Type INTEGER; T_Start_Time BIGINT;

    -- instance_id, latest-full, latest-full-or-diff
    theInstanceId INTEGER;
	theBackupProps INTEGER;
    previousFull INTEGER;
    nextFull VARCHAR; -- require VARCHAR type (see below ~+)
    endOfList INTEGER;

    -- supported backup types
    IsTypeExchange BOOLEAN := FALSE;
    IsTypeMSSQL BOOLEAN := FALSE;
    IsTypeVMESX BOOLEAN := FALSE;
    IsTypeHyperV BOOLEAN := FALSE;
    IsTypeOracle BOOLEAN := FALSE;
    IsTypeSharePt BOOLEAN := FALSE;
    IsTypeUCS BOOLEAN := FALSE;
    IsTypeNdmp BOOLEAN := FALSE;

    -- Assume result is not null
    IsResultNull INTEGER := 1;

    -- placeholder for query
    theQuery TEXT;
    theTempTable VARCHAR;
    -- filter
    theFilter TEXT;

    -- parameters passed to function
    theBackupId ALIAS FOR $1;
    theBTypes ALIAS FOR $2;

BEGIN

    theInstanceId := 0; theBackupProps := 0; previousFull := 0; nextFull := ''; endOfList := 0;
    
        -- Get instance_id for this backup
    SELECT instance_id, properties INTO theInstanceId, theBackupProps
	FROM bp.backups WHERE bp.backups.backup_no = theBackupId;

    -- properties = same imported status (& 32),
    -- properties = not securesync (& 1 = 0),
    -- status = final ( > 64)
    theFilter := '(properties & 32) = (' || theBackupProps || ' & 32)'
	             || ' AND (properties & 1) = 0 AND (status & 64) > 0';
    
    -- Get instance_id for this backup
    theInstanceId := (SELECT instance_id FROM bp.backups WHERE bp.backups.backup_no = theBackupId);
    
    -- Get type (Sharepoint, Oracle, UCS, NoDeps, Supported Apps - MSSQL, VMESX, HyperV)
    -- Note: the order of theBTypes matters with theBTypes[1] indicating FULL backup: ordering of arguments
    -- are tightly coupled to bpGetRelatedBackupsOptimized() function in bp/libsrc/bpsql.c
    
    IsTypeExchange := (SELECT theBTypes[1] = 5);
    IsTypeMSSQL := (SELECT theBTypes[1] = 9);
    IsTypeVMESX := (SELECT theBTypes[1] = 12);
    IsTypeHyperV := (SELECT theBTypes[1] = 1031);
    IsTypeOracle := (SELECT theBTypes[1] = 1034);
    IsTypeSharePt := (SELECT theBTypes[1] = 1036);
    IsTypeUCS := (SELECT theBTypes[1] = 1039);
    IsTypeNdmp := (SELECT theBTypes[1] = 4098);
    
    -- temp table name
    theTempTable := (SELECT 'bpGRB' || md5(random()::VARCHAR));

    -- Get the group corresponding to this instance_id, backup_no and type in (full, diff, incr) 
    theQuery := 'CREATE TEMPORARY TABLE '|| theTempTable || ' AS
    (SELECT backup_no, type, status, properties, exit_code, start_time, sync_resv, tape_no FROM bp.backups
    WHERE instance_id = '|| theInstanceId || ' AND type IN ('||theBTypes[1]||','||theBTypes[2]||','||theBTypes[3]||'));';
    
    EXECUTE theQuery;

    theQuery := 'SELECT count(1) FROM '|| theTempTable;

    EXECUTE theQuery INTO IsResultNull;
   
    --RAISE NOTICE '% % ', theTempTable, IsResultNull;
 
    IF (IsResultNull = 0) THEN
        FOR theRecord IN (SELECT 0,0,0,0,0,0)
        LOOP
            RETURN NEXT theRecord;
        END LOOP;
    ELSE

        -- NOTE1: For previousFull and nextFull, we have additional filters (status & 36888) = 0
        --       the status mask is: ST_FAILED ST_CANCELLED ST_CANCELLED_SENT ST_PROC_DIED
        -- Get previousFull
        theQuery := '(SELECT backup_no FROM ' || theTempTable || ' rg WHERE rg.type = ' || theBTypes[1]
            || ' AND rg.backup_no <= ' || theBackupId || ' AND ' || theFilter || ' AND (status & 36888) = 0 ORDER BY rg.backup_no DESC LIMIT 1);';

        EXECUTE theQuery INTO previousFull;

        -- Get nextFull
        /* --~+ NOTE2: nextFull should be set to a string value 'NULL' if no rows are returned by the query
           --~+ so that COALESCE(NULL,endOfList) = endOfList. The implementation is
           --~+ (a) equivalent to quote_nullable() function which does NOT exist in PG8.1.23 (CentOS5) and
           --~+ (b) is required to prevent occurrence of postgres ERROR:  cannot EXECUTE a null querystring
        */
        theQuery := '(SELECT backup_no FROM ' || theTempTable || ' rg WHERE rg.type = ' || theBTypes[1]
          || ' AND rg.backup_no > ' || theBackupId || ' AND ' || theFilter || ' AND (status & 36888) = 0 ORDER BY rg.backup_no ASC LIMIT 1)';

        EXECUTE 'SELECT CASE WHEN NOT EXISTS (' || theQuery || ') THEN ''NULL'' ELSE ' || theQuery || '::VARCHAR END;' INTO nextFull;
    

        -- Get end-of-list (latest)
        -- get the increment [ ie, max(backup_no)+1 ] for this instance_id and therefore guarantee the upper bound (end backup)
        theQuery := 'SELECT MAX(backup_no)+1 FROM ' || theTempTable; 
	EXECUTE theQuery INTO endOfList;

        -- For supported types, return rows for related backups
        IF (IsTypeExchange OR IsTypeMSSQL OR IsTypeVMESX OR IsTypeHyperV OR IsTypeOracle OR IsTypeSharePt OR IsTypeUCS OR IsTypeNdmp) THEN
            --FOR theRecord IN SELECT * FROM bpGetRelatedBackupsResults
            FOR theRecord IN EXECUTE 'SELECT rg.backup_no, rg.status, rg.properties, rg.exit_code, rg.type, rg.start_time
		FROM '|| theTempTable ||' rg WHERE rg.backup_no >= '|| previousFull ||' AND rg.backup_no != '|| theBackupId ||
		' AND rg.backup_no < COALESCE(' || nextFull || ',' || endOfList || ') AND sync_resv <> 1 AND tape_no != 4294967295 
		AND '|| theFilter || ' ORDER BY backup_no;'

            LOOP 
                RETURN NEXT theRecord; 
            END LOOP;
        ELSE
            FOR theRecord IN (SELECT 0,0,0,0,0,0)
            LOOP
                RETURN NEXT theRecord;
            END LOOP;
        END IF;
    END IF;

    theQuery := 'DROP TABLE ' || theTempTable;
    EXECUTE theQuery;
    -- clean up (we are using PG8.1.23(CentOS5)  which does not support CREATE TEMPORARY TABLE DROP ON COMMIT)
    -- DROP TABLE bpGetRelatedBackupsByInstanceID, bpGetRelatedBackupsResults;
    --TRUNCATE TABLE bpGetRelatedBackupsResults;
    RETURN;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;
---------------------------------------------------------------------------------------

---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION update_summary_charts(subsystem VARCHAR,
	elapsed_time BIGINT, read_size BIGINT, write_size BIGINT, compress_size BIGINT,
	read_time BIGINT, write_time BIGINT, compress_time BIGINT)
RETURNS BOOLEAN AS
$BODY$
-- This signature is called by the backup subsystem, and modifies these columns
-- in the summary_charts table related to backup metrics:
--   backup_elapsed_time: add to the sum of total elapsed backup time
--   backup_count: increment the number of backups for today
--   backup_read_size: sum of total data sent by clients
--   backup_write_size: either compress_size or read-size if not compressed
--   backup_compress_size: add to sum of compressed data written
--   backup_read_time: sum of total secs reading on data channel
--   backup_write_time: sum of total secs writing data buffers
--   backup_compress_time: sum of total secs in calls to compression engine

	DECLARE today DATE = now()::date;

BEGIN
	IF (subsystem <> 'backup')
	THEN 
		RAISE NOTICE 'Invalid subsystem: %, this function is for backup.',
			subsystem;
		RETURN FALSE;
	END IF;

	IF EXISTS (SELECT TRUE FROM bp.summary_charts WHERE insert_date = today)
	THEN
		-- Update an existing record for today
		UPDATE bp.summary_charts SET
			backup_elapsed_time = backup_elapsed_time + elapsed_time,
			backup_count = backup_count + 1,
			backup_read_size = backup_read_size + read_size,
			backup_write_size = backup_write_size + write_size,
			backup_compress_size = backup_compress_size + compress_size,
			backup_read_time = backup_read_time + read_time,
			backup_write_time = backup_write_time + write_time,
			backup_compress_time = backup_compress_time + compress_time
		WHERE insert_date = today;
	ELSE
		-- Insert a new record for today
		INSERT INTO bp.summary_charts (backup_elapsed_time, backup_count,
			backup_read_size, backup_write_size, backup_compress_size,
			backup_read_time, backup_write_time, backup_compress_time, insert_date)
		VALUES (elapsed_time, 1, read_size, write_size, compress_size,
			read_time, write_time, compress_time, today);
	END IF;

	RETURN TRUE;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

CREATE OR REPLACE FUNCTION update_summary_charts(subsystem VARCHAR,
		elapsed_time BIGINT, data_size BIGINT)
RETURNS BOOLEAN AS
$BODY$
-- This signature is called by either the restore or archive subsystem.
-- For restore, these columns in summary_charts are modified:
--   restore_size_total: sum of total data sent during restores
--   restore_elapsed_time: sum of total time in restore
--   restore_count: increment the number of restore jobs for today.
-- For archive, these columns in summary_charts are modified:
--   archive_size_total: add to sum of total data archived
--   archive_elapsed_time: add to the sum to total elapsed archiving time
--   archive_count: increment the number of archive jobs for today.

	DECLARE today DATE = now()::date;

BEGIN
	IF (subsystem <> 'restore' AND subsystem <> 'archive') 
	THEN
		RAISE NOTICE 'Invalid subsystem: %, this function is restore or archive.',
			subsystem;
		RETURN FALSE;
	END IF;

	IF EXISTS (SELECT TRUE FROM bp.summary_charts WHERE insert_date = today)
	THEN
		-- Update an existing record for today
		IF (subsystem = 'archive')
		THEN
			UPDATE bp.summary_charts SET
				archive_size_total = archive_size_total + data_size,
				archive_elapsed_time = archive_elapsed_time + elapsed_time,
				archive_count = archive_count + 1
			WHERE insert_date = today;
		ELSEIF (subsystem = 'restore')
		THEN
			UPDATE bp.summary_charts SET
				restore_size_total = restore_size_total + data_size,
				restore_elapsed_time = restore_elapsed_time + elapsed_time,
				restore_count = restore_count + 1
			WHERE insert_date = today;
		ELSE 
			RAISE NOTICE 'Invalid subsystem: %', subsystem;
			RETURN FALSE;
		END IF;
	ELSE
		-- Insert a new record for today
		IF (subsystem = 'archive')
		THEN
			INSERT INTO bp.summary_charts(insert_date, 
					archive_size_total, archive_elapsed_time, archive_count)
			VALUES(today, data_size, elapsed_time, 1);
		ELSEIF (subsystem = 'restore')
		THEN
			INSERT INTO bp.summary_charts(insert_date,
					restore_size_total, restore_elapsed_time, restore_count)
			VALUES(today, data_size, elapsed_time, 1);
		ELSE 
			RAISE NOTICE 'Invalid subsystem: %', subsystem;
			RETURN FALSE;
		END IF;
	END IF;

	RETURN TRUE;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

CREATE OR REPLACE FUNCTION update_summary_charts(subsystem VARCHAR,
	read_size BIGINT, write_size BIGINT, read_time BIGINT, write_time BIGINT)
RETURNS BOOLEAN AS
$BODY$
-- This signature is called by the replication subsystem (on a source), and 
-- modifies these columns in the summary_charts table: 
--   replication_read_size: size of the backup that was replicated
--   replication_write_size: bytes sent to the target
--   replication_read_time: sum of total secs spent reading/preparing data
--   replication_write_time: sum of total secs sending data to target

	DECLARE today DATE = now()::date;

BEGIN
	IF (subsystem <> 'replication')
	THEN 
		RAISE NOTICE 'Invalid subsystem: %, this function is for replication.',
			subsystem;
		RETURN FALSE;
	END IF;

	IF EXISTS (SELECT TRUE FROM bp.summary_charts WHERE insert_date = today)
	THEN
		-- Update an existing record for today
		UPDATE bp.summary_charts SET
			replication_read_size = replication_read_size + read_size,
			replication_write_size = replication_write_size + write_size,
			replication_read_time = replication_read_time + read_time,
			replication_write_time = replication_write_time + write_time
		WHERE insert_date = today;
	ELSE
		-- Insert a new record for today
		INSERT INTO bp.summary_charts (replication_read_size,
			replication_write_size, replication_read_time,
			replication_write_time, insert_date)
		VALUES (read_size, write_size, read_time, write_time, today);
	END IF;

	RETURN TRUE;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;
---------------------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION update_storage_history (id INTEGER, storage_size BIGINT, 
	 storage_used BIGINT, storage_free BIGINT)
RETURNS BOOLEAN AS
$BODY$
-- These columns in storage_history are modified :
--   mb_size: size of the storage 'id' in MB.
--   mb_used: used space of storage 'id' in MB.
--   mb_free: available free space on the storage 'id' in MB.
	DECLARE today DATE = now()::date;

BEGIN
	IF EXISTS (SELECT TRUE FROM bp.storage_history WHERE storage_id = id AND insert_date = today)
	THEN
		-- Update an existing record for today
		UPDATE bp.storage_history SET
			mb_size = storage_size,
			mb_used = storage_used,
			mb_free = storage_free
		WHERE storage_id = id AND insert_date = today;
	ELSE
		-- Insert a new record for today
			INSERT INTO bp.storage_history(storage_id, insert_date, 
					mb_size, mb_used, mb_free)
			VALUES(id, today, storage_size, storage_used, storage_free);
	END IF;

	RETURN TRUE;
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;
---------------------------------------------------------------------------------------

DROP FUNCTION conditionally_drop_function(varchar,varchar);

