-- Add statements/functions for updating the database.
-- This script is primarily used for doing upgrades to the schema
-- in future releases.  It will run on fresh install after the
-- base schema scripts have run.
SET search_path TO public;

--
-- Create table modification functions
--
CREATE OR REPLACE FUNCTION table_exists(sch_name VARCHAR,tbl_name VARCHAR)
RETURNS BOOLEAN AS
$BODY$BEGIN
-- table names are case insensitive, but where VARCHAR compare function is
        RETURN ((SELECT COUNT(*) FROM pg_catalog.pg_tables WHERE
        lower(schemaname)=lower(sch_name) AND lower(tablename)=lower(tbl_name))>0);
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

-- Returns TRUE if the specified index exists
CREATE OR REPLACE FUNCTION index_exists(sch_name VARCHAR,idx_name VARCHAR)
RETURNS BOOLEAN AS
$BODY$BEGIN
-- index names are case insensitive, but VARCHAR compare function is
    RETURN ((
	SELECT COUNT(*) FROM pg_catalog.pg_class c, pg_catalog.pg_namespace n
    WHERE c.relkind = 'i'
    	AND c.relnamespace = n.oid
		AND lower(n.nspname) = lower(sch_name)
		AND lower(c.relname) = lower(idx_name)
	) > 0);

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

-- Returns TRUE if the specified table column exists
CREATE OR REPLACE FUNCTION column_exists(sch_name VARCHAR,tbl_name VARCHAR,col_name VARCHAR)
RETURNS BOOLEAN AS
$BODY$BEGIN

	RETURN((
	SELECT COUNT(*) FROM pg_catalog.pg_class c, pg_catalog.pg_namespace n, pg_catalog.pg_attribute a
	WHERE a.attnum > 0
		AND NOT a.attisdropped
		AND a.attrelid = c.oid
		AND c.relnamespace = n.oid
		AND pg_catalog.pg_table_is_visible(c.oid)
		AND lower(n.nspname) = lower(sch_name)
		AND lower(c.relname) = lower(tbl_name)
		AND lower(a.attname) = lower(col_name)
	) > 0);

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

-- Returns TRUE if the specified view exists
CREATE OR REPLACE FUNCTION view_exists(sch_name VARCHAR,view_name VARCHAR)
RETURNS BOOLEAN AS
$BODY$BEGIN
    RETURN ((
	SELECT COUNT(*) FROM pg_catalog.pg_class c, pg_catalog.pg_namespace n
    WHERE c.relkind = 'v'
    	AND c.relnamespace = n.oid
		AND lower(n.nspname) = lower(sch_name)
		AND lower(c.relname) = lower(view_name)
	) > 0);
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

-- Renames conflicts for a given table and column name.  Returns the given
-- requested_name so that the function can be used inside INSERT/UPDATE
-- statements.  If something had to be renamed and canonical_obj_name is
-- supplied, an alert specifying the change is raised.
CREATE OR REPLACE FUNCTION clear_name(sch_name VARCHAR, tbl_name VARCHAR, col_name VARCHAR, requested_name VARCHAR, canonical_obj_name VARCHAR)
RETURNS VARCHAR AS
$BODY$
	DECLARE i INTEGER := 1;
	DECLARE qprefix VARCHAR;
	DECLARE qsuffix VARCHAR;
	DECLARE qstr VARCHAR;
	DECLARE name_exists BOOLEAN;
	DECLARE new_name VARCHAR;
	DECLARE alert_str VARCHAR;
	DECLARE first_char VARCHAR;
	DECLARE len INTEGER;
BEGIN
	-- Store query prefix and suffix, the parts before/after the requested name
	qprefix := 'SELECT COUNT(*) > 0 FROM ' || quote_ident(sch_name) || '.' || quote_ident(tbl_name)
	        || ' WHERE ' || quote_ident(col_name) || ' = ';
	qsuffix := ' LIMIT 1';
	qstr := qprefix || quote_literal(requested_name) || qsuffix;

	-- If the name exists, loop through numbers until we have an unused name; else exit
	EXECUTE qstr INTO name_exists;
	IF (name_exists = true) THEN
		LOOP
			-- Set new name: "requested_name-customer" or "requested_name-customer2" are ideal, but we
			-- can go with "requested_name-2" or some other shorthand if we're tight on space
			len := char_length(requested_name);
			new_name := requested_name || '-customer';
			IF (i = 1) THEN
				IF (len >= 39) THEN -- 1 or less chars available, take 39 chars of requested_name and append "2"
					i := 2;
					new_name := substring(requested_name from 1 for 39) || '2';
				ELSE -- 2 or more chars available, take requested_name and append as much of "-customer" as will fit
					new_name := requested_name || substring('-customer' from 1 for (40 - len));
				END IF;
			ELSE
				len := len + char_length(i);
				IF (len >= 40) THEN -- the name and number fill up the field, store as much of name as possible and append i
					new_name := substring(requested_name from 1 for (40 - char_length(i))) || i;
				ELSE -- there is extra room besides name and number, so store name, as much of "-customer" as we can, and i
					new_name := requested_name || substring('-customer' from 1 for (40 - len)) || i;
				END IF;
			END IF;
			qstr := qprefix || quote_literal(new_name) || qsuffix;

			-- Does the name exist?  If not, rename the value
			EXECUTE qstr INTO name_exists;
			IF (name_exists = false) THEN
				qstr := 'UPDATE ' || quote_ident(sch_name) || '.' || quote_ident(tbl_name)
				     || ' SET ' || quote_ident(col_name) || ' = ' || quote_literal(new_name)
				     || ' WHERE ' || quote_ident(col_name) || ' = ' || quote_literal(requested_name);
				EXECUTE qstr;

				-- Raise an alert about renaming this value
				IF (canonical_obj_name != '') THEN
					first_char := substring(canonical_obj_name from 1 for 1);
					alert_str := 'Update: A';
					IF (first_char = 'a' OR first_char = 'e' OR first_char = 'i' OR first_char = 'o' OR first_char = 'u') THEN
						alert_str := alert_str || 'n';
					END IF;
					alert_str := alert_str || ' ' || canonical_obj_name || ' with name "' || requested_name
					          || '" was found and renamed to "' || new_name || '".';
					INSERT INTO bp.alerts(alert_text, source_id, target, severity, created, updated, closed)
					 VALUES (alert_str, (SELECT source_id FROM bp.alert_source WHERE source_name = 'Process'), 'update', 3,
					 (SELECT FLOOR(EXTRACT(EPOCH FROM NOW()))), (SELECT FLOOR(EXTRACT(EPOCH FROM NOW()))), true);
				END IF;

				EXIT; -- Exit loop
			END IF;

			i := i + 1;
		END LOOP;
	END IF;
	RETURN requested_name;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

-- Function to conditionally add new rows to lookup tables if the row doesn't exist already
CREATE OR REPLACE FUNCTION bp.ifn1(
	p_schtbl VARCHAR,
	p_idcol VARCHAR,
	p_idval VARCHAR)
RETURNS VARCHAR AS
$BODY$
	DECLARE qstr VARCHAR;
BEGIN
	qstr := 'INSERT INTO ' || p_schtbl || '(' || p_idcol || ')'
	     || ' SELECT ' || quote_literal(p_idval) 
	     || ' WHERE (SELECT COUNT(*) FROM ' || p_schtbl
	            || ' WHERE ' || p_idcol || ' = ' || quote_literal(p_idval) || ') = 0';
	EXECUTE qstr;
--	RETURN p_schtbl||'(\''||p_idval||'\')';
	RETURN p_schtbl||E'(\''||p_idval||E'\')';
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

-- Function to conditionally add new rows to lookup tables if the row doesn't exist already
CREATE OR REPLACE FUNCTION bp.ifn2(
	p_schtbl VARCHAR,
	p_idcol VARCHAR,
	p_col2 VARCHAR,
	p_idval VARCHAR,
	p_val2 VARCHAR)
RETURNS VARCHAR AS
$BODY$
	DECLARE qstr VARCHAR;
BEGIN
	qstr := 'INSERT INTO ' || p_schtbl || '(' || p_idcol || ', ' || p_col2 || ')'
	     || ' SELECT ' || quote_literal(p_idval) || ', ' || quote_literal(p_val2)
	     || ' WHERE (SELECT COUNT(*) FROM ' || p_schtbl
	            || ' WHERE ' || p_idcol || ' = ' || quote_literal(p_idval) || ') = 0';
	EXECUTE qstr;
--	RETURN p_schtbl||'(\''||p_idval||'\', \''||p_val2||'\')';
	RETURN p_schtbl||E'(\''||p_idval||E'\', \''||p_val2||E'\')';
END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

set search_path to bp;
--
-- Create views for checking bloat values of tables
--
--------------------------------------------------------------------------------------------------------------
-- datawidth
--------------------------------------------------------------------------------------------------------------
CREATE OR REPLACE VIEW bloat_datawidth_bpdb AS
SELECT ns.nspname AS schemaname, tbl.oid AS relid, tbl.relname,
  CASE WHEN every(avg_width IS NOT NULL) 
  THEN round((SUM((1-null_frac)*avg_width) + MAX(null_frac) * 24)::NUMERIC, 2) ELSE NULL END AS datawidth
  FROM pg_attribute att JOIN pg_class tbl ON att.attrelid = tbl.oid
  JOIN pg_namespace ns ON ns.oid = tbl.relnamespace LEFT JOIN pg_stats s
  ON s.schemaname=ns.nspname AND s.tablename = tbl.relname /*AND s.inherited=false*/ AND
  s.attname=att.attname AND s.schemaname = 'bp'
  WHERE att.attnum > 0 AND tbl.relkind in ('r', 'i') GROUP BY 1,2,3;


--------------------------------------------------------------------------------------------------------------
-- bloat
--------------------------------------------------------------------------------------------------------------
CREATE OR REPLACE VIEW bloat_tables_bpdb AS
SELECT bdw.schemaname, bdw.relname, bdw.datawidth,
  cc.reltuples::bigint, cc.relpages::bigint,
  ceil(cc.reltuples*bdw.datawidth/current_setting('block_size')::NUMERIC)::bigint AS expectedpages,
  --100 - (cc.reltuples*100*bdw.datawidth)/(current_setting('block_size')::NUMERIC*cc.relpages) AS bloatpct
  (100 - (cc.reltuples*100*bdw.datawidth)/(current_setting('block_size')::NUMERIC*cc.relpages))::NUMERIC(5,2) AS bloatpct
  FROM bloat_datawidth_bpdb bdw JOIN pg_class cc ON cc.oid = bdw.relid
  AND bdw.schemaname = 'bp' AND bdw.datawidth IS NOT NULL AND cc.relpages > 1 ORDER BY bloatpct DESC;


--------------------------------------------------------------------------------------------------------------
-- table size
--------------------------------------------------------------------------------------------------------------
CREATE OR REPLACE VIEW size_tables_bpdb AS
SELECT schemaname, tablename,
  pg_relation_size(schemaname || '.' || tablename) AS size_p,
  pg_total_relation_size(schemaname || '.' || tablename) AS siz,
  pg_total_relation_size(schemaname || '.' || tablename) AS total_size_p,
  pg_total_relation_size(schemaname || '.' || tablename) - pg_relation_size(schemaname || '.' || tablename) AS index_size,
  (100*(pg_total_relation_size(schemaname || '.' || tablename) - pg_relation_size(schemaname || '.' || tablename))) /
  CASE WHEN pg_total_relation_size(schemaname || '.' || tablename) = 0 THEN 1
  ELSE pg_total_relation_size(schemaname || '.' || tablename) END AS index_pct
  FROM pg_tables WHERE schemaname='bp' ORDER BY siz DESC LIMIT 10;



SET search_path TO bp,public;

--
-- Modify tables
--
CREATE OR REPLACE FUNCTION update_bp()
RETURNS INT4 AS
$BODY$BEGIN

-- Drop views first so that table updates
-- are reflected when views are re-created
IF view_exists('bp', 'restorable_backups') THEN
	DROP VIEW restorable_backups;
END IF;
IF view_exists('bp', 'successful_backups') THEN
	DROP VIEW successful_backups;
END IF;
IF view_exists('bp', 'sm_backups') THEN
	DROP VIEW sm_backups;
END IF;

-- Add storage_usage column to the bp.storage table
IF NOT column_exists('bp', 'storage', 'storage_usage') THEN
	ALTER TABLE storage ADD COLUMN storage_usage VARCHAR  NOT NULL DEFAULT 'backup';
END IF;

-- Add is_active and check_timestamp columns to the bp.systems table
IF NOT column_exists('bp', 'systems', 'is_active') THEN
	ALTER TABLE systems ADD COLUMN is_active BOOLEAN NOT NULL DEFAULT TRUE;
END IF;
IF NOT column_exists('bp', 'systems', 'check_timestamp') THEN
	ALTER TABLE systems ADD COLUMN check_timestamp BIGINT NOT NULL DEFAULT 0;
END IF;
IF NOT column_exists('bp', 'systems', 'version') THEN
	ALTER TABLE systems ADD COLUMN version VARCHAR NOT NULL DEFAULT 'Unknown';
END IF;
IF NOT column_exists('bp', 'systems', 'role') THEN
	ALTER TABLE systems ADD COLUMN role VARCHAR NOT NULL DEFAULT 'Unknown';
END IF;
IF NOT column_exists('bp', 'systems', 'storage_id') THEN
	ALTER TABLE systems ADD COLUMN storage_id INTEGER NOT NULL DEFAULT 1 REFERENCES storage(storage_id);
END IF;

IF NOT column_exists('bp', 'operation_progress', 'pid_to_kill') THEN
	ALTER TABLE operation_progress ADD COLUMN pid_to_kill BIGINT DEFAULT 0;
END IF;

IF NOT column_exists('bp','node_applications','visible') THEN
	ALTER TABLE node_applications ADD COLUMN visible BOOLEAN NOT NULL DEFAULT TRUE;
END IF;

-- Increase length limit of names from 32 to 40 characters
ALTER TABLE bp.calendars ALTER COLUMN name TYPE VARCHAR(40);
ALTER TABLE bp.schedules ALTER COLUMN name TYPE VARCHAR(40);
ALTER TABLE bp.selection_lists ALTER COLUMN name TYPE VARCHAR(40);
ALTER TABLE bp.option_lists ALTER COLUMN name TYPE VARCHAR(40);

IF NOT table_exists('bp', 'schedule_client_assoc') THEN
	-- Only do this if upgrading from a version that has application schema,
	-- release 4.1.x doesn't have these tables.
	IF table_exists('bp', 'application_schedules') THEN
		--
		-- Create view of schedules, lcs records, profiles, and option/selection lists
		--
		CREATE OR REPLACE VIEW schedule_profiles AS
	  	SELECT schedule_lcs.schedule_id, lcs_client_assoc.*, profiles.profile_path
	  	FROM schedule_lcs
	  	RIGHT JOIN lcs_client_assoc USING (lcs_no)
	  	JOIN (SELECT DISTINCT ON (lcs_no, node_no, type) lcs_no, node_no, type, profile_path
			FROM lcs_client ORDER BY lcs_no, node_no, type, profile_path) AS profiles
	    	USING (lcs_no, node_no, type);

		--
		-- Create view of application schedules, profiles, and backup options
		--
		CREATE OR REPLACE VIEW application_profiles AS
			SELECT DISTINCT application_schedule_id AS schedule_id, node_no,
				summary AS type, instance_id, verify_level, device_name,
				'schedules/' || node_name || '_' || instance_id || '_' ||
				CASE
					WHEN summary=5 THEN 'exchmast_'
					WHEN summary=6 THEN 'exchdiff_'
					WHEN summary=7 THEN 'exchincr_'
					WHEN summary=10 THEN 'mssqlfull_'
					WHEN summary=11 THEN 'mssqldiff_'
					WHEN summary=12 THEN 'mssqltrans_'
					WHEN summary=14 THEN 'vmesxfull_'
					WHEN summary=15 THEN 'vmesxdiff_'
					WHEN summary=16 THEN 'vmesxincr_'
					WHEN summary=18 THEN 'hypervfull_'
					WHEN summary=19 THEN 'hypervdiff_'
					WHEN summary=20 THEN 'hypervincr_'
					WHEN summary=28 THEN 'ndmpfull_'
					WHEN summary=29 THEN 'ndmpdiff_'
					WHEN summary=30 THEN 'ndmpincr_'
				END || application_schedule_id || '.spr' AS profile_path
			FROM application_schedules JOIN events USING(calendar_id)
				JOIN option_lists USING(option_list_id)
				JOIN schedule_instances USING(application_schedule_id)
				JOIN nodes USING(node_no);
	END IF;
ELSE
	CREATE OR REPLACE VIEW schedule_profiles AS
	  SELECT schedules.schedule_id, schedule_client_assoc.instance_id, schedule_client_assoc.type, 
		schedule_client_assoc.option_list_id, schedule_client_assoc.selection_incl_id, schedule_client_assoc.selection_excl_id, 
		application_instances.node_no, 'schedules/' || node_name || '_' ||
		CASE
			WHEN schedule_client_assoc.type=1 THEN 'mast'
			WHEN schedule_client_assoc.type=2 THEN 'incr'
			WHEN schedule_client_assoc.type=3 THEN 'sele'
			WHEN schedule_client_assoc.type=4 THEN 'iabg'
			WHEN schedule_client_assoc.type=17 THEN 'inc4ever'
		ELSE 'err' END
		|| '_' || schedules.schedule_id || '-' || schedule_client_assoc.instance_id || '.spr' AS profile_path
	  FROM schedules
	  RIGHT JOIN schedule_client_assoc USING (schedule_id)
	  JOIN application_instances USING (instance_id)
	  JOIN nodes USING (node_no) WHERE schedules.app_id = 1 AND schedules.run_only = FALSE;

	-- Create view of application schedules, profiles, and backup options
	CREATE OR REPLACE VIEW application_profiles AS
	  SELECT DISTINCT schedule_id, node_no,
		summary AS type, instance_id, verify_level, device_name,
		'schedules/' || node_name || '_' || instance_id || '_' ||
		CASE
			WHEN summary=5 THEN 'exchmast_'
			WHEN summary=6 THEN 'exchdiff_'
			WHEN summary=7 THEN 'exchincr_'
			WHEN summary=10 THEN 'mssqlfull_'
			WHEN summary=11 THEN 'mssqldiff_'
			WHEN summary=12 THEN 'mssqltrans_'
			WHEN summary=14 THEN 'vmesxfull_'
			WHEN summary=15 THEN 'vmesxdiff_'
			WHEN summary=16 THEN 'vmesxincr_'
			WHEN summary=18 THEN 'hypervfull_'
			WHEN summary=19 THEN 'hypervdiff_'
			WHEN summary=20 THEN 'hypervincr_'
			WHEN summary=28 THEN 'ndmpfull_'
			WHEN summary=29 THEN 'ndmpdiff_'
			WHEN summary=30 THEN 'ndmpincr_'
		END || schedule_id || '.spr' AS profile_path
	FROM schedules JOIN events USING(calendar_id)
		JOIN schedule_client_assoc USING(schedule_id, option_list_id)
		JOIN option_lists USING(option_list_id)
		JOIN application_instances USING (instance_id)
		JOIN nodes USING(node_no) WHERE schedules.app_id != 1 AND run_only = FALSE;
END IF;

	RETURN 0;
END;
$BODY$
LANGUAGE 'plpgsql' VOLATILE;

SELECT update_bp();
DROP FUNCTION update_bp();




--
-- Add new values into lookup tables
--

-- Add new alert types
SELECT bp.ifn2('alert_source', 'source_id', 'source_name', '6', 'Disk');
SELECT bp.ifn2('alert_source', 'source_id', 'source_name', '7', 'CEP');
SELECT bp.ifn2('alert_source', 'source_id', 'source_name', '8', 'DPU');
SELECT bp.ifn2('alert_source', 'source_id', 'source_name', '9', 'RAID');
SELECT bp.ifn2('alert_source', 'source_id', 'source_name', '10', 'Backup');
SELECT bp.ifn2('alert_source', 'source_id', 'source_name', '11', 'Archiving');
SELECT bp.ifn2('alert_source', 'source_id', 'source_name', '12', 'Retention');
SELECT bp.ifn2('alert_source', 'source_id', 'source_name', '13', 'Disaster Recovery');

-- Add new sync_type names
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '4', 'LocalDir');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '5', 'SQL');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '6', 'DPUstate');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '7', 'Exchange');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '8', 'MSExch Full');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '9', 'MSExch Incr');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '10', 'MSExch Diff');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '11', 'MSSQL Full');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '12', 'MSSQL Diff');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '13', 'MSSQL Trans');
UPDATE bp.sync_type SET type_name = 'Differential' WHERE type_id = 2
	AND type_name != 'Differential';
UPDATE bp.sync_type SET type_name = 'MSExch Diff' WHERE type_id = 10
	AND type_name != 'MSExch Diff';
UPDATE bp.sync_type SET type_name = 'MSSQL Diff' WHERE type_id = 12
	AND type_name != 'MSSQL Diff';
UPDATE bp.sync_type SET type_name = 'MSSQL Trans' WHERE type_id = 13
	AND type_name != 'MSSQL Trans';
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '14', 'ESX VM Full');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '15', 'ESX VM Diff');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '16', 'ESX VM Incr');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '17', 'Hyper-V Full');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '18', 'Hyper-V Diff');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '19', 'Hyper-V Incr');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '20', 'System Metadata');
UPDATE bp.sync_type SET type_name = 'Hyper-V Full' WHERE type_id = 17
	AND type_name != 'Hyper-V Full';
UPDATE bp.sync_type SET type_name = 'Hyper-V Diff' WHERE type_id = 18
	AND type_name != 'Hyper-V Diff';
UPDATE bp.sync_type SET type_name = 'Hyper-V Incr' WHERE type_id = 19
	AND type_name != 'Hyper-V Incr';
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '21', 'Oracle Full');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '22', 'Oracle Incr');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '23', 'SharePoint Full');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '24', 'SharePoint Diff');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '25', 'Incremental');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '26', 'UCS Service Profile Full');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '27', 'NDMP Full');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '28', 'NDMP Diff');
SELECT bp.ifn2('sync_type', 'type_id', 'type_name', '29', 'NDMP Incr');

-- Add new operation_names values for name_value column
SELECT bp.ifn1('operation_names', 'name_value', 'migrate');
SELECT bp.ifn1('operation_names', 'name_value', 'prepare');
SELECT bp.ifn1('operation_names', 'name_value', 'replicate');
SELECT bp.ifn1('operation_names', 'name_value', 'wait');
SELECT bp.ifn1('operation_names', 'name_value', 'rebuild');
SELECT bp.ifn1('operation_names', 'name_value', 'no_activity');
SELECT bp.ifn1('operation_names', 'name_value', 'processing');
SELECT bp.ifn1('operation_names', 'name_value', 'hash');
SELECT bp.ifn1('operation_names', 'name_value', 'dedup');
SELECT bp.ifn1('operation_names', 'name_value', 'redup');
SELECT bp.ifn1('operation_names', 'name_value', 'unhash');


-- Add new values for 3ware migration to disk_status table
SELECT bp.ifn2('disk_status', 'status_id', 'status', '10', 'MIGRATING');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '11', 'MIGRATE-PAUSED');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '12', 'INIT-PAUSED');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '13', 'RECOVERY');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '14', 'UNKNOWN');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '15', 'DCB-DATA-CHECK');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '16', 'DEGRADED-INIT');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '17', 'DEGRADED-MGRT');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '18', 'DEGRADED-RBLD');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '19', 'DEGRADED-VRFY');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '20', 'DEVICE-ERROR');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '21', 'DGRD-RBLD-INIT');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '22', 'DGRD-RBLD-MGRT');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '23', 'DGRD-RBLD-VRFY');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '24', 'ECC-ERROR');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '25', 'NOT-SUPPORTED');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '26', 'READ-FAILURE');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '27', 'READ-TIMEOUT');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '28', 'REBUILD-INIT');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '29', 'REBUILD-MGRT');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '30', 'REBUILD-VERIFY');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '31', 'OFFLINE-JBOD');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '32', 'ORPHAN');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '33', 'SMART-FAILURE');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '34', 'UNSUPP-DCB');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '35', 'UNCONV-DCB');
SELECT bp.ifn2('disk_status', 'status_id', 'status', '36', 'VERIFY-PAUSED');

-- Create pending migration operations
INSERT INTO bp.operation_status (operation_id, subsystem, description, status)
	SELECT 1, 'Securesync',
	    'Transition to using sync needed for BareMetals', 'pending' 
	WHERE NOT EXISTS (SELECT operation_id FROM bp.operation_status
			WHERE operation_id = 1);
INSERT INTO operation_status (operation_id, subsystem, description, status)
	SELECT 2, 'Installation',
		'Migrate file metadata for 5.0.2 schema', 'pending'
	WHERE NOT EXISTS (SELECT TRUE FROM operation_status WHERE operation_id = 2);
INSERT INTO operation_status (operation_id, subsystem, description, status)
	SELECT 3, 'Profiles',
		'Migrate templates and schedules for 5.1 unicode', 'pending'
	WHERE NOT EXISTS (SELECT TRUE FROM operation_status WHERE operation_id = 3);
-- 6.2.0: operation_id=5 is added for archive data migration
INSERT INTO operation_status (operation_id, subsystem, description, status)
	SELECT 5, 'Archiving',
		'Migrate data from archive_files to its child tables', 'pending'
	WHERE NOT EXISTS (SELECT TRUE FROM operation_status WHERE operation_id = 5);
-- 6.1.0: operation_id=4 is added below when sync,priority columns are added
-- to the application_instances table and values are migrated from nodes.


--
-- Create view of schedules, lcs records, profiles, and option/selection lists
--

--
-- Add default entries for calendars and schedules.
-- If adding any events or calendars, be sure to adjust the sequence number.
--
-- PostgreSQL database dump created using pg_dump -a -D -t <tablename> >> dumpfile.sql
--

SET check_function_bodies = false;


CREATE OR REPLACE FUNCTION populate_cal_strat()
RETURNS INT4 AS
$BODY$
DECLARE
	Vstr VARCHAR;
BEGIN

IF (SELECT COUNT(*) FROM bp.calendars)=0
THEN

	-- SEQUENCES -- there's room for up to 50 calendars, 50 strategies, and 200 events.
	-- This is to avoid complicated sequence adjustments if any more calendars or
	-- strategies are added, especially if the user has saved any of their own.
	ALTER SEQUENCE bp.strategies_strategy_id_seq RESTART WITH 51;
	ALTER SEQUENCE bp.calendars_calendar_id_seq RESTART WITH 51;
	ALTER SEQUENCE bp.events_event_id_seq RESTART WITH 201;

	-- STRATEGIES
	INSERT INTO strategies (strategy_id, name, description, read_only)
	 VALUES (5, 'Quarterly Master, Weekly Diff',
	         'Master each quarter, Differentials once each week', true);
	INSERT INTO strategies (strategy_id, name, description, read_only)
	 VALUES (8, 'Beginning of Month Master',
	         'Master Backup the first day of each month', true);
	INSERT INTO strategies (strategy_id, name, description, read_only)
	 VALUES (12, 'Weekday Selectives',
	         'Selective backup each weekday', true);

	-- STRATEGY EVENTS
	INSERT INTO events (event_id, calendar_id, strategy_id, summary,
	                     dtstart, freq, "interval", bymonthday, description)
	 VALUES (9, NULL, 5, 1, 1199156400, 4, 3, -1, 'quarterly master');
	INSERT INTO events (event_id, calendar_id, strategy_id, summary,
	                    dtstart, freq, "interval", bymonthday, description)
	 VALUES (10, NULL, 5, 2, 1199156400, 3, NULL, NULL, 'weekly differential');
	INSERT INTO events (event_id, calendar_id, strategy_id, summary,
	                    dtstart, freq, "interval", bymonthday, description)
	 VALUES (13, NULL, 8, 1, 1199156400, 4, NULL, 1, 'monthly master');
	INSERT INTO events (event_id, calendar_id, strategy_id, summary,
	                    dtstart, freq, "interval", bymonthday, description)
	 VALUES (17, NULL, 12, 3, 1199156400, 3, NULL, NULL, 'weekday selectives');

	-- STRATEGY BYDAY
	INSERT INTO byday (event_id, day_no) VALUES (10, 6);
	INSERT INTO byday (event_id, day_no) VALUES (17, 1);
	INSERT INTO byday (event_id, day_no) VALUES (17, 2);
	INSERT INTO byday (event_id, day_no) VALUES (17, 3);
	INSERT INTO byday (event_id, day_no) VALUES (17, 4);
	INSERT INTO byday (event_id, day_no) VALUES (17, 5);

	-- CALENDARS
	INSERT INTO calendars (calendar_id, name, description, family, read_only)
	 VALUES (5, 'Quarterly Master, Weekly Diff',
	         'Master each quarter, Differentials once each week', 'Any', true);
	INSERT INTO calendars (calendar_id, name, description, family, read_only)
	 VALUES (8, 'Beginning of Month Master',
	         'Master Backup the first day of each month', 'Any', true);
	INSERT INTO calendars (calendar_id, name, description, family, read_only)
	 VALUES (12, 'Weekday Selectives',
	         'Selective backup each weekday', 'Any', true);

	-- CALENDAR EVENTS
	INSERT INTO events (event_id, calendar_id, strategy_id, summary,
	                    dtstart, freq, "interval", bymonthday, description)
	 VALUES (32, 5, NULL, 1, 1199156400, 4, 3, -1, 'quarterly master');
	INSERT INTO events (event_id, calendar_id, strategy_id, summary,
	                    dtstart, freq, "interval", bymonthday, description)
	 VALUES (33, 5, NULL, 2, 1199156400, 3, NULL, NULL, 'weekly differential');
	INSERT INTO events (event_id, calendar_id, strategy_id, summary,
	                    dtstart, freq, "interval", bymonthday, description)
	 VALUES (36, 8, NULL, 1, 1199156400, 4, NULL, 1, 'monthly master');
	INSERT INTO events (event_id, calendar_id, strategy_id, summary,
	                    dtstart, freq, "interval", bymonthday, description)
	 VALUES (40, 12, NULL, 3, 1199156400, 3, NULL, NULL, 'weekday selectives');

	-- CALENDAR BYDAY
	INSERT INTO byday (event_id, day_no) VALUES (33, 6);
	INSERT INTO byday (event_id, day_no) VALUES (40, 1);
	INSERT INTO byday (event_id, day_no) VALUES (40, 2);
	INSERT INTO byday (event_id, day_no) VALUES (40, 3);
	INSERT INTO byday (event_id, day_no) VALUES (40, 4);
	INSERT INTO byday (event_id, day_no) VALUES (40, 5);
ELSE 
	-- Update "Incr" to "Diff"
	UPDATE strategies SET name = 'Quarterly Master, Weekly Diff', 
	       description = 'Master each quarter, Differentials once each week'
	   WHERE strategy_id = 5 AND name != 'Quarterly Master, Weekly Diff';
	UPDATE events SET description = 'weekly differential'
	   WHERE event_id = 10 AND description != 'weekly differential';
	UPDATE calendars SET name = 'Quarterly Master, Weekly Diff', 
	       description = 'Master each quarter, Differentials once each week'
	   WHERE calendar_id = 5 AND name != 'Quarterly Master, Weekly Diff';
	UPDATE events SET description = 'weekly differential'
	   WHERE event_id = 33 AND description != 'weekly differential';
END IF;

-- Add the "Friday Master, Daily Diff" calendar and strategy
IF (SELECT COUNT(*) FROM bp.calendars WHERE calendar_id = 22) = 0
THEN
	-- Add strategy
	PERFORM clear_name('bp', 'strategies', 'name',
	                   'Friday Master, Daily Incr', 'strategy');
	INSERT INTO bp.strategies (strategy_id, name, description, read_only)
	 VALUES (22, 'Friday Master, Daily Diff',
	 'Master each Friday, Differentials daily during the week', true);
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (68, NULL, 22, 1, 1199127600, 3, NULL, NULL, 'Friday master');
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (69, NULL, 22, 2, 1199156400, 3, NULL, NULL, 'daily differential');
	INSERT INTO byday (event_id, day_no) VALUES (68, 5);
	INSERT INTO byday (event_id, day_no) VALUES (69, 0);
	INSERT INTO byday (event_id, day_no) VALUES (69, 1);
	INSERT INTO byday (event_id, day_no) VALUES (69, 2);
	INSERT INTO byday (event_id, day_no) VALUES (69, 3);
	INSERT INTO byday (event_id, day_no) VALUES (69, 4);

	-- Add calendar
	PERFORM clear_name('bp', 'calendars', 'name',
	                   'Friday Master, Daily Incr', 'calendar');
	INSERT INTO bp.calendars (calendar_id, name, description, family, read_only)
	 VALUES (22, 'Friday Master, Daily Diff',
	 'Master each Friday, Differentials daily during the week', 'Any', true);
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (70, 22, NULL, 1, 1199127600, 3, NULL, NULL, 'Friday master');
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (71, 22, NULL, 2, 1199156400, 3, NULL, NULL, 'daily differential');
	INSERT INTO byday (event_id, day_no) VALUES (70, 5);
	INSERT INTO byday (event_id, day_no) VALUES (71, 0);
	INSERT INTO byday (event_id, day_no) VALUES (71, 1);
	INSERT INTO byday (event_id, day_no) VALUES (71, 2);
	INSERT INTO byday (event_id, day_no) VALUES (71, 3);
	INSERT INTO byday (event_id, day_no) VALUES (71, 4);
ELSE 
	UPDATE bp.strategies SET name='Friday Master, Daily Diff',
         description = 'Master each Friday, Differentials daily during the week'
		 WHERE strategy_id = 22 AND description != 'Master each Friday, Differentials daily during the week';
	UPDATE bp.events SET description = 'daily differential' WHERE event_id = 69 AND description != 'daily differential';
	UPDATE bp.calendars SET name = 'Friday Master, Daily Diff',
         description = 'Master each Friday, Differentials daily during the week'
		 WHERE calendar_id = 22 AND description != 'Master each Friday, Differentials daily during the week';
	UPDATE bp.events SET description = 'daily differential' WHERE event_id = 71 AND description != 'daily differential';
	
END IF;

-- Add the "Friday BareMetal" calendar and strategy
IF (SELECT COUNT(*) FROM bp.calendars WHERE calendar_id = 23) = 0
THEN
	-- Add strategy
	PERFORM clear_name('bp', 'strategies', 'name',
	                   'Saturday BareMetal', 'strategy');
	INSERT INTO strategies (strategy_id, name, description, read_only)
	 VALUES (23, 'Saturday BareMetal',
	         'Hot BareMetal backup each week', true);
	INSERT INTO events (event_id, calendar_id, strategy_id, summary,
	                    dtstart, freq, "interval", bymonthday, description)
	 VALUES (72, NULL, 23, 4, 1199127600, 3, NULL, NULL, 'weekly bare metal');
	INSERT INTO byday (event_id, day_no) VALUES (72, 6);

	-- Add calendar
	PERFORM clear_name('bp', 'calendars', 'name',
	                   'Saturday BareMetal', 'calendar');
	INSERT INTO calendars (calendar_id, name, description, family, read_only)
	 VALUES (23, 'Saturday BareMetal',
	         'Hot BareMetal backup each week', 'Windows', true);
	INSERT INTO events (event_id, calendar_id, strategy_id, summary,
	                    dtstart, freq, "interval", bymonthday, description)
	 VALUES (73, 23, NULL, 4, 1199127600, 3, NULL, NULL, 'weekly bare metal');
	INSERT INTO byday (event_id, day_no) VALUES (73, 6);
ELSE
	-- Strategy and calendar names before 5.0.2 had "Bare Metal"
	PERFORM clear_name('bp', 'strategies', 'name',
	                   'Saturday Bare Metal', 'strategy');
	UPDATE bp.strategies SET name = 'Saturday BareMetal',
	 description = 'Hot BareMetal backup each week' WHERE strategy_id = 23;
	UPDATE bp.events SET description = 'weekly baremetal' WHERE event_id = 72;

	PERFORM clear_name('bp', 'calendars', 'name',
	                   'Saturday Bare Metal', 'calendar');
	UPDATE bp.calendars SET name = 'Saturday BareMetal',
	 description = 'Hot BareMetal backup each week' WHERE calendar_id = 23;
	UPDATE bp.events SET description = 'weekly baremetal' WHERE event_id = 73;
END IF;


-- Add the "Friday Master, Daily Diff, Hourly Incr" calendar and strategy
IF (SELECT COUNT(*) FROM bp.calendars WHERE calendar_id = 24) = 0
THEN
	-- Add strategy
	PERFORM clear_name('bp', 'strategies', 'name',
	                   'Friday Master, Daily Diff, Hourly Incr', 'strategy');
	INSERT INTO bp.strategies (strategy_id, name, description, read_only)
	 VALUES (24, 'Friday Master, Daily Diff, Hourly Incr',
	         'Master each Friday, Differentials daily during the week, Incrementals every hour', true);
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (74, NULL, 24, 1, 1199127600, 3, NULL, NULL, 'Friday master');
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (75, NULL, 24, 2, 1199156400, 3, NULL, NULL, 'daily differential');
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (76, NULL, 24, 17, 1199156400, 2, NULL, NULL, 'hourly incremental');

	-- Add BYDAYS events
	INSERT INTO byday (event_id, day_no) VALUES (74, 5);
	INSERT INTO byday (event_id, day_no) VALUES (75, 0);
	INSERT INTO byday (event_id, day_no) VALUES (75, 1);
	INSERT INTO byday (event_id, day_no) VALUES (75, 2);
	INSERT INTO byday (event_id, day_no) VALUES (75, 3);
	INSERT INTO byday (event_id, day_no) VALUES (75, 4);

	-- Add BYHOUR events
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 0);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 1);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 2);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 3);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 4);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 5);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 6);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 7);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 8);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 9);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 10);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 11);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 12);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 13);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 14);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 15);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 16);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 17);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 18);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 19);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 20);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 21);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 22);
	INSERT INTO byhour (event_id, hour_no) VALUES (76, 23);

	-- Add calendar events
	PERFORM clear_name('bp', 'calendars', 'name',
	                   'Friday Master, Daily Diff, Hourly Incr', 'calendar');
	INSERT INTO bp.calendars (calendar_id, name, description, family, read_only)
	 VALUES (24, 'Friday Master, Daily Diff, Hourly Incr',
	         'Master each Friday, Differentials daily during the week, Incrementals every hour', 'Any', true);
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (77, 24, NULL, 1, 1199127600, 3, NULL, NULL, 'Friday master');
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
		               dtstart, freq, "interval", bymonthday, description)
	 VALUES (78, 24, NULL, 2, 1199156400, 3, NULL, NULL, 'daily differential');
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (79, 24, NULL, 17, 1199156400, 2, NULL, NULL, 'hourly incremental');

	-- ADD BYDAY events
	INSERT INTO byday (event_id, day_no) VALUES (77, 5);
	INSERT INTO byday (event_id, day_no) VALUES (78, 0);
	INSERT INTO byday (event_id, day_no) VALUES (78, 1);
	INSERT INTO byday (event_id, day_no) VALUES (78, 2);
	INSERT INTO byday (event_id, day_no) VALUES (78, 3);
	INSERT INTO byday (event_id, day_no) VALUES (78, 4);

	-- Add BYHOUR events
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 0);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 1);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 2);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 3);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 4);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 5);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 6);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 7);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 8);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 9);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 10);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 11);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 12);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 13);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 14);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 15);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 16);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 17);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 18);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 19);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 20);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 21);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 22);
	INSERT INTO byhour (event_id, hour_no) VALUES (79, 23);
END IF;

-- Add the "Incremental Forever" calendar and strategy
IF (SELECT COUNT(*) FROM bp.calendars WHERE calendar_id = 25) = 0
THEN
	-- Add strategy
	PERFORM clear_name('bp', 'strategies', 'name',
	                   'Incremental Forever', 'strategy');
	INSERT INTO bp.strategies (strategy_id, name, description, read_only)
	 VALUES (25, 'Incremental Forever',
	         'Incremental backup every hour', true);
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (80, NULL, 25, 17, 1199156400, 2, NULL, NULL, 'incremental forever');

	-- Add BYHOUR events
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 0);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 1);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 2);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 3);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 4);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 5);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 6);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 7);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 8);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 9);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 10);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 11);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 12);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 13);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 14);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 15);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 16);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 17);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 18);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 19);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 20);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 21);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 22);
	INSERT INTO byhour (event_id, hour_no) VALUES (80, 23);

	-- Add calendar events
	PERFORM clear_name('bp', 'calendars', 'name',
	                   'Incremental Forever', 'calendar');
	INSERT INTO bp.calendars (calendar_id, name, description, family, read_only)
	 VALUES (25, 'Incremental Forever',
	         'Incremental backup every hour', 'Any', true);
	INSERT INTO bp.events (event_id, calendar_id, strategy_id, summary,
	                       dtstart, freq, "interval", bymonthday, description)
	 VALUES (81, 25, NULL, 17, 1199156400, 2, NULL, NULL, 'incremental forever');

	-- Add BYHOUR events
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 0);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 1);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 2);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 3);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 4);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 5);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 6);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 7);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 8);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 9);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 10);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 11);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 12);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 13);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 14);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 15);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 16);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 17);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 18);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 19);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 20);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 21);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 22);
	INSERT INTO byhour (event_id, hour_no) VALUES (81, 23);
END IF;

	CREATE OR REPLACE FUNCTION bp.update_cal_strat()
	RETURNS VARCHAR AS
	$UCS_BODY$
	DECLARE
		Vkeep_ids CONSTANT INT4[] := ARRAY[5,8,12,22,23];
	BEGIN

	-- If strategy id 1 exists and is read-only, then update the records
	IF (SELECT read_only = TRUE AS tf
		  FROM bp.strategies WHERE strategy_id = 1) = TRUE
	THEN
		-- Remove all strategies not in the Vkeep_ids array
		DELETE FROM bp.strategies
		 WHERE strategy_id < 50 AND NOT strategy_id = ANY (Vkeep_ids);

		-- Remove all calendars not in the Vkeep_ids array that aren't
		-- referenced.  If they are referenced, mark them as read/write.
		DELETE FROM bp.calendars
		 WHERE calendar_id < 50 AND (NOT calendar_id = ANY (Vkeep_ids))
		 AND calendar_id NOT IN
		  (SELECT calendar_id FROM bp.schedules WHERE calendar_id IS NOT NULL);
		UPDATE bp.calendars SET read_only = FALSE
		 WHERE calendar_id < 50 AND NOT calendar_id = ANY (Vkeep_ids);

		RETURN 'calendars and strategies updated successfully';
	ELSE
		RETURN 'calendars and strategies already updated';
	END IF;

	END;$UCS_BODY$
	LANGUAGE 'plpgsql' VOLATILE;

	SELECT INTO Vstr update_cal_strat();
	RAISE NOTICE '%', Vstr;
	DROP FUNCTION update_cal_strat();

-- TODO remove this before 6.0 is released
-- Remove byminute events for incremental calendars
DELETE FROM bp.byminute WHERE event_id IN (76,79,80,81);

	RETURN 0;

END;
$BODY$
LANGUAGE 'plpgsql' VOLATILE;

SELECT populate_cal_strat();
DROP FUNCTION populate_cal_strat();
--
-- End of populating calendar and strategy default values
--



--
-- Populate default selection lists
--
CREATE OR REPLACE FUNCTION populate_selection_lists()
RETURNS INT4 AS
$BODY$BEGIN

IF (SELECT COUNT(*) FROM bp.selection_lists)=0
THEN

	-- There's room for up to 100 default selection lists. This is
	-- to avoid complicated sequence adjustments if any more lists 
	-- are added, especially if the user has saved any of their own.
	ALTER SEQUENCE bp.selection_lists_selection_list_id_seq RESTART WITH 101;

	PERFORM clear_name('bp', 'selection_lists', 'name', 'Include home directories (Unix)', 'selection list');
	PERFORM clear_name('bp', 'selection_lists', 'name', 'Include home directories (Linux)', 'selection list');
	PERFORM clear_name('bp', 'selection_lists', 'name', 'Include C: volume', 'selection list');
	PERFORM clear_name('bp', 'selection_lists', 'name', 'Select .pst files', 'selection list');

	INSERT INTO selection_lists (selection_list_id, name, description, type, family, node_no, read_only, netmounts, registry, tempfiles, readmounts, allmounts) VALUES (1, 'Include home directories (Unix)', 'Include home directories on Unix clients', 'inclusion', 'Unix', NULL, false, false, false, false, false, false);
	INSERT INTO selection_list_filenames (selection_list_id, filename) VALUES (1, '@@@:/home');
	INSERT INTO selection_lists (selection_list_id, name, description, type, family, node_no, read_only, netmounts, registry, tempfiles, readmounts, allmounts) VALUES (2, 'Include home directories (Linux)', 'Include home directories on Linux clients', 'inclusion', 'Linux', NULL, false, false, false, false, false, false);
	INSERT INTO selection_list_filenames (selection_list_id, filename) VALUES (2, '@@@:/home');
	INSERT INTO selection_lists (selection_list_id, name, description, type, family, node_no, read_only, netmounts, registry, tempfiles, readmounts, allmounts) VALUES (3, 'Include C: volume', 'Include all files on C: volume', 'inclusion', 'Windows', NULL, false, false, false, false, false, false);
	INSERT INTO selection_list_filenames (selection_list_id, filename) VALUES (3, 'C:/');
	INSERT INTO selection_lists (selection_list_id, name, description, type, family, node_no, read_only, netmounts, registry, tempfiles, readmounts, allmounts) VALUES (4, 'Select .pst files', 'Select all files with extension .pst for inclusion or exclusion', 'any', 'Windows', NULL, false, false, false, false, false, false);
	INSERT INTO selection_list_filenames (selection_list_id, filename) VALUES (4, '*.pst');

END IF;

RETURN 0;

END;
$BODY$
LANGUAGE 'plpgsql' VOLATILE;

SELECT populate_selection_lists();
DROP FUNCTION populate_selection_lists();
--
-- End of populating default selection lists
--



--
-- Populate default option lists
--
CREATE OR REPLACE FUNCTION populate_option_lists()
RETURNS INT4 AS
$BODY$BEGIN

IF (SELECT COUNT(*) FROM bp.option_lists)=0
THEN

	-- There's room for up to 100 default option lists. This is
	-- to avoid complicated sequence adjustments if any more lists 
	-- are added, especially if the user has saved any of their own.
	ALTER SEQUENCE bp.option_lists_option_list_id_seq RESTART WITH 101;

	PERFORM clear_name('bp', 'option_lists', 'name', 'Default Any', 'option list');
	INSERT INTO option_lists (option_list_id, name, description, type, family, read_only, device_name, read_locking, speed_option, verify_level, depth, comment, raw_device, raw_device_command, before_command, after_command, make_catalog) VALUES (1, 'Default Any', 'Default backup options for all clients', 'backup', 'Any', true, NULL, 0, false, 1, 0, 'Default options', NULL, NULL, NULL, NULL, true);

END IF;

-- Exchange was added in a subsequent version, so use a different test
IF (SELECT TRUE FROM bp.option_lists WHERE option_list_id = 3) IS NULL
THEN

	PERFORM clear_name('bp', 'option_lists', 'name', 'Exchange Master', 'option list');
	PERFORM clear_name('bp', 'option_lists', 'name', 'Exchange Differential', 'option list');
	-- Add Exchange option lists
	INSERT INTO option_lists (option_list_id, name, description, type, family, read_only, device_name, read_locking, speed_option, verify_level, depth, comment, raw_device, raw_device_command, before_command, after_command, make_catalog) VALUES (2, 'Exchange Master', 'Exchange master backup-specific options', 'backup', 'Windows', true, NULL, 0, false, 0, 0, '', NULL, NULL, 'CNT:c:/pcbp/Exchange.dir/ExchMaster.bat', NULL, true);
	INSERT INTO option_lists (option_list_id, name, description, type, family, read_only, device_name, read_locking, speed_option, verify_level, depth, comment, raw_device, raw_device_command, before_command, after_command, make_catalog) VALUES (3, 'Exchange Differential', 'Exchange differential backup-specific options', 'backup', 'Windows', true, NULL, 0, false, 0, 0, '', NULL, NULL, 'CNT:c:/pcbp/Exchange.dir/ExchDiffer.bat', NULL, true);

END IF;

RETURN 0;

END;
$BODY$
LANGUAGE 'plpgsql' VOLATILE;

SELECT populate_option_lists();
DROP FUNCTION populate_option_lists();


--
-- End of populating default option lists
--


-- Set statistics of some tables to improve query plans.  The column
-- backup_files.end_backup_no is adjusted below, after its creation.
ALTER TABLE bp.backups ALTER COLUMN backup_no SET STATISTICS 50;
ALTER TABLE bp.backups ALTER COLUMN node_no SET STATISTICS 50;
ALTER TABLE bp.backup_files ALTER COLUMN backup_no SET STATISTICS 50;
ALTER TABLE bp.backup_files ALTER COLUMN dir_no SET STATISTICS 50;
ALTER TABLE bp.backup_files_non_ctar ALTER COLUMN backup_no SET STATISTICS 50;
ALTER TABLE bp.backup_files_non_ctar ALTER COLUMN dir_no SET STATISTICS 50;
ALTER TABLE bp.archive_files ALTER COLUMN archive_id SET STATISTICS 50;
ALTER TABLE bp.archive_files ALTER COLUMN dir_no SET STATISTICS 50;
ALTER TABLE bp.dirs ALTER COLUMN dir_no SET STATISTICS 50;
ALTER TABLE bp.dirs ALTER COLUMN dname SET STATISTICS 50;
ALTER TABLE bp.dirs_non_ctar ALTER COLUMN dir_no SET STATISTICS 50;
ALTER TABLE bp.dirs_non_ctar ALTER COLUMN dname SET STATISTICS 50;


DROP FUNCTION clear_name(sch_name VARCHAR, tbl_name VARCHAR, col_name VARCHAR, requested_name VARCHAR, canonical_obj_name VARCHAR);


--
-- Changes required for 4.2 Application-specific agents
--
-- Function to add a application_lookup row, if the row doesn't exist already
CREATE OR REPLACE FUNCTION bp.add_application_lookup(p_app_id INTEGER, p_name VARCHAR, p_type VARCHAR)
RETURNS VARCHAR AS
$BODY$BEGIN

	INSERT INTO bp.application_lookup(app_id, name, type) SELECT p_app_id, p_name, p_type
	    WHERE (SELECT TRUE FROM bp.application_lookup WHERE app_id = p_app_id) IS NULL;
	RETURN p_app_id||' - '||p_name;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

-- Add application entries
-- See bpdb.sql for details
SELECT bp.add_application_lookup(1, 'file-level', 'file-level');
SELECT bp.add_application_lookup(2, 'Exchange 2003', 'Exchange');
SELECT bp.add_application_lookup(3, 'Exchange 2007', 'Exchange');
SELECT bp.add_application_lookup(4, 'Exchange 2010', 'Exchange');
SELECT bp.add_application_lookup(5, 'Exchange 2013', 'Exchange');
SELECT bp.add_application_lookup(21, 'SQL Server 2005',    'SQL Server');
SELECT bp.add_application_lookup(22, 'SQL Server 2008',    'SQL Server');
SELECT bp.add_application_lookup(23, 'SQL Server 2008 R2', 'SQL Server');
SELECT bp.add_application_lookup(24, 'SQL Server 2012', 'SQL Server');
SELECT bp.add_application_lookup(25, 'SQL Server 2014', 'SQL Server');
SELECT bp.add_application_lookup(30, 'Archive', 'Archive');
SELECT bp.add_application_lookup(40, 'VMware', 'VMware');
SELECT bp.add_application_lookup(50, 'Hyper-V 2008 R2', 'Hyper-V');
SELECT bp.add_application_lookup(51, 'Hyper-V 2012', 'Hyper-V');
-- 7.0.0
SELECT bp.add_application_lookup(100, 'Oracle 11', 'Oracle');
SELECT bp.add_application_lookup(101, 'Oracle 12', 'Oracle');
SELECT bp.add_application_lookup(102, 'Oracle 10', 'Oracle');
SELECT bp.add_application_lookup(110, 'SharePoint 2007', 'SharePoint');
SELECT bp.add_application_lookup(111, 'SharePoint 2010', 'SharePoint');
SELECT bp.add_application_lookup(112, 'SharePoint 2013', 'SharePoint');
SELECT bp.add_application_lookup(60, 'System Metadata', 'System Metadata');
-- 7.3.0
SELECT bp.add_application_lookup(120, 'UCS Service Profile', 'UCS Service Profile');
-- 7.4.0
SELECT bp.add_application_lookup(130, 'NDMP Device', 'NDMP Device');

DROP FUNCTION bp.add_application_lookup(INTEGER, VARCHAR, VARCHAR);

-- Add client application entries for file-level backups
INSERT INTO bp.node_applications (node_no, app_id, current)
   SELECT n.node_no, 1, TRUE
     FROM bp.nodes AS n LEFT JOIN bp.node_applications AS a ON (n.node_no = a.node_no)
     WHERE a.node_no IS NULL ORDER BY n.node_no;
INSERT INTO bp.application_instances (node_no, app_id)
   SELECT n.node_no, 1
     FROM bp.nodes AS n LEFT JOIN bp.application_instances AS a ON (n.node_no = a.node_no)
     WHERE a.node_no IS NULL ORDER BY n.node_no;


CREATE OR REPLACE FUNCTION update_bp()
RETURNS INT4 AS
$BODY$BEGIN

	-- Backups now refer to application instances
	IF NOT column_exists('bp', 'backups', 'instance_id') THEN
		ALTER TABLE bp.backups ADD COLUMN instance_id INTEGER
			 REFERENCES bp.application_instances (instance_id) ON DELETE SET NULL;
		UPDATE bp.backups SET instance_id = ai.instance_id
		  FROM bp.application_instances AS ai
		  WHERE bp.backups.node_no = ai.node_no AND ai.app_id = 1;
	END IF;

	-- Jobs now refer to application instances
	IF NOT column_exists('bp', 'jobs', 'instance_id') THEN
		ALTER TABLE bp.jobs ADD COLUMN instance_id INTEGER
			 REFERENCES bp.application_instances (instance_id) ON DELETE SET NULL;
		UPDATE bp.jobs SET instance_id = ai.instance_id
		  FROM bp.application_instances AS ai
		  WHERE bp.jobs.node_no = ai.node_no AND ai.app_id = 1;
	END IF;

	-- Support jobs for application schedules
	IF NOT column_exists('bp', 'jobs', 'schedule_id') THEN
		ALTER TABLE bp.jobs ADD COLUMN schedule_id INTEGER;
	END IF;

	-- Client version is stored in the nodes table
	IF NOT column_exists('bp', 'nodes', 'version') THEN
		ALTER TABLE bp.nodes ADD COLUMN version VARCHAR DEFAULT 'Unknown';
	END IF;

	-- Store application instance discovery timestamp
	IF NOT column_exists('bp', 'application_instances', 'discovery_time') THEN
		ALTER TABLE bp.application_instances
			ADD COLUMN discovery_time BIGINT NOT NULL
			DEFAULT extract(epoch from now())::bigint;
	END IF;

	-- Update raid_drives tables
	IF NOT column_exists('bp', 'raid_drives', 'partition') THEN
		ALTER TABLE bp.raid_drives ADD COLUMN partition INTEGER NOT NULL DEFAULT 0;
	END IF;

	-- Update devices table 
	IF NOT column_exists('bp', 'devices', 'is_deleted') THEN
		ALTER TABLE bp.devices ADD COLUMN is_deleted BOOLEAN NOT NULL DEFAULT FALSE;
	END IF;
	IF NOT column_exists('bp', 'devices', 'dedup') THEN
		ALTER TABLE bp.devices ADD COLUMN dedup BOOLEAN NOT NULL DEFAULT TRUE;
	END IF;
	IF NOT column_exists('bp', 'devices', 'storage_id') THEN
		ALTER TABLE bp.devices ADD COLUMN storage_id INTEGER NOT NULL DEFAULT 1 REFERENCES storage(storage_id);
	END IF;

	IF NOT column_exists('bp', 'hash_values', 'sis_offset') THEN
		ALTER TABLE bp.hash_values ADD COLUMN sis_offset BIGINT DEFAULT NULL;
	END IF;

	--IF NOT index_exists('bp', 'deduplicated_backups_hash_values_index') THEN
	--	CREATE INDEX deduplicated_backups_hash_values_index ON bp.backups_hash_values(deduplicated) WHERE deduplicated IS NULL;
	--	ALTER TABLE bp.backups_hash_values ALTER hash_id SET STATISTICS 1000;
	--END IF;
	
	IF NOT index_exists('bp', 'backups_hash_values_hash_id_index') THEN
		CREATE INDEX backups_hash_values_hash_id_index ON bp.backups_hash_values (hash_id);
	END IF;

	--IF NOT index_exists('bp', 'backup_files_backup_no_index') THEN
	--	CREATE INDEX backup_files_backup_no_index ON bp.backup_files (backup_no, end_backup_no);
	--END IF;

	IF NOT index_exists('bp', 'hash_values_sis_offset_index') THEN
		CREATE INDEX hash_values_sis_offset_index ON bp.hash_values (sis_offset);
	END IF;

	IF NOT index_exists('bp', 'dirs_dname_lcase') THEN
		CREATE INDEX dirs_dname_lcase ON bp.dirs(lower(dname));
	END IF;

	IF NOT index_exists('bp', 'dirs_non_ctar_dname_lcase') THEN
		CREATE INDEX dirs_non_ctar_dname_lcase ON bp.dirs_non_ctar(lower(dname));
	END IF;

	-- Add default SNMP manager, notifications.unitrends.com
	INSERT INTO bp.snmp_config (destination, enabled)
		SELECT 'notifications.unitrends.com', TRUE
		WHERE NOT EXISTS (SELECT * FROM bp.snmp_config
			WHERE destination = 'notifications.unitrends.com');

	IF column_exists('bp', 'archive_profiles', 'is_legacy') THEN
		ALTER TABLE bp.archive_profiles ALTER COLUMN is_legacy DROP DEFAULT;
		ALTER TABLE bp.archive_profiles ALTER COLUMN is_legacy TYPE INTEGER
			USING CASE WHEN is_legacy=FALSE THEN 1 ELSE 0 END;
		ALTER TABLE bp.archive_profiles RENAME COLUMN is_legacy TO revision;
	END IF;

	IF NOT column_exists('bp', 'archive_profiles', 'purge') THEN
		ALTER TABLE bp.archive_profiles ADD COLUMN purge BOOLEAN NOT NULL DEFAULT FALSE;
	END IF;

	-- In 5.0.2, backup_files became temporal via the end_backup_no column
	IF NOT column_exists('bp', 'backup_files', 'end_backup_no') THEN
		DROP INDEX backup_files_backup_dir_idx;
		ALTER TABLE bp.backup_files DISABLE TRIGGER ALL;
		ALTER TABLE bp.backup_files ADD COLUMN end_backup_no INTEGER NOT NULL DEFAULT 0;
		CREATE INDEX backup_files_backup_dir_idx ON bp.backup_files(dir_no, backup_no, end_backup_no);
		ALTER TABLE bp.backup_files ENABLE TRIGGER ALL;
	END IF;
	-- Improve query plan by raising statistics level; most columns are done
	-- above, but this column may not exist until this point in the script.
	ALTER TABLE bp.backup_files ALTER COLUMN end_backup_no SET STATISTICS 50;

	-- If backup_files is temporal, backup_files_non_ctar should be, too
	IF NOT column_exists('bp', 'backup_files_non_ctar', 'end_backup_no') THEN
		DROP INDEX backup_files_non_ctar_backup_dir_idx;
		ALTER TABLE bp.backup_files_non_ctar DISABLE TRIGGER ALL;
		ALTER TABLE bp.backup_files_non_ctar ADD COLUMN end_backup_no INTEGER NOT NULL DEFAULT 0;
		CREATE INDEX backup_files_non_ctar_backup_dir_idx ON bp.backup_files_non_ctar(dir_no, backup_no, end_backup_no);
		ALTER TABLE bp.backup_files_non_ctar ENABLE TRIGGER ALL;
	END IF;
	-- Improve query plan by raising statistics level; most columns are done
	-- above, but this column may not exist until this point in the script.
	ALTER TABLE bp.backup_files_non_ctar ALTER COLUMN end_backup_no SET STATISTICS 50;

	-- Update vmware schedule references to allow vCenter association.  In 6.0,
	-- this is migrated away but the references need fixing before migration.
	IF EXISTS(SELECT TRUE FROM pg_catalog.pg_constraint WHERE conname = 'vmware_schedules_esx_uuid_fkey') THEN
		-- Allow esx_uuid to refer to vCenter servers
		ALTER TABLE bp.vmware_schedules DROP CONSTRAINT vmware_schedules_esx_uuid_fkey;
--		UPDATE bp.vmware_schedules SET esx_uuid =
--		 (SELECT COALESCE(vcenter_uuid, esx_uuid)
--		 FROM vmware_esx_servers LEFT JOIN vmware_vcenter_assoc USING (esx_uuid)
--		 WHERE esx_uuid = vmware_schedules.esx_uuid);

		ALTER TABLE application_schedules DISABLE TRIGGER ALL;
		ALTER TABLE vmware_schedules DISABLE TRIGGER ALL;
		-- Make schedules reference vCenter servers
		UPDATE bp.vmware_schedules
			SET esx_uuid = COALESCE(vva.vcenter_uuid, vmware_schedules.esx_uuid)
			FROM vmware_vcenter_assoc vva
			WHERE vva.esx_uuid = vmware_schedules.esx_uuid;
		-- Make application_instances.key1 be the vCenter name
		UPDATE bp.application_instances SET key1 = vcenter.name
			FROM bp.vmware_vms vm, bp.vmware_vcenter_assoc assoc,
			  bp.vmware_vcenter_servers vcenter
			WHERE application_instances.instance_id = vm.instance_id
			  AND vm.esx_uuid = assoc.esx_uuid
			  AND assoc.vcenter_uuid = vcenter.vcenter_uuid;
		-- Turn off policy flag for all but one schedule per vCenter server
		UPDATE bp.application_schedules
			SET add_new_instances = FALSE
			WHERE add_new_instances = TRUE
			AND application_schedule_id IN
			 (SELECT application_schedule_id
			  FROM bp.application_schedules
			  JOIN bp.vmware_schedules USING (application_schedule_id))
			AND application_schedule_id NOT IN
			 (SELECT MIN(application_schedule_id)
			  FROM bp.application_schedules
			  JOIN bp.vmware_schedules USING (application_schedule_id)
			  GROUP BY esx_uuid);
		ALTER TABLE application_schedules ENABLE TRIGGER ALL;
		ALTER TABLE vmware_schedules ENABLE TRIGGER ALL;
	END IF;

	-- In 6.0, add fields for archiving support of tapes and autochangers
	IF NOT column_exists('bp', 'archive_profiles', 'target_slots') THEN
		ALTER TABLE bp.archive_profiles
			ADD COLUMN target_slots VARCHAR NOT NULL DEFAULT '';
	END IF;
	IF NOT column_exists('bp', 'archive_disks', 'start_mark') THEN
		ALTER TABLE bp.archive_disks
			ADD COLUMN start_mark INTEGER NOT NULL DEFAULT 0;
	END IF;
	IF NOT column_exists('bp', 'archive_disks', 'first_archive') THEN
		ALTER TABLE bp.archive_disks
			ADD COLUMN first_archive INTEGER NOT NULL DEFAULT -1;
	END IF;

	-- 6.0 correction for primary key correction.
	-- 6.1: adjust implementation so it only happens at most once.
	IF index_exists('bp', 'nas_pkey') THEN
		ALTER TABLE bp.nas DROP CONSTRAINT nas_pkey;
		ALTER TABLE bp.nas ADD CONSTRAINT nas_pkey2 PRIMARY KEY (address, port, protocol, share_name);
	END IF;
	IF index_exists('bp', 'fc_pkey') THEN
		ALTER TABLE bp.fc DROP CONSTRAINT fc_pkey;
		ALTER TABLE bp.fc ADD CONSTRAINT fc_pkey2 PRIMARY KEY (wwpn, lun);
	END IF;

	-- 6.1.0: Support vaulting of individual application instances.
	IF NOT column_exists('bp', 'application_instances', 'sync') THEN
		ALTER TABLE bp.application_instances 
			ADD COLUMN sync BOOLEAN NOT NULL DEFAULT FALSE;
		ALTER TABLE bp.application_instances 
			ADD COLUMN priority INTEGER NOT NULL DEFAULT 500;
		-- Migrate the priority and sync values from nodes (we don't need
		-- an EXISTS clause since all records in application_instances have
		-- a corresponding row in nodes).
		UPDATE bp.application_instances SET priority = 
			(SELECT bp.nodes.priority FROM bp.nodes
			 WHERE bp.nodes.node_no = bp.application_instances.node_no);
		UPDATE bp.application_instances SET sync = true FROM bp.nodes
			 WHERE bp.nodes.node_no = bp.application_instances.node_no
			 AND (bp.nodes.properties & 1) = 1
			 AND bp.application_instances.current = true;
		INSERT INTO bp.operation_status
			(operation_id, subsystem, description, status)
			SELECT 4, 'AppVaulting',
			'Modify sync for app instances dependent on Securesync settings',
			'pending'
			WHERE NOT EXISTS (SELECT TRUE FROM bp.operation_status
				WHERE operation_id = 4);
	END IF;

	-- 6.2.0 intermediate between Beta & Final release, add WVF app support
	-- Add wvf to application_instances also affects fresh upgrade from pre-6.2
	IF NOT column_exists('bp', 'application_instances', 'wvf') THEN
		ALTER TABLE bp.application_instances
			ADD COLUMN wvf BOOLEAN NOT NULL DEFAULT FALSE; 
	END IF;
	-- This migrates Beta 6.2.0 to Final for WVF
	IF column_exists('bp', 'virtual_clients', 'last_backup_no') THEN
		UPDATE bp.application_instances SET wvf = true FROM bp.virtual_clients
			WHERE bp.virtual_clients.node_no = bp.application_instances.node_no
			AND bp.application_instances.app_id = 1;
		INSERT INTO bp.virtual_clients_appinst
			(virtual_id, instance_id, last_backup_no, last_backup_time)
	    	SELECT virtual_id, instance_id, last_backup_no, last_backup_time
		    FROM bp.virtual_clients JOIN bp.application_instances
			USING(node_no) WHERE app_id=1;
		ALTER TABLE bp.virtual_clients DROP COLUMN last_backup_no;
		ALTER TABLE bp.virtual_clients DROP COLUMN last_backup_time;
		ALTER TABLE bp.virtual_clients
			ADD COLUMN rdu BOOLEAN NOT NULL DEFAULT TRUE;
	END IF;

	-- 6.3.0
	IF NOT column_exists('bp', 'vmware_esx_servers', 'num_cpu_sockets') THEN
		ALTER TABLE bp.vmware_esx_servers
			ADD COLUMN num_cpu_sockets INTEGER NOT NULL DEFAULT 1;
	END IF;
	IF NOT column_exists('bp', 'nodes', 'num_cpu_sockets') THEN
		ALTER TABLE bp.nodes
			ADD COLUMN num_cpu_sockets INTEGER NOT NULL DEFAULT 1;
	END IF;
	IF NOT column_exists('bp', 'internal', 'is_lvm') THEN
		ALTER TABLE bp.internal
			ADD COLUMN is_lvm BOOLEAN NOT NULL DEFAULT FALSE;
	END IF;
	IF NOT column_exists('bp', 'nodes', 'last_update') THEN
		ALTER TABLE bp.nodes ADD COLUMN last_update BIGINT DEFAULT 0;
	END IF;
	IF NOT column_exists('bp', 'nodes', 'credential_id') THEN
		ALTER TABLE bp.nodes ADD COLUMN credential_id INTEGER DEFAULT NULL
		REFERENCES client_credentials (credential_id) ON DELETE SET NULL;
	END IF;

	-- 7.0.0
	IF NOT column_exists('bp','client_credentials','display_name') THEN
		ALTER TABLE client_credentials
			ADD COLUMN display_name VARCHAR DEFAULT NULL;
	END IF;

	IF NOT column_exists('bp','application_instances','credential_id') THEN
		ALTER TABLE application_instances
			ADD COLUMN credential_id INTEGER DEFAULT NULL
			REFERENCES client_credentials (credential_id) ON DELETE SET NULL;
	END IF;

	IF NOT column_exists('bp','application_instances','app_aware') THEN
		ALTER TABLE application_instances
			ADD COLUMN app_aware BOOLEAN DEFAULT NULL;
	END IF;

	IF NOT column_exists('bp','application_instances','identity') THEN
		ALTER TABLE application_instances
			ADD COLUMN identity VARCHAR DEFAULT NULL;
	END IF;

	IF NOT column_exists('bp','systems','identity') THEN
		ALTER TABLE systems ADD COLUMN identity VARCHAR UNIQUE DEFAULT NULL;
	END IF;

	-- No ON DELETE action, enforce cleanup of grandchild backups
	IF NOT column_exists('bp','nodes','system_id') THEN
		ALTER TABLE nodes ADD COLUMN system_id INTEGER DEFAULT NULL
		REFERENCES systems(system_id);
	END IF;

	IF NOT column_exists('bp','nodes','gcname') THEN
		ALTER TABLE nodes ADD COLUMN gcname VARCHAR DEFAULT NULL;
	END IF;

	IF NOT column_exists('bp','users','vault_user') THEN
		ALTER TABLE users ADD COLUMN vault_user BOOLEAN DEFAULT FALSE;
	END IF;

	IF NOT column_exists('bp', 'archives', 'deduped') THEN
		ALTER TABLE bp.archives ADD COLUMN deduped BOOLEAN NOT NULL DEFAULT FALSE;
	END IF;

	IF NOT column_exists('bp', 'archive_restore_specs', 'gc_mode') THEN
		ALTER TABLE bp.archive_restore_specs ADD COLUMN gc_mode BOOLEAN NOT NULL DEFAULT FALSE;
	END IF;

	IF NOT column_exists('bp', 'archive_profiles', 'dedup') THEN
		ALTER TABLE bp.archive_profiles ADD COLUMN dedup BOOLEAN NOT NULL DEFAULT FALSE;
	END IF;

	IF NOT column_exists('bp', 'archive_profiles', 'fastseed') THEN
		ALTER TABLE bp.archive_profiles ADD COLUMN fastseed BOOLEAN NOT NULL DEFAULT FALSE;
	END IF;

	IF NOT column_exists('bp','sds_stats','destination') THEN
		ALTER TABLE sds_stats ADD COLUMN destination VARCHAR DEFAULT NULL;
	END IF;

	IF (SELECT attnotnull
			FROM pg_catalog.pg_class c, pg_catalog.pg_namespace n, pg_catalog.pg_attribute a
			WHERE a.attrelid = c.oid AND c.relnamespace = n.oid AND lower(n.nspname) = 'bp'
			AND lower(c.relname) = 'replication_queue' AND lower(a.attname) = 'vault') = FALSE
	THEN
		DELETE FROM bp.replication_queue WHERE vault IS NULL;
		ALTER TABLE bp.replication_queue ALTER COLUMN vault SET NOT NULL;
	END IF;

	-- app_aware in application_instances depends on credential_id in client_credentials
	CREATE OR REPLACE FUNCTION client_credentials_turn_off_app_aware() RETURNS TRIGGER AS $client_creds_app_aware_tag$
	BEGIN
		-- If credential_id is set to NULL, all referring app_aware instances are false
		UPDATE bp.application_instances SET app_aware = FALSE
		WHERE bp.application_instances.credential_id = OLD.credential_id;
		RETURN OLD;
	END
	$client_creds_app_aware_tag$ LANGUAGE 'plpgsql';

	DROP FUNCTION client_credentials_turn_off_app_aware() cascade;
	-- I don't think if exists works, hence the create/drop/cascade
	CREATE OR REPLACE FUNCTION client_credentials_turn_off_app_aware() RETURNS TRIGGER AS $client_creds_app_aware_tag$
	BEGIN
		-- If credential_id is set to NULL, all referring app_aware instances are false
		UPDATE bp.application_instances SET app_aware = FALSE
		WHERE bp.application_instances.credential_id = OLD.credential_id;
		RETURN OLD;
	END
	$client_creds_app_aware_tag$ LANGUAGE 'plpgsql';

	CREATE TRIGGER client_credentials_turn_off_app_aware BEFORE DELETE ON client_credentials
		FOR EACH ROW EXECUTE PROCEDURE client_credentials_turn_off_app_aware();
	-- end app_aware

	-- 7.0.0-2 patch
	IF NOT EXISTS(SELECT TRUE FROM pg_catalog.pg_constraint WHERE 
			conname = 'application_instances_identity_unique') THEN
		-- The 2 simple conditions require no cleanup before applying the
		-- constraint: [1] no grandclients, or [2] no duplicates.
		-- If cleanup is required, it will be handled later in postinstall.
		IF (SELECT COUNT(node_no) FROM bp.nodes WHERE (properties & 32)=32) = 0
		THEN
			ALTER TABLE bp.application_instances ADD CONSTRAINT
			application_instances_identity_unique UNIQUE(identity);
		ELSIF (SELECT COUNT(node_no) FROM bp.nodes
				JOIN bp.application_instances USING(node_no)
				WHERE (properties & 32)=32 GROUP BY identity
				ORDER BY count DESC LIMIT 1) <= 1 THEN
			ALTER TABLE bp.application_instances ADD CONSTRAINT
			application_instances_identity_unique UNIQUE(identity);
		END IF;
	END IF; -- end of checking unique constraint on identity

	-- 7.1.0
	IF NOT column_exists('bp', 'node_applications', 'credential_id') THEN
		ALTER TABLE bp.node_applications ADD COLUMN credential_id
		INTEGER DEFAULT NULL REFERENCES client_credentials (credential_id)
		ON DELETE SET NULL;
	END IF;
	
	-- 7.2.0
	IF NOT column_exists('bp' , 'replication_queue' , 'rep_start') THEN
		ALTER TABLE replication_queue 
			ADD COLUMN rep_start BIGINT DEFAULT NULL;
	END IF;

	IF NOT column_exists('bp', 'vmware_vms', 'parent_type') THEN
		ALTER TABLE bp.vmware_vms ADD COLUMN parent_type INTEGER NOT NULL DEFAULT 0;
		ALTER TABLE bp.vmware_vms ADD COLUMN parent_key VARCHAR NOT NULL DEFAULT '';
	END IF;

	IF NOT column_exists('bp','systems','device_no') THEN
		ALTER TABLE bp.systems ADD COLUMN device_no
		INTEGER DEFAULT NULL REFERENCES devices (device_no);
	END IF;

	IF NOT column_exists('bp','vmware_vm_disks','disk_mode') THEN
		ALTER TABLE bp.vmware_vm_disks ADD COLUMN disk_mode
		INTEGER DEFAULT 0;
	END IF;

	IF NOT column_exists('bp', 'notifications', 'log') THEN
		ALTER TABLE notifications ADD COLUMN log BOOLEAN NOT NULL DEFAULT false;
	END IF;

	IF NOT column_exists('bp', 'notifications', 'psa') THEN
		ALTER TABLE notifications ADD COLUMN psa BOOLEAN NOT NULL DEFAULT false;
	END IF;
	
	IF NOT column_exists('bp', 'retention_limits', 'legal_hold') THEN
		ALTER TABLE retention_limits ADD COLUMN legal_hold INTEGER NOT NULL DEFAULT 0;
	END IF;
	
	IF NOT column_exists('bp', 'backups', 'legal_hold') THEN
		ALTER TABLE backups ADD COLUMN legal_hold INTEGER NOT NULL DEFAULT 0;
	END IF;
	
	-- 7.3.0
	IF NOT column_exists('bp', 'nodes', 'generic_property') THEN
		ALTER TABLE nodes ADD COLUMN generic_property INTEGER DEFAULT 0;
	END IF;

	IF NOT column_exists('bp', 'vmware_vms', 'is_template') THEN
		ALTER TABLE bp.vmware_vms ADD COLUMN is_template BOOLEAN NOT NULL DEFAULT FALSE;
	END IF;

	IF NOT column_exists('bp', 'archives', 'vmware_template') THEN
		ALTER TABLE bp.archives ADD COLUMN vmware_template BOOLEAN NOT NULL DEFAULT FALSE;
	END IF;

	-- 7.4.0
	IF NOT column_exists('bp', 'virtual_clients', 'vm_name') THEN
		ALTER TABLE bp.virtual_clients ADD COLUMN vm_name VARCHAR;
	END IF;
	IF NOT column_exists('bp', 'virtual_clients', 'vmware_host') THEN
		ALTER TABLE bp.virtual_clients ADD COLUMN vmware_host INTEGER
		REFERENCES bp.vmware_connections(connection_id) ON DELETE RESTRICT;
	END IF;
	IF NOT column_exists('bp', 'virtual_clients', 'hyperv_host') THEN
		ALTER TABLE bp.virtual_clients ADD COLUMN hyperv_host INTEGER
		REFERENCES bp.nodes(node_no) ON DELETE RESTRICT;
	END IF;
	IF NOT column_exists('bp', 'virtual_clients', 'vm_guid') THEN
		ALTER TABLE bp.virtual_clients ADD COLUMN vm_guid VARCHAR UNIQUE;
	END IF;
	IF NOT column_exists('bp', 'virtual_clients', 'ip_addr') THEN
		ALTER TABLE bp.virtual_clients ADD COLUMN ip_addr VARCHAR UNIQUE;
	END IF;
	IF NOT column_exists('bp', 'backups', 'certified') THEN
		ALTER TABLE bp.backups ADD COLUMN certified INTEGER DEFAULT NULL;
	END IF;
        IF NOT column_exists('bp', 'archives', 'orig_backup_prop') THEN
                ALTER TABLE bp.archives ADD COLUMN orig_backup_prop BIGINT NOT NULL DEFAULT 0;
        END IF;

	-- start cloudhook bp.nas columns for 7.5.0
	IF NOT column_exists('bp', 'nas', 'max_size') THEN
		ALTER TABLE bp.nas ADD COLUMN max_size INTEGER DEFAULT 0;
	END IF;
	IF NOT column_exists('bp', 'nas', 'compressed') THEN
		ALTER TABLE bp.nas ADD COLUMN compressed INTEGER DEFAULT 0;
	END IF;
	IF NOT column_exists('bp', 'nas', 'fs_passwd') THEN
		ALTER TABLE bp.nas ADD COLUMN fs_passwd VARCHAR;
	END IF;
	IF NOT column_exists('bp', 'nas', 'prefix') THEN
		ALTER TABLE bp.nas ADD COLUMN prefix VARCHAR;
	END IF;
	-- end cloudhook bp.nas columns

        -- start bp.archive_tapes columns for 7.5.0
        IF NOT column_exists('bp', 'archive_tapes', 'cap_status') THEN
                ALTER TABLE bp.archive_tapes ADD COLUMN cap_status INTEGER;
        END IF;
        IF NOT column_exists('bp', 'archive_tapes', 'mb_written') THEN
                ALTER TABLE bp.archive_tapes ADD COLUMN mb_written INTEGER;
        END IF;
        IF NOT column_exists('bp', 'archive_tapes', 'is_prepared') THEN
                ALTER TABLE bp.archive_tapes ADD COLUMN is_prepared INTEGER NOT NULL DEFAULT 0;
        END IF;
        -- end bp.archive_tapes columns

	-- 8.0.0
		IF NOT column_exists('bp', 'schedules', 'vm_regex') THEN
			ALTER TABLE bp.schedules ADD COLUMN vm_regex VARCHAR DEFAULT NULL;
		END IF;

	-- start rest api storage columns for 8.1
		IF NOT column_exists('bp', 'storage', 'mb_size') THEN
			ALTER TABLE bp.storage ADD COLUMN mb_size BIGINT DEFAULT 0;
		END IF;
		IF NOT column_exists('bp', 'storage', 'mb_free') THEN
			ALTER TABLE bp.storage ADD COLUMN mb_free BIGINT DEFAULT 0;
		END IF;
		IF NOT column_exists('bp', 'storage', 'mb_used') THEN
			ALTER TABLE bp.storage ADD COLUMN mb_used BIGINT DEFAULT 0;
		END IF;
		IF NOT column_exists('bp', 'storage', 'mb_to_purge') THEN
			ALTER TABLE bp.storage ADD COLUMN mb_to_purge BIGINT DEFAULT 0;
		END IF;
		IF NOT column_exists('bp', 'storage', 'is_deleted') THEN
			ALTER TABLE bp.storage ADD COLUMN is_deleted BOOLEAN DEFAULT FALSE;
		END IF;
		IF NOT column_exists('bp', 'storage', 'is_purging') THEN
			ALTER TABLE bp.storage ADD COLUMN is_purging BOOLEAN DEFAULT FALSE;
		END IF;
		IF NOT column_exists('bp', 'storage', 'is_default') THEN
			ALTER TABLE bp.storage ADD COLUMN is_default BOOLEAN DEFAULT FALSE;
		END IF;
		IF NOT column_exists('bp', 'storage', 'stateless_storage_type') THEN
			ALTER TABLE bp.storage ADD COLUMN stateless_storage_type VARCHAR DEFAULT NULL;
		END IF;
		IF column_exists('bp', 'storage', 'storage_usage') THEN
			UPDATE bp.storage SET storage_usage='backup' WHERE storage_usage IS NULL AND storage_name='Internal';
		END IF;
	-- end rest api storage columns

	-- Other 8.1.0 modifications
	IF column_exists('bp', 'sync_summary', 'protected_assets') THEN
		ALTER TABLE bp.sync_summary RENAME COLUMN protected_assets TO registered_assets;
	END IF;
	CREATE OR REPLACE RULE insert_not_allowed AS ON INSERT TO bp.inventory_sync
		DO INSTEAD NOTHING;
	IF NOT column_exists('bp', 'summary_counts', 'rpo_days') THEN
		ALTER TABLE bp.summary_counts ADD COLUMN rpo_days INTEGER DEFAULT 7;
	END IF;
	IF NOT column_exists('bp', 'summary_counts', 'rpo_protected') THEN
		ALTER TABLE bp.summary_counts ADD COLUMN rpo_protected INTEGER DEFAULT 0;
	END IF;
	CREATE OR REPLACE RULE no_insert_summary_counts AS
		ON INSERT TO bp.summary_counts DO INSTEAD NOTHING;
	IF NOT column_exists('bp', 'forum_users', 'salesforce_id') THEN
		ALTER TABLE bp.forum_users ADD COLUMN salesforce_id VARCHAR NOT NULL;
	END IF;

	 -- NDMP
	IF column_exists('bp', 'ndmp_nas', 'node_no') THEN
		IF (select count(1) from bp.modification_time where table_name = 'ndmp_nas')=0
		THEN
			Insert into bp.modification_time (table_name, replicate) values ('ndmp_nas', 't');
			ALTER TABLE bp.ndmp_nas DISABLE TRIGGER ALL;
			CREATE TRIGGER update_modification_time_for_update_trig AFTER UPDATE ON bp.ndmp_nas FOR EACH ROW EXECUTE PROCEDURE update_modification_time();
			CREATE TRIGGER update_modification_time_for_insert_trig AFTER INSERT OR DELETE ON bp.ndmp_nas FOR EACH STATEMENT EXECUTE PROCEDURE update_modification_time();
			ALTER TABLE bp.ndmp_nas ENABLE TRIGGER ALL;
		END IF;
	END IF;
        -- END NDMP

	-- More 8.1.0 changes
	IF NOT column_exists('bp', 'vmware_vms', 'vm_instance_uuid') THEN
		ALTER TABLE bp.vmware_vms ADD COLUMN vm_instance_uuid VARCHAR UNIQUE;
	END IF;

	IF index_exists('bp', 'vmware_vms_pkey') THEN
		ALTER TABLE bp.vmware_vms DROP CONSTRAINT vmware_vms_pkey;
		ALTER TABLE bp.vmware_vms ADD CONSTRAINT vmware_vms_pkey2 UNIQUE (vm_uuid, vm_instance_uuid);
	END IF;

	RETURN 0;

END;
$BODY$
LANGUAGE 'plpgsql' VOLATILE;

SELECT update_bp();
DROP FUNCTION update_bp();

-- For appliance agent backup, bypass the trigger check
CREATE OR REPLACE FUNCTION bp.archives_instance_in_archive_set()
	RETURNS TRIGGER AS $app_sched_tag$
DECLARE
	exists BOOLEAN := FALSE;
BEGIN
	IF (NEW.archive_instance_id IS NULL AND NEW.orig_backup_no IS NULL) THEN
		RETURN NEW;
	END IF;
	IF (NEW.backtype = 1038) THEN
		RETURN NEW;
	END IF;

	-- Check instance assignments
	SELECT INTO exists TRUE
	  FROM bp.archive_profile_instances api
	  JOIN bp.archive_sets USING (profile_id)
	  WHERE archive_sets.archive_set_id = NEW.archive_set_id
	    AND api.archive_instance_id = NEW.archive_instance_id
	  LIMIT 1;

	IF (exists = TRUE) THEN
		RETURN NEW;
	END IF;

	-- Check client assignments
	SELECT INTO exists TRUE
	  FROM bp.archive_profile_clients
	  JOIN bp.archive_sets USING (profile_id)
	  JOIN bp.nodes USING (node_no)
	  JOIN bp.archive_instances ai ON (nodes.node_name = ai.client_name)
	  WHERE ai.instance_id IS NOT NULL
	    AND archive_sets.archive_set_id = NEW.archive_set_id
	    AND ai.archive_instance_id = NEW.archive_instance_id
	  LIMIT 1;

	IF (exists = TRUE) THEN
		RETURN NEW;
	ELSE
		RAISE EXCEPTION 'Archive instance % must be in archive set %',
			NEW.archive_instance_id, NEW.archive_set_id;
		RETURN NULL;
	END IF;
END
$app_sched_tag$ LANGUAGE 'plpgsql';

-- drop lookups table for 6.0.1 and transfer values to sequences
CREATE OR REPLACE FUNCTION update_bp()
RETURNS INT4 AS
$BODY$
	DECLARE result BOOLEAN;
BEGIN
	IF table_exists('bp', 'lookups') THEN
		select restart_sequence('nodes_node_no_seq', 'SELECT misc1 FROM lookups WHERE lookup_type=4 and lookup_no=3') into result;
		select restart_sequence('filesets_fileset_no_seq', 'SELECT misc1 FROM lookups WHERE lookup_type=4 and lookup_no=4') into result;
		select restart_sequence('tapes_tape_no_seq', 'SELECT misc1 FROM lookups WHERE lookup_type=4 and lookup_no=5') into result;
		select restart_sequence('jobs_job_no_seq', 'SELECT misc1 FROM lookups WHERE lookup_type=4 and lookup_no=7') into result;
		select restart_sequence('backups_backup_no_seq', 'SELECT misc1 FROM lookups WHERE lookup_type=4 and lookup_no=8') into result;
		select restart_sequence('lcs_lcs_no_seq', 'SELECT misc1 FROM lookups WHERE lookup_type=4 and lookup_no=10') into result;
		DROP TABLE lookups;
	END IF;

	RETURN 0;
END;
$BODY$
LANGUAGE 'plpgsql' VOLATILE;

SELECT update_bp();
DROP FUNCTION update_bp();

--
-- Create view of restorable backups
--
CREATE OR REPLACE VIEW bp.restorable_backups AS
	SELECT * FROM backups
	  WHERE (properties & 1) = 0  -- not synthetic
	  AND tape_no != 4294967295   -- not logically deleted
	  AND (status & 64) > 0;      -- complete

--
-- Create view of successful backups (used for purging, etc)
--
CREATE OR REPLACE VIEW bp.successful_backups AS
	SELECT * FROM backups
	  WHERE (properties & 1) = 0  -- not synthetic
	  AND tape_no != 4294967295   -- not logically deleted
	  AND (status & 64) > 0       -- job_final
	  AND (status & 36888) = 0;   -- not failed/cancelled/proc_died

--
-- Create "small backups" view for diagnostic querying of the backups
-- table; one row should fit within a 1280-pixel-wide viewing window.
--
CREATE OR REPLACE VIEW bp.sm_backups AS
	SELECT b.backup_no, n.node_name, n.node_no, b.instance_id, b.type,
	  to_timestamp(b.start_time) AS time, b.status, b.properties,
	  b.total_netpaths AS files, b.total_megs AS mb_size,
	  CASE WHEN b.tape_no = 4294967295 THEN '(deleted)'
	    ELSE COALESCE(d.dev_name, '(unknown)') END AS device
	  FROM backups b
	  JOIN nodes n USING (node_no)
	  LEFT JOIN filesets f ON (f.fileset_no = b.fileset_no)
	  LEFT JOIN tapes t ON (t.tape_no = f.tape_no)
	  LEFT JOIN devices d ON (d.dev_nr_name = t.location);

-- 7.2 Function to add psa_tools row, if the row doesn't already exist
CREATE OR REPLACE FUNCTION bp.add_psa_tool(tool_id INTEGER, tool_name VARCHAR) 
RETURNS VARCHAR AS
$BODY$BEGIN

	INSERT INTO bp.psa_tools(psa_tool_id, psa_tool_name)
	SELECT tool_id, tool_name WHERE
	(SELECT TRUE FROM bp.psa_tools WHERE psa_tool_id = tool_id) IS NULL;
	RETURN tool_id||' - '||tool_name;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

SELECT bp.add_psa_tool(1, 'ConnectWise');
DROP FUNCTION bp.add_psa_tool(INTEGER, VARCHAR);

-- 7.6 Function to add nvp user preference default rows, if the row doesn't already exist
-- Note: the factory default value must match _FACT_DEFAULT_ defined in bpl.h
CREATE OR REPLACE FUNCTION bp.add_nvp_fact_defaults(nvptype VARCHAR, nvpname VARCHAR, itemname VARCHAR, itemval VARCHAR)
RETURNS VARCHAR AS
$BODY$
DECLARE old_val VARCHAR := NULL;

BEGIN
  SELECT INTO old_val item_value FROM bp.nvp WHERE nvp_type = nvptype AND nvp_name = nvpname AND item_name = itemname;
  IF old_val IS NULL
  THEN
	INSERT INTO bp.nvp (nvp_type, nvp_name, item_name, item_value) VALUES (nvptype, nvpname, itemname, itemval);
  ELSEIF old_val <> itemval
  THEN
	UPDATE bp.nvp SET item_value = itemval  WHERE nvp_type = nvptype AND nvp_name = nvpname AND item_name = itemname;
  END IF;

  RETURN nvptype;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'dash_active_job_refresh','10');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'dash_alert_refresh','10');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'dash_backup_refresh','300');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'dash_replication_refresh','60');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'dash_restore_refresh','10');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'dash_storage_refresh','10');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'dash_forum_posts_refresh','60');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'dash_daily_feed_refresh','60');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'page_protect_refresh','10');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'page_restore_refresh','10');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'page_jobs_refresh','10');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'language','en');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'show_setup_wizard','true');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'show_help_overlay','true');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'show_eula','true');
SELECT bp.add_nvp_fact_defaults('user_preference',' !!-factory-default!!', 'dashboard_settings','');
DROP FUNCTION bp.add_nvp_fact_defaults(VARCHAR,VARCHAR,VARCHAR, VARCHAR);

--
-- Drop functions used in table modification, RUN THIS LAST!
--
DROP FUNCTION public.table_exists(sch_name VARCHAR,tbl_name VARCHAR);
DROP FUNCTION public.index_exists(sch_name VARCHAR,idx_name VARCHAR);
DROP FUNCTION public.column_exists(sch_name VARCHAR,tbl_name VARCHAR,col_name VARCHAR);

-- Function to add a application_lookup row, if the row doesn't exist already
CREATE OR REPLACE FUNCTION bp.schedule_vm_container_lookup(p_container_id INTEGER, p_hypervisor VARCHAR, p_name VARCHAR)
RETURNS VARCHAR AS
$BODY$BEGIN

	  INSERT INTO bp.schedule_vm_container_lookup(container_id, hypervisor, name) 
	  SELECT p_container_id, p_hypervisor, p_name WHERE 
	  (SELECT TRUE FROM bp.schedule_vm_container_lookup WHERE container_id = p_container_id) IS NULL;
	  RETURN p_container_id||' - '||p_name;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

-- Add application entries
-- See bpdb.sql for details
SELECT bp.schedule_vm_container_lookup(4, 'VMware', 'Datacenters');
SELECT bp.schedule_vm_container_lookup(5, 'VMware', 'ESX Clusters');
SELECT bp.schedule_vm_container_lookup(6, 'VMware', 'ESX Servers');
SELECT bp.schedule_vm_container_lookup(7, 'VMware', 'Datastores');
SELECT bp.schedule_vm_container_lookup(8, 'VMware', 'Resource Pools');
SELECT bp.schedule_vm_container_lookup(9, 'VMware',    'vApps');
SELECT bp.schedule_vm_container_lookup(10, 'VMware',    'Folders');
SELECT bp.schedule_vm_container_lookup(11, 'VMware',    'VM DisplayName');

DROP FUNCTION bp.schedule_vm_container_lookup(INTEGER, VARCHAR, VARCHAR);

CREATE OR REPLACE FUNCTION bp.schedule_vm_regex_lookup(p_regex_action_id INTEGER, p_name VARCHAR)
RETURNS VARCHAR AS
$BODY$BEGIN

	  INSERT INTO bp.schedule_vm_regex_lookup(regex_action_id, name) 
	  SELECT p_regex_action_id, p_name WHERE 
	  (SELECT TRUE FROM bp.schedule_vm_regex_lookup WHERE regex_action_id = p_regex_action_id) IS NULL;
	  RETURN p_regex_action_id||' - '||p_name;

END$BODY$
LANGUAGE 'plpgsql' VOLATILE;

-- Add application entries
-- See bpdb.sql for details
SELECT bp.schedule_vm_regex_lookup(0, 'Equal');
SELECT bp.schedule_vm_regex_lookup(1, 'Not Equal');
SELECT bp.schedule_vm_regex_lookup(2, 'Contains');
SELECT bp.schedule_vm_regex_lookup(3, 'Not Contains');
SELECT bp.schedule_vm_regex_lookup(4, 'Starts with');
SELECT bp.schedule_vm_regex_lookup(5, 'Ends With');

DROP FUNCTION bp.schedule_vm_regex_lookup(INTEGER, VARCHAR);


